{
Copyright (C) 2002-2008 Massimo Melina (www.rejetto.com)

This file is part of HFS ~ HTTP File Server.

    HFS is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    HFS is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with HFS; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
}
unit editDlg;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ComCtrls, ToolWin, StdCtrls, utilLib, Menus, ExtCtrls, strUtils;

type
  TeditFrm = class(TForm)
    ToolBar1: TToolBar;
    okBtn: TToolButton;
    undoBtn: TToolButton;
    defaultBtn: TToolButton;
    cancelBtn: TToolButton;
    helpBtn: TToolButton;
    applyBtn: TToolButton;
    importBtn: TToolButton;
    ToolButton3: TToolButton;
    ToolButton2: TToolButton;
    exportBtn: TToolButton;
    ToolButton4: TToolButton;
    ToolButton5: TToolButton;
    box: TMemo;
    wrapChk: TCheckBox;
    sbar: TStatusBar;
    findBox: TEdit;
    Label1: TLabel;
    goBtn: TToolButton;
    fontBtn: TToolButton;
    fontDlg: TFontDialog;
    ToolButton1: TToolButton;
    procedure okBtnClick(Sender: TObject);
    procedure cancelBtnClick(Sender: TObject);
    procedure undoBtnClick(Sender: TObject);
    procedure defaultBtnClick(Sender: TObject);
    procedure helpBtnClick(Sender: TObject);
    procedure applyBtnClick(Sender: TObject);
    procedure importBtnClick(Sender: TObject);
    procedure exportBtnClick(Sender: TObject);
    procedure wrapChkClick(Sender: TObject);
    procedure findBoxKeyPress(Sender: TObject; var Key: Char);
    procedure goBtnClick(Sender: TObject);
    procedure fontBtnClick(Sender: TObject);
    procedure boxKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
  public
    default: string;
    help: string;
    helpFrm: Tform;
    onSetText: procedure(sender:Tobject; text:string);
    procedure findNext();
    procedure updateSbar();
  end;

implementation

{$R *.dfm}

procedure TeditFrm.okBtnClick(Sender: TObject);
begin
if assigned(onSetText) then onSetText(self, box.text);
close();
end;

procedure TeditFrm.cancelBtnClick(Sender: TObject);
begin close() end;

procedure TeditFrm.undoBtnClick(Sender: TObject);
begin box.Undo() end;

procedure TeditFrm.defaultBtnClick(Sender: TObject);
begin box.text:=default end;

procedure TeditFrm.helpBtnClick(Sender: TObject);
var
  re: TRichEdit;
begin
if assigned(helpFrm) then
  begin
  helpFrm.show;
  exit;
  end;
helpFrm:=Tform.create(self);
helpFrm.Width:=620;
helpFrm.height:=300;
helpFrm.caption:='HELP';
re:=TRichEdit.create(helpFrm);
helpFrm.InsertControl(re);
re.Align:=alClient;
re.ReadOnly:=TRUE;
re.ScrollBars:=ssBoth;
re.Text:='prova';
copyRTFtoRichEdit(help, re);
helpFrm.show;
end;

procedure TeditFrm.applyBtnClick(Sender: TObject);
begin if assigned(onSetText) then onSetText(self, box.text) end;

procedure TeditFrm.importBtnClick(Sender: TObject);
var
  dlg: TopenDialog;
  s: string;
begin
dlg:=TopenDialog.create(self);
dlg.DefaultExt:='tpl';
dlg.Filter:='Template|*.tpl';
if dlg.Execute then
  begin
  s:=loadFile(dlg.FileName);
  if AnsiStartsStr(#$EF#$BB#$BF, s) then // UTF8 BOM
    begin
    delete(s,1,3);
    s:=UTF8toAnsi(s);
    end
  else
    if utf8test(s) then
      s:=firstNonEmpty(UTF8toAnsi(s), s);
  box.text:=s;
  end;
dlg.free;
end;

procedure TeditFrm.exportBtnClick(Sender: TObject);
var
  dlg: TSaveDialog;
begin
dlg:=TSaveDialog.create(self);
dlg.DefaultExt:='tpl';
dlg.Filter:='Template|*.tpl';
if dlg.Execute then
  savefile(dlg.FileName, optUTF8(utf8test(box.text), box.text));
dlg.free;
end;

procedure TeditFrm.findNext();
var
  i: integer;
  pattern, full: string;
begin
pattern:=lowercase(findBox.Text);
full:=lowercase(box.text);
if pattern = '' then exit;
i:=posEx(pattern, full, box.SelStart+box.SelLength);
if i = 0 then
  begin
  i:=pos(pattern, full);
  if i = 0 then
    begin
    msgDlg('Not found');
    exit;
    end;
  if msgDlg('Not found.'#13'Restart from the beginning?', MB_YESNO) = IDNO then
    exit;
  end;
box.SelStart:=i-1;
box.SelLength:=length(pattern);
end; // findNext

procedure TeditFrm.wrapChkClick(Sender: TObject);
var
  bak: integer;
begin
bak:=box.SelStart;
box.WordWrap:=wrapChk.Checked;
if box.wordwrap then box.ScrollBars:=ssVertical
else box.ScrollBars:=ssBoth;
box.SelStart:=bak;
end;

procedure TeditFrm.updateSbar();
begin
sbar.SimpleText:=format('Bytes: %d - Line: %d - Col: %d',
  [length(box.text), 1+box.caretpos.y, 1+box.caretpos.x]);
end;

procedure TeditFrm.findBoxKeyPress(Sender: TObject; var Key: Char);
begin
if key = #13 then
  begin
  key:=#0;
  findNext();
  end;
end;

procedure TeditFrm.goBtnClick(Sender: TObject);
begin findNext() end;

procedure TeditFrm.fontBtnClick(Sender: TObject);
begin
fontDlg.Font.assign(box.Font);
if not fontDlg.Execute() then exit;
box.Font.Assign(fontDlg.font);
end;

procedure TeditFrm.boxKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
if shift = [ssCtrl] then
  if key = ord('A') then
    box.SelectAll();
if key = VK_F3 then
  findNext()
end;

end.
