{
Copyright (C) 2002-2008 Massimo Melina (www.rejetto.com)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


HTTP Server Lib

==== TO DO
* upload bandwidth control (can it be done without multi-threading?)

}
{$I- }

unit HSlib;

interface

uses
  wsocket, classes, messages, winprocs, forms, extctrls, sysutils, contnrs;

const
  VERSION = '2.5.1';

type
  ThttpSrv=class;

  ThttpConn=class;

  ThttpMethod=( HM_UNK, HM_GET, HM_POST, HM_HEAD );

  ThttpEvent=(
    HE_OPEN,            // server is listening
    HE_CLOSE,           // server does not listen anymore
    HE_CONNECTED,       // a client just connected
    HE_DISCONNECTED,    // client communication terminated
    HE_GOT,             // other peer sent sth
    HE_SENT,            // we sent sth
    HE_REQUESTED,       // a full request has been submitted
    HE_REPLIED,         // the reply has been sent
    HE_POST_FILE,       // new file is posted
    HE_POST_MORE_FILE,  // more data has come for the previous file
    HE_POST_END_FILE,   // last file done
    HE_POST_END,        // POST section terminated
    HE_LAST_BYTE_DONE,  // useful to count full downloads
    HE_CANT_OPEN_FILE   // error
  );

  ThttpConnState=(
    HCS_IDLE,               // connected but idle
    HCS_REQUESTING,         // getting request
    HCS_POSTING,            // getting post data
    HCS_REPLYING,           // a reply is pending
    HCS_REPLYING_HEADER,    // sending header
    HCS_REPLYING_BODY,      // sending body
    HCS_DISCONNECTED        // disconnected
  );

  ThttpReplyMode=(
    HRM_REPLY,              // reply header+body
    HRM_REPLY_HEADER,       // reply header only
    HRM_DENY,               // answer a deny code
    HRM_UNAUTHORIZED,       // bad user/pwd
    HRM_NOT_FOUND,          // answer a not-found code
    HRM_BAD_REQUEST,        // answer a bad-request code
    HRM_INTERNAL_ERROR,     // answer an internal-error code
    HRM_CLOSE,              // close connection with no reply
    HRM_IGNORE,             // does nothing, connection remains open
    HRM_METHOD_NOT_ALLOWED, // answer a method-not-allowed code
    HRM_REDIRECT,           // redirection to another URL
    HRM_OVERLOAD,           // server is overloaded, retry later
    HRM_TOO_LARGE,          // the request has exceeded the max length allowed
    HRM_MOVED,              // moved permanently to another url
    HRM_NOT_MODIFIED        // use the one in your cache, client
  );

  ThttpReply = record
    mode: ThttpReplyMode;
    header: string;            // full raw header (optional)
    contentType: string;       // ContentType header (optional)
    additionalHeaders: string; // these are appended to predefined headers (opt)
    bodyMode :(
      RBM_FILE,         // variable body specifies a file
      RBM_STRING,       // variable body specifies byte content
      RBM_STREAM        // refer to bodyStream
    );
    body: string;    // specifies reply body according to bodyMode
    bodyStream: Tstream;   // note: the stream is automatically freed 
    firstByte, lastByte: int64;  // body interval for partial replies (206)
    realm: string;   // this will appear in the authentication dialog
    url: string;     // used for redirections
    reason: string;  // customized reason phrase
    end;

  ThttpRequest = record
    full: string;           // the raw request, byte by byte
    method: ThttpMethod;
    url: string;
    ver: string;
    firstByte, lastByte: int64;  // body interval for partial requests
    headers: string;        // http headers, separated by CRLF
    user,pwd: string;
    end;

  ThttpPost = record
    boundary: string;       // multipart form-data boundary
    length: int64;        // multipart form-data length
    header: string;         // contextual header
    data: string;           // post data chunk
    filename: string;       // name of posted file
    end;

  TspeedLimiter = class
    maxSpeed, availableBandwidth: integer;
    constructor create(max:integer=MAXINT);
    end;

  ThttpConn = class
  protected
    srv: ThttpSrv;        // reference to the server
    stream: Tstream;
    P_address: string;
    P_port: string;
    brecvd: int64;          // bytes received from the client
    bsent: int64;           // bytes sent to the client
    bsent_body: int64;      // bytes sent to the client (current body only)
    bsent_bodies: int64;    // bytes sent to the client (for all bodies)
    P_requestCount: integer;
    P_destroying: boolean;  // destroying is in progress
    persistent: boolean;
    disconnecting: boolean;  // disconnected() has been called
    dontFulFil: boolean;
    firstPostFile: boolean;
    lastPostItemStart: int64;
    // post handling
    inBoundaries: boolean;    // we are between form-data boundaries
    postDataReceived: int64; // bytes received in post data
    // used to calculate actual speed
    lastBsent, lastBrecvd: int64;
    lastSpeedTime: int64;
    P_speedOut, P_speedIn: real;

    buffer: string;       // internal buffer for incoming data
    // event handlers
    procedure disconnected(Sender: TObject; Error: Word);
    procedure dataavailable(Sender: TObject; Error: Word);
    procedure senddata(sender:Tobject; bytes:integer);
    procedure datasent(sender:Tobject; error:word);
    function  fullBodySize():int64;
    function  partialBodySize():int64;
    function  initInputStream():boolean;
    function  sendNextChunk(max:integer=MAXINT):integer;
    function  getBytesToSend():int64;
    function  getBytesToPost():int64;
    function  getBytesGot():int64;
    function  getPostedLastItem():int64;
    procedure notify(ev:ThttpEvent);
    procedure calculateSpeed();
	  procedure sendheader(h:string='');
		function  replyHeader_mode(mode:ThttpReplyMode):string;
		function  replyHeader_code(code:integer):string;
    procedure processInputBuffer();
    procedure clearPostPart();
    procedure clearRequest();
    procedure clearReply();
  public
    sock: Twsocket;             // client-server communication socket
    state: ThttpConnState;      // what is doing now with this
    request: ThttpRequest;      // it requests
    reply: ThttpReply;          // we serve
    post: ThttpPost;            // it posts
    data: pointer;              // user data
    paused: boolean;            // while (not paused) do senddata()
    ignoreSpeedLimit: boolean;
    speedLimiter: TspeedLimiter;
    constructor create(server:ThttpSrv);
    destructor Destroy; override;
    procedure disconnect();
    procedure addHeader(s:string); // append an additional header line
    function  getHeader(h:string):string;  // extract the value associated to
                                           // the header from request.headers
    property address:string read P_address;      // other peer ip address
    property port:string read P_port;            // other peer port
    property requestCount:integer read P_requestCount;
    property bytesToSend:int64 read getBytesToSend;
    property bytesToPost:int64 read getBytesToPost;
    property bytesSent:int64 read bsent_bodies;
    property bytesSentLastItem:int64 read bsent_body;
    property bytesPartial:int64 read partialBodySize;
    property bytesFullBody:int64 read fullBodySize;
    property bytesGot:int64 read getBytesGot;
    property bytesPosted:int64 read postDataReceived;
    property bytesPostedLastItem:int64 read getPostedLastItem;
    property speedIn:real read P_speedIn;  // (bytes_recvd/s)
    property speedOut:real read P_speedOut;  // (bytes_sent/s)
    property disconnectedByServer:boolean read disconnecting;
    property destroying:boolean read P_destroying;
    end;

  ThttpSrv = class
  protected
    timer: Ttimer;
    lockTimerevent: boolean;
    lastHertz: Tdatetime;

    P_port: string;
    P_autoFree: boolean;
    P_speedIn, P_speedOut: real;
    P_maxSpeed: integer;  // limit B/s
    bsent, brecvd: int64;
    freq: int64;
    availableBandwidth: integer;
    procedure setPort(v:string);
    function  getActive():boolean;
    procedure setActive(v:boolean);
    procedure connected(Sender: TObject; Error: Word);
    procedure disconnected(Sender: TObject; Error: Word);
    procedure bgexception(Sender: TObject; E:Exception; var CanClose:Boolean);
    procedure setAutoFree(v:boolean);
    procedure setMaxSpeed(v:integer);
    procedure notify(ev:ThttpEvent; conn:ThttpConn);
    procedure hertzEvent();
    procedure timerEvent(sender:Tobject);
    procedure calculateSpeed();
    procedure processDisconnecting();
  public
    sock: Twsocket;     // listening socket
    conns: Tlist;       // full list of connected clients
    disconnecting: Tlist; // list of pending disconnections
    offlines: Tlist;    // disconnected clients to be freed
    q: Tlist;           // clients waiting for data to be sent
    data: pointer;      // user data
    speedLimiters: TobjectList;
    persistentConnections: boolean;  // disconnect clients after they're served
    onEvent: procedure(event:ThttpEvent; conn:ThttpConn) of object;
    constructor create(); overload;
    destructor Destroy(); override;
    property active:boolean read getActive write setActive; // r we listening?
    property port:string read P_port write setPort;
    property bytesSent:int64 read bsent;
    property bytesReceived:int64 read brecvd;
    property speedIn:real read P_speedIn;  // (bytes_recvd/s)
    property speedOut:real read P_speedOut;  // (bytes_sent/s)
    property autoFreeDisconnectedClients: boolean read P_autoFree write setAutoFree;
    property maxSpeed: integer read P_maxSpeed write setMaxSpeed;
    function start(onAddress:string='*'):boolean; // returns true if all is ok
    procedure stop();
    procedure disconnectAll(wait:boolean=FALSE);
    procedure freeConnList(l:Tlist);
    end;

const
  TIMER_HZ = 100;
  MINIMUM_CHUNK_SIZE = 2*1024;
  MAXIMUM_CHUNK_SIZE = 512*1024;
  HRM2CODE: array [ThttpReplyMode] of integer = (200, 200, 403, 401, 404, 400,
  	500, 0, 0, 405, 302, 503, 414, 301, 304 );
  METHOD2STR: array [ThttpMethod] of string = ('UNK','GET','POST','HEAD');
  HRM2STR: array [ThttpReplyMode] of string = ('Head+Body', 'Head only', 'Deny',
    'Unauthorized', 'Not found', 'Bad request', 'Internal error', 'Close',
    'Ignore', 'Unallowed method', 'Redirect', 'Overload', 'Request too large',
    'Moved permanently', 'Not Modified');

{ split S in position where SS is found, the first part is returned
  the second part following SS is left in S }
function chop(ss:string; var s:string):string; overload;
// same as before, but separator is I
function chop(i:integer; var s:string):string; overload;
// same as before, but specifying separator length
function chop(i, l:integer; var s:string):string; overload;
// same as chop(lineterminator, s)
function chopLine(var s:string):string; overload;
// decode/decode url
function decodeURL(url:string; utf8:boolean=TRUE):string;
function encodeURL(url:string; nonascii:boolean=TRUE; spaces:boolean=TRUE;
  unicode:boolean=FALSE):string;
// returns true if address is not suitable for the internet
function isLocalIP(ip:string):boolean;
// base64 encoding
function base64encode(s:string):string;
function base64decode(s:string):string;
// an ip address where we are listening
function getIP():string;
// ensure a string ends with a specific string
procedure includeTrailingString(var s:string; ss:string);

implementation

uses
  Windows;
const
  CRLF = #13#10;
  MAX_REQUEST_LENGTH = 2*1024;
  MAX_INPUT_BUFFER_LENGTH = 256*1024;
  // used as body content when the user did not specify any
  HRM2BODY: array [ThttpReplyMode] of string = (
  	'200 - OK',
    '200 - OK (header only)',
    '403 - You are not allowed to access this file',
    '401 - You are not authorized to access this file',
    '404 - File not found',
    '400 - Bad request',
    '500 - Internal server error',
    '',
    '',
    '405 - Method not allowed',
    '302 - <a href="%url%">Redirection to %url%</a>'
    +'<script language=javascript>window.location="%url%"</script>',
    '503 - Server is overloaded, retry later',
    '414 - The request has exceeded the max length allowed',
    '301 - Moved permanently to <a href="%url%">%url%</a>',
    '' // RFC2616: The 304 response MUST NOT contain a message-body
  );

procedure includeTrailingString(var s:string; ss:string);
begin if copy(s, length(s)-length(ss)+1, length(ss)) <> ss then s:=s+ss end;

function charToUnicode(c:char):dword;
begin stringToWideChar(c,@result,4) end;

function isLocalIP(ip:string):boolean;
var
  r: record d,c,b,a:byte end;
begin
dword(r):=WSocket_ntohl(WSocket_inet_addr(pchar(ip)));
result:=(r.a in [0,10,23,127])
  or (r.a = 192) and ((r.b = 168) or (r.b = 0) and (r.c = 2))
  or (r.a = 169) and (r.b = 254)
  or (r.a = 172) and (r.b in [16..31])
end; // isLocalIP

function ifThen(c:boolean; a:string; b:string=''):string; overload;
begin if c then result:=a else result:=b end;

function ifThen(c:boolean; a:integer; b:integer=0):integer; overload;
begin if c then result:=a else result:=b end;

function min(a,b:integer):integer;
begin if a < b then result:=a else result:=b end;

function base64encode(s:string):string;
const
  TABLE='ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/';
type
  Ttriple=array [0..2] of byte;
var
  p: ^Ttriple;
  i: integer;
begin
result:='';
p:=@s[1];
for i:=1 to length(s) div 3 do
  begin
  result:=result+TABLE[1+p[0] shr 2]
  	+TABLE[1+(p[0] and 3) shl 4+p[1] shr 4]
    +TABLE[1+(p[1] and 15) shl 2+p[2] shr 6]
    +TABLE[1+(p[2] and 63)];
  inc(p);
  end;
if length(s) mod 3 > 0 then
	result:=result+TABLE[1+p[0] shr 2]+TABLE[1+(p[0] and 3) shl 4+p[1] shr 4]
    +ifThen(length(s) mod 3=1,'==',TABLE[1+(p[1] and 15) shl 2+p[2] shr 6]+'=');
end; // base64encode

function base64decode(s:string):string;
const
  TABLE:array[#43..#122] of byte=(
  	62,0,0,0,63,52,53,54,55,56,57,58,59,60,61,0,0,0,0,0,0,0,0,1,2,3,4,5,6,7,
    8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,0,0,0,0,0,0,26,27,28,
    29,30,31,32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,48,49,50,51);
var
  i: integer;
begin
result:='';
i:=1;
while i <= length(s) do
  begin
	result:=result
  	+chr(TABLE[s[i]] shl 2+TABLE[s[i+1]] shr 4)
    +ifThen(s[i+2]<>'=', chr(TABLE[s[i+1]] shl 4+TABLE[s[i+2]] shr 2))
    +ifThen(s[i+3]<>'=', chr(TABLE[s[i+2]] shl 6+TABLE[s[i+3]]));
  inc(i,4);
  end;
end; // base64decode

function decodeURL(url:string; utf8:boolean=TRUE):string;
var
  i, l: integer;
  c: char;
begin
setLength(result, length(url));
l:=0;
i:=1;
while i<=length(url) do
  begin
  if (url[i] = '%') and (i+2 <= length(url)) then
    try
      c:=char(strToInt( '$'+url[i+1]+url[i+2] ));
      inc(i,2);
    except c:=url[i]
    end
  else
    c:=url[i];

  inc(i);
  inc(l);
  result[l]:=c;
  end;
setLength(result, l);
if utf8 then
  begin
  url:=utf8ToAnsi(result);
  // if the string is not UTF8 compliant, the result is empty
  if url > '' then result:=url;
  end;
end; // decodeURL

function encodeURL(url:string; nonascii:boolean=TRUE; spaces:boolean=TRUE;
  unicode:boolean=FALSE):string;
var
  i: integer;
  encodePerc, encodeUni: set of char;
begin
result:='';
encodeUni:=[];
if nonascii then encodeUni:=[#128..#255];
encodePerc:=[#0..#31,'#','%','+','$','&',':',';','=','?','@','|','"','^','`',
  ',','<','>','{','}'];
if spaces then include(encodePerc,' ');
if not unicode then
  begin
  encodePerc:=encodePerc+encodeUni;
  encodeUni:=[];
  end;
for i:=1 to length(url) do
	if url[i] in encodePerc then
    result:=result+'%'+intToHex(ord(url[i]),2)
  else if url[i] in encodeUni then
    result:=result+'&#'+intToStr(charToUnicode(url[i]))+';'
  else
    result:=result+url[i];
end; // encodeURL

function getIP():string;
var
  i: integer;
  ips: Tstrings;
begin
ips:=LocalIPlist();
case ips.count of
  0: result:='';
  1: result:=ips[0];
  else
    i:=0;
    while (i < ips.count-1) and isLocalIP(ips[i]) do
      inc(i);
    result:=ips[i];
  end;
end; // getIP

function replyHeader_IntPositive(name:string; int:int64):string;
begin
result:='';
if int >= 0 then result:=name+': '+intToStr(int)+CRLF;
end;

function replyHeader_Str(name:string; str:string):string;
begin
result:='';
if str > '' then result:=name+': '+str+CRLF;
end;

function chop(i, l:integer; var s:string):string; overload;
begin
if i=0 then
  begin
  result:=s;
  s:='';
  exit;
  end;
result:=copy(s,1,i-1);
delete(s,1,i-1+l);
end; // chop

function chop(ss:string; var s:string):string;
begin result:=chop(pos(ss,s),length(ss),s) end;

function chop(i:integer; var s:string):string;
begin result:=chop(i,1,s) end;

function chopLine(var s:string):string;
begin
result:=chop(#10,s);
if (result>'') and (result[length(result)]=#13) then
  setlength(result, length(result)-1);
end; // chopline

/////// SERVER

function ThttpSrv.start(onAddress:string='*'):boolean;
begin
result:=FALSE;
if active or not assigned(sock) then exit;
try
  if onAddress = '' then onAddress:='*';
  if (onAddress = '') or (onAddress = '*') then sock.addr:='0.0.0.0'
  else sock.addr:=onAddress;
  sock.port:=port;
  sock.proto:='6';
  sock.listen();
  P_port:=sock.getxport();
  result:=TRUE;
  notify(HE_OPEN, NIL);
except
  end;
end; // start

procedure ThttpSrv.stop();
begin if assigned(sock) then sock.Close() end;

procedure ThttpSrv.connected(Sender: TObject; Error: Word);
begin if error=0 then ThttpConn.create(self) end;

procedure ThttpSrv.disconnected(Sender: TObject; Error: Word);
begin notify(HE_CLOSE, NIL) end;

constructor ThttpSrv.create();
begin
sock:=TWSocket.create(NIL);
sock.OnSessionAvailable:=connected;
sock.OnSessionClosed:=disconnected;
sock.OnBgException:=bgexception;
conns:=Tlist.create;
offlines:=Tlist.create;
q:=Tlist.create;
disconnecting:=Tlist.create;
maxSpeed:=MAXINT;
QueryPerformanceFrequency(freq);
timer:=Ttimer.create(NIL);
timer.OnTimer:=timerEvent;
timer.Interval:=1000 div TIMER_HZ;
timer.Enabled:=TRUE;
Port:='80';
autoFreeDisconnectedClients:=TRUE;
persistentConnections:=TRUE;
speedLimiters:=TobjectList.create();
end; // create

destructor ThttpSrv.destroy();
begin
freeAndNIL(timer);
// it should be safe enough to mute exceptions here
try sock.Close() except end;
disconnectAll(TRUE);
freeAndNIL(sock);
freeConnList(conns);
processDisconnecting();
// the following line is causing some AV, and it's not truly needed, is it?
//freeConnList(offlines);
freeAndNIL(conns);
freeAndNIL(disconnecting);
freeAndNIL(offlines);
freeAndNIL(q);
end; // destroy

procedure ThttpSrv.hertzEvent();
var
  i: integer;
begin
if now()-lastHertz < 1/(24*60*60) then exit;
lastHertz:=now();
calculateSpeed();
availableBandwidth:=maxSpeed;
for i:=0 to speedLimiters.Count-1 do
  try
    with speedLimiters[i] as TspeedLimiter do
      availableBandwidth:=maxSpeed;
  except end;
end; // hertzEvent

procedure ThttpSrv.processDisconnecting();
var
  c: ThttpConn;
begin
while disconnecting.Count > 0 do
  begin
  c:=disconnecting.first();
  disconnecting.delete(0);
  c.state:=HCS_DISCONNECTED;
  q.remove(c);
  conns.remove(c);
  offlines.Add(c);
  notify(HE_DISCONNECTED, c);
  end;
end; // processDisconnecting

procedure ThttpSrv.timerEvent(sender:Tobject);

  procedure processPipelines();
  var
    i: integer;
  begin
  i:=conns.Count;
  while i > 0 do
    begin
    dec(i);
    with ThttpConn(conns[i]) do
      if (state in [HCS_IDLE, HCS_DISCONNECTED]) and (buffer > '') then
        processInputBuffer();
    end;
  end; // processPipelines

  procedure processQ();
  var
    c: ThttpConn;
    toQ: Tlist;
    chunkLimit: integer;
  begin
  toQ:=Tlist.create;
  while q.count > 0 do
    begin
    c:=q.first();
    q.delete(0);
    if c = NIL then continue;
    try
      chunkLimit:=ifThen(c.paused, 0, MAXINT);
      if not c.ignoreSpeedLimit then
        begin
        if assigned(c.speedLimiter) then
          chunkLimit:=min(chunkLimit, c.speedLimiter.availableBandwidth);
        if availableBandwidth >= 0 then
          chunkLimit:=min(chunkLimit, availableBandwidth);
        end;
      if chunkLimit <= 0 then
        begin
        toQ.Add(c);
        continue;
        end;
      // serve the pending connection with a data chunk
      if not c.destroying and (c.state <> HCS_DISCONNECTED)
      and assigned(c.sock) and (c.sock.State = wsConnected) then
        begin
        chunkLimit:=c.sendNextChunk(chunkLimit);
        dec(availableBandwidth, chunkLimit);
        if assigned(c.speedLimiter) then
          dec(c.speedLimiter.availableBandwidth, chunkLimit);
        end;
    except end;
    end;
  q.assign(toQ, laOR); 
  toQ.Free;
  end; // processQ

begin
hertzEvent();

lockTimerevent:=TRUE;
try
  processDisconnecting();
  if autoFreeDisconnectedClients then freeConnList(offlines);
  processPipelines();
  processQ();
finally
  lockTimerevent:=FALSE
  end;
end; // timerEvent

procedure ThttpSrv.notify(ev:ThttpEvent; conn:ThttpConn);
begin if assigned(onEvent) then onEvent(ev, conn) end;

function Thttpsrv.getActive():boolean;
begin result:=assigned(sock) and (sock.State=wsListening) end;

procedure ThttpSrv.setActive(v:boolean);
begin
if v = active then exit;
if v then start() else stop()
end; // setactive

procedure ThttpSrv.freeConnList(l:Tlist);
var
  c: ThttpConn;
begin
while l.count > 0 do
  try
    c:=l.first();
    l.delete(0);
    try c.free except end;
  finally
  end;
end; // freeConnList

procedure ThttpSrv.calculateSpeed();
var
  i: integer;
begin
P_speedOut:=0;
P_speedIn:=0;
i:=0;
while i < conns.count do
  begin
  ThttpConn(conns[i]).calculateSpeed();
  P_speedOut:=P_speedOut+ThttpConn(conns[i]).speedOut;
  P_speedIn:=P_speedIn+ThttpConn(conns[i]).speedIn;
  inc(i);
  end;
end; // calculateSpeed

procedure ThttpSrv.setPort(v:string);
begin
if active then
  raise Exception.Create(classname+': cannot change port while active');
P_port:=v
end; // setPort

procedure ThttpSrv.disconnectAll(wait:boolean=FALSE);
var
  i: integer;
  clone: Tlist;
begin
// on disconnection <conns> list changes. clone it for safer enumeration.
clone:=Tlist.Create;
clone.Assign(conns);
// cast disconnection
for i:=0 to clone.count-1 do
  ThttpConn(clone[i]).disconnect();
if wait then
  for i:=0 to clone.count-1 do
    if conns.IndexOf(clone[i]) >= 0 then
      ThttpConn(clone[i]).sock.WaitForClose();
clone.free;
end; // disconnectAll

procedure ThttpSrv.setAutoFree(v:boolean);
begin P_autofree:=v end;

procedure ThttpSrv.setMaxSpeed(v:integer);
begin if v < 0 then P_maxSpeed:=MAXINT else P_maxSpeed:=v end;

procedure ThttpSrv.bgexception(Sender: TObject; E: Exception;
  var CanClose: Boolean);
begin
canClose:=FALSE;
end;

////////// CLIENT

constructor ThttpConn.create(server:ThttpSrv);
begin
// init socket
sock:=Twsocket.create(NIL);
sock.Dup(server.sock.Accept);
sock.OnDataAvailable:=dataavailable;
sock.OnSessionClosed:=disconnected;
sock.onSendData:=senddata;
sock.onDataSent:=datasent;
sock.LineMode:=FALSE;
sock.flushtimeout:=0;

P_address:=sock.GetPeerAddr();
P_port:=sock.GetPeerPort();
state:=HCS_IDLE;
srv:=server;
srv.conns.Add(self);
clearRequest();
clearReply();
QueryPerformanceCounter(lastSpeedTime);

server.notify(HE_CONNECTED, self);
if reply.mode <> HRM_CLOSE then exit;
dontFulFil:=TRUE;
disconnect();
end;

destructor ThttpConn.destroy;
begin
P_destroying:=TRUE;
if assigned(sock) then
  begin
  sock.Shutdown(SD_BOTH);
  sock.WaitForClose();
  end;
srv.offlines.remove(self);
freeAndNil(stream);
freeAndNIL(sock);
end; // destroy

procedure ThttpConn.calculateSpeed();
var
	bytes: int64;
  now: int64;
  elapsed: Tdatetime;
begin
QueryPerformanceCounter(now);
elapsed:=(now-lastSpeedTime)/srv.freq;
lastSpeedTime:=now;

bytes:=bsent-lastBsent;
lastBsent:=bsent;
P_speedOut:=bytes/elapsed;

bytes:=brecvd-lastBrecvd;
lastBrecvd:=brecvd;
P_speedIn:=bytes/elapsed;
end; // calculateSpeed

procedure ThttpConn.disconnected(Sender: TObject; Error: Word);
begin srv.disconnecting.Add(self) end;

function ThttpConn.getHeader(h:string):string;
var
  s,l:string;
begin
result:='';
if request.method = HM_UNK then exit;
s:=request.headers;
  repeat
  l:=chopline(s);
  if compareText(chop(':',l), h) = 0 then
    begin
    if (l>'') and (l[1]=' ') then delete(l,1,1);
    result:=l;
    exit;
    end;
  until s='';
end; // getHeader

procedure ThttpConn.clearRequest();
begin
request.method:=HM_UNK;
request.ver:='';
request.url:='';
request.firstByte:=-1;
request.lastByte:=-1;
request.headers:='';
request.user:='';
request.pwd:='';
end; // clearRequest

procedure ThttpConn.clearPostPart();
begin
post.data:='';
post.header:='';
post.filename:='';
end; // clearPostPart

procedure ThttpConn.clearReply();
begin
reply.header:='';
reply.bodyMode:=RBM_STRING;
reply.body:='';
reply.additionalHeaders:='';
reply.mode:=HRM_IGNORE;
reply.firstByte:=request.firstByte;
reply.lastByte:=request.lastByte;
reply.realm:='Password protected resource';
reply.reason:='';
end; // clearReply

procedure ThttpConn.processInputBuffer();

  function parseHeader():boolean;
  var
    r, s: string;
    i: integer;
  begin
  result:=FALSE;
  r:=request.full;

  // find first blank space
  for i:=1 to 10 do
    if i > length(r) then exit
    else if r[i]=' ' then break;

  clearRequest();

  s:=uppercase(chop(i, r));
  if s='GET' then request.method:=HM_GET else
  if s='POST' then request.method:=HM_POST else
  if s='HEAD' then request.method:=HM_HEAD else
    exit; // unhandled method or invalid http request

  request.url:=chop(' ', r);

  s:=uppercase(chopLine(r));
  // if 'HTTP/' is not found, chop returns S
  if chop('HTTP/',s) = '' then request.ver:=s;

  request.headers:=r;

  s:=getHeader('Range');
  if pos('bytes=',s) = 1 then
    begin
    chop('bytes=',s);
    r:=chop('-',s);
    try
      if r>'' then request.firstByte:=strToInt64(r);
      if s>'' then request.lastByte:=strToInt64(s);
    except end;
    end;

  s:=getHeader('Authorization');
  if pos('Basic',s) = 1 then
    begin
    chop('Basic ',s);
    s:=base64decode(s);
    request.user:=chop(':',s);
    request.pwd:=s;
    end;

  s:=getHeader('Connection');
  persistent:=srv.persistentConnections and
    ((pos('Keep-Alive',s)=1) or (request.ver >= '1.1') and (pos('close',s)=0));

  s:=getHeader('Content-Type');
  if pos('multipart',s) = 1 then
    begin
    chop('boundary=',s);
    post.boundary:='--'+s;
    end;
  try
    post.length:=strToInt64(getHeader('Content-Length'));
    // the browser may not support 2GB+ files.
    // this workaround works only for files under 4GB.
    if post.length < 0 then inc(post.length, int64(2) shl 31);
  except
    end;

  result:=TRUE;
  end; // parseHeader

  procedure handlePostData();
  var
    i: integer;
    s, t: string;
  begin
    repeat
    { I tried to speed up the code below, for large files. When we have the
    { buffer stuffed with file bytes only, we can avoid calling pos() and
    { chunk(). Unexpectedly this did not speed up anything. CPU and throughput
    { did not change. I report the try here below so you don't waste your time.

    if (bytesPosted < post.length-length(post.boundary))
    and (post.filename > '') then
      begin
      post.data:=buffer;
      buffer:='';
      notify(HE_POST_MORE_FILE);
      break;
      end;
    }
    i:=pos(post.boundary, buffer);
    if i = 0 then
      begin
      if post.filename > '' then
        { we were already receiving a file, listener should be notified about this
        { data, but only the data we are sure doesn't overlap the boundary }
        begin
        post.data:=chop(length(buffer)-length(post.boundary), 0, buffer);
        if post.data > '' then notify(HE_POST_MORE_FILE);
        end;
      break;
      end;
    // extract the (sub)header. we consider it to be shorter than 1000.
    s:=copy(buffer, i+length(post.boundary), 1000);
    // the end has come
    if copy(s,1,4) = '--'+CRLF then
      begin
      chop('--'+CRLF, buffer);
      if not firstPostFile then notify(HE_POST_END_FILE);
      notify(HE_POST_END);
      state:=HCS_REPLYING;
      break;
      end;
    // we wait for the header to be complete
    if pos(CRLF+CRLF, s) = 0 then break;
    { the data is processed in a second step. the data processed below is
    { related to the previous header, still stored in post.header }
    post.data:=chop(i, length(post.boundary), buffer);
    // if data was found, we must chop the CRLF between data and boundary
    setlength(post.data, length(post.data)-2);
    if (post.data > '') and (post.filename > '') then notify(HE_POST_MORE_FILE);
    chopLine(buffer);
    clearPostPart();
    post.header:=chop(CRLF+CRLF, buffer);
    // parse the header part
    t:=post.header;
    while t > '' do
      begin
      s:=chopLine(t);
      if s = '' then break;
      chop('filename=',s);
      if (s > '') and (s[1] = '"') then s:=copy(s,2,length(s)-2); // dequote
      if s > '' then
        begin
        delete(s, 1, lastDelimiter('/\',s));
        post.filename:=s;
        break; // ATM we get only filename from the header, so we can break just after that
        end;
      end;
    if post.filename = '' then continue;
    // we get here upon a new file, so if there is a previous one, we must notify its end
    if not firstPostFile then notify(HE_POST_END_FILE);
    firstPostFile:=FALSE;
    notify(HE_POST_FILE);
    lastPostItemStart:=bytesPosted-length(buffer);
    until false;
  end; // handlePostData

  procedure handleHeaderData();
  var
    i, sepLen: integer;
  begin
  // try to identify header length and position
  i:=pos(CRLF+CRLF, buffer);
  sepLen:=4;
  if i <= 0 then
    begin
    i:=pos(#13#13, buffer);
    sepLen:=2;
    end;
  if i <= 0 then
    begin
    if (buffer>'') and (pos(#3,buffer)>0) then  // CTRL+C
      begin
      reply.mode:=HRM_CLOSE;
      disconnect();
      end;
    if length(buffer) > MAX_REQUEST_LENGTH then
      begin
      reply.mode:=HRM_TOO_LARGE;
      sendHeader(replyheader_mode(reply.mode));
      end;
    exit;
    end;
  request.full:=chop(i,sepLen,buffer);
  if not parseHeader() then exit;
  if request.method = HM_POST then
    begin
    state:=HCS_POSTING;
    firstPostFile:=TRUE;
    clearPostPart();
    end
  else state:=HCS_REPLYING;
  end; // handleHeaderData

  function replyHeader_OK(contentLength:int64=-1):string;
  begin
  result:=replyheader_code(200)
    +format('Content-Length: %d'+CRLF, [contentLength]);
  end; // replyHeader_OK

  function replyHeader_PARTIAL( firstB, lastB, totalB:int64):string;
  begin
  result:=replyheader_code(206)
    +format('Content-Range: bytes %d-%d/%d'+CRLF+'Content-Length: %d'+CRLF,
          [firstB, lastB, totalB, lastB-firstB+1 ])
  end; // replyheader_PARTIAL

begin
if state = HCS_IDLE then
  begin
  state:=HCS_REQUESTING;
  reply.contentType:='text/html';
  end;
case state of
  HCS_POSTING: handlePostData();
  HCS_REQUESTING:
    begin
    handleHeaderData();
    if state = HCS_POSTING then
      begin
      postDataReceived:=length(buffer);
      handlePostData();
      end;
    end;
  end;
if state <> HCS_REPLYING then exit;
// handle reply
clearReply();
inc(P_requestCount);
notify(HE_REQUESTED);
if not initInputStream() then
  begin
  reply.mode:=HRM_INTERNAL_ERROR;
  reply.contentType:='text/html';
  notify(HE_CANT_OPEN_FILE);
  end;
case reply.mode of
  HRM_CLOSE: disconnect();
  HRM_IGNORE: ;
  HRM_NOT_FOUND,
  HRM_BAD_REQUEST,
  HRM_METHOD_NOT_ALLOWED,
  HRM_INTERNAL_ERROR,
  HRM_OVERLOAD,
  HRM_NOT_MODIFIED,
  HRM_DENY: sendHeader( replyheader_mode(reply.mode) );
  HRM_UNAUTHORIZED:
    sendHeader(replyheader_mode(reply.mode)
      +replyHeader_Str('WWW-Authenticate','Basic realm="'+reply.realm+'"') );
  HRM_REDIRECT, HRM_MOVED:
    sendHeader(replyheader_mode(reply.mode)+'Location: '+reply.url );
  HRM_REPLY, HRM_REPLY_HEADER:
    if stream = NIL then
      sendHeader( replyHeader_code(404) )
    else
      if (request.firstByte >= bytesFullBody)
      or (request.lastByte >= bytesFullBody) then
        sendHeader( replyHeader_code(400) )
      else
        if reply.header > '' then
          sendHeader()
        else
          if (request.firstByte <= 0) and (request.lastByte < 0) then
            sendHeader( replyHeader_OK(bytesFullBody) )
          else
            with reply do
              sendHeader(replyHeader_PARTIAL(firstByte,lastByte,bytesFullBody));
  end;//case
end; // processInputBuffer

procedure ThttpConn.dataavailable(Sender: TObject; Error: Word);
var
  s: string;
begin
if error <> 0 then exit;
s:=sock.ReceiveStr();
inc(brecvd, length(s));
inc(srv.brecvd, length(s));
if (s = '') or dontFulFil then exit;
if state = HCS_POSTING then inc(postDataReceived, length(s));
buffer:=buffer+s;
if length(buffer) > MAX_INPUT_BUFFER_LENGTH then
  begin
  disconnect();
  sock.Abort(); // please, brutally
  exit;
  end;
notify(HE_GOT);
processInputBuffer();
end; // dataavailable

procedure ThttpConn.senddata(sender:Tobject; bytes:integer);
begin
if bytes <= 0 then exit;
inc(bsent, bytes);
inc(srv.bsent, bytes);
if state = HCS_REPLYING_BODY then
  begin
  inc(bsent_body, bytes);
  inc(bsent_bodies, bytes);
  end;
notify(HE_SENT);
end; // senddata

procedure ThttpConn.datasent(sender:Tobject; error:word);
var
  notifyReplied: boolean;
begin
if state in [HCS_IDLE, HCS_POSTING, HCS_REPLYING] then exit;
if (state = HCS_REPLYING_HEADER) and (reply.mode <> HRM_REPLY_HEADER) then
  begin
  state:=HCS_REPLYING_BODY;
  // set up a default body for errors with no body set
  if ((stream = NIL) or (stream.size = 0)) and (reply.mode <> HRM_REPLY) then
    begin
    reply.bodyMode:=RBM_STRING;
    reply.body:=HRM2BODY[reply.mode];
    if reply.mode in [HRM_REDIRECT, HRM_MOVED] then
      reply.body:=stringReplace(reply.body, '%url%', reply.url, [rfReplaceAll]);
    initInputStream();
    end;
  end;
if (state = HCS_REPLYING_BODY) and (bytesToSend > 0) then
  begin
  if (srv.maxSpeed = MAXINT) and not paused then sendNextChunk()
  else srv.q.Add(self);
  exit;
  end;
notifyReplied:=FALSE;
if (state in [HCS_REPLYING_HEADER, HCS_REPLYING_BODY, HCS_DISCONNECTED])
and (bytesSent > 0) and (bytesToSend = 0) then
  begin
  notifyReplied:=TRUE;
  state:=HCS_IDLE;
  end;

if not persistent or not (reply.mode in [HRM_REPLY, HRM_REPLY_HEADER]) then
  disconnect()
else
  { we must check the socket state, because a disconnection could happen while
  { this method is executing }
  if sock.State <> wsClosed then state:=HCS_IDLE;
if notifyReplied then
  begin
  if stream.Position = stream.size then notify(HE_LAST_BYTE_DONE);
  notify(HE_REPLIED);
  end;
// once the event has been notified, we reset the current counter
if state = HCS_IDLE then bsent_body:=0;

freeAndNil(stream);
end; // datasent

procedure ThttpConn.disconnect();
begin
if disconnecting then exit;
disconnecting:=TRUE;
if sock = NIL then exit;
sock.Shutdown(SD_BOTH);
sock.CloseDelayed();
end;

function ThttpConn.fullBodySize():int64;
begin if stream = NIL then result:=0 else result:=stream.Size end;

function ThttpConn.partialBodySize():int64;
begin
if (reply.lastByte<0) and (reply.firstByte<0) then result:=bytesFullBody
else result:=reply.lastByte-reply.firstByte+1
end; // partialBodySize

function ThttpConn.initInputStream():boolean;
var
  i: integer;
begin
result:=FALSE;
FreeAndNil(stream);
try
  case reply.bodyMode of
    RBM_STRING: stream:=TStringStream.create(reply.body);
    RBM_FILE:
      begin
      i:=fileopen(reply.body, fmOpenRead+fmShareDenyNone);
      if i = -1 then exit;
      stream:=TFileStream.Create(i);
      end;
    RBM_STREAM: stream:=reply.bodyStream;
    end;
  if (request.firstByte > 0) and (reply.mode = HRM_REPLY) then
    stream.Seek(request.firstByte, soBeginning);

  with reply do
    if (firstByte < 0) and (lastByte < 0) then
      begin
      firstByte:=0;
      lastbyte:=bytesFullBody-1;
      end
    else
      if lastByte < 0 then lastbyte:=bytesFullBody-1
      else
        if firstbyte < 0 then
          begin
          firstByte:=bytesFullBody-lastByte;
          lastByte:=bytesFullBody;
          end;
  result:=TRUE;
except end;
end; // initInputStream

function ThttpConn.sendNextChunk(max:integer=MAXINT):integer;
var
  n: int64;
  buf: string;
begin
result:=0;
if stream = NIL then exit;
n:=trunc(speedOut*1.5);
// the following line helps fast networks to reach max speed sooner
// in a test, a 3MB file has been downloaded locally at doubled speed
if (n = 0) or (bytesSentLastItem = 0) then n:=max;
if n > MAXIMUM_CHUNK_SIZE then n:=MAXIMUM_CHUNK_SIZE;
if n < MINIMUM_CHUNK_SIZE then n:=MINIMUM_CHUNK_SIZE;
if n > max then n:=max;
if n > bytesToSend then n:=bytesToSend;
if n = 0 then exit;
setLength(buf, n);
n:=stream.read(buf[1], n);
setLength(buf, n);
try result:=sock.SendStr(buf)
except end; // the socket may be accidentally closed
if result < n then stream.Seek(n-result, soCurrent);
end; // sendNextChunk

function ThttpConn.getBytesToSend():int64;
begin result:=bytesPartial-bsent_body end;

function ThttpConn.getBytesToPost():int64;
begin result:=post.length-bytesPosted end;

function ThttpConn.getbytesGot():int64;
begin result:=length(buffer) end;

function ThttpConn.getPostedLastItem():int64;
begin result:=bytesPosted-length(buffer)-lastPostItemStart end;

procedure ThttpConn.notify(ev:ThttpEvent);
begin srv.notify(ev, self) end;

procedure ThttpConn.sendheader(h:string='');
begin
state:=HCS_REPLYING_HEADER;
if reply.header = '' then reply.header:=h;
includeTrailingString(reply.header, CRLF);
reply.header:=reply.header+reply.additionalHeaders;
includeTrailingString(reply.header, CRLF);
try sock.SendStr(reply.header+CRLF);
except end;
end; // sendHeader

function replycode2reason(code:integer):string;
begin
case code of
  200: result:='OK';
  206: result:='Partial Content';
  301: result:='Moved Permanently';
  302: result:='Found';
  400: result:='Bad Request';
  401: result:='Unauthorized';
  403: result:='Forbidden';
  404: result:='Not Found';
  405: result:='Method Not Allowed';
  414: result:='Request-URI Too Large';
  500: result:='Internal Server Error';
  503: result:='Service Unavailable';
  else result:='';
  end;
end; // replycode2reason

function ThttpConn.replyHeader_code(code:integer):string;
begin
if reply.reason = '' then reply.reason:=replycode2reason(code);
result:=format('HTTP/1.1 %d %s'+CRLF, [code,reply.reason])
  + replyHeader_Str('Content-Type',reply.contentType)
end;

function ThttpConn.replyHeader_mode(mode:ThttpReplyMode):string;
begin result:=replyHeader_code(HRM2CODE[mode]) end;

procedure ThttpConn.addHeader(s:string);
begin reply.additionalHeaders:=reply.additionalHeaders+s+CRLF end;

constructor TspeedLimiter.create(max:integer=MAXINT);
begin
maxSpeed:=max;
availableBandwidth:=max;
end;

end.
