{
Copyright (C) 2002-2008 Massimo Melina (www.rejetto.com)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


This lib ensures only one instance of the software does run
}
unit monoLib;

interface

function init(id:string):boolean; // FALSE on error
function isMaster():boolean;
function fetchSlaveParams():string; // empty string if the queue is empty
procedure sendParams();
function getError():string;

var
  onSlaveParams: procedure(params:string);
  working: boolean = FALSE;

implementation

uses
  windows, messages, forms, classes;

type
  Tmono = class
    function hook(var msg:TMessage):boolean;
    end;

var
  msgID: Thandle;
  master: boolean;
  myWhnd, peerWhnd: Thandle;
  mono: Tmono;
  paramsQ: TStrings;
  error: string;

const
  MSG_WHEREAREYOU = 1;
  MSG_HEREIAM = 2;
  MSG_PARAMS = 3;

function getError():string;
begin result:=error end;

function isMaster():boolean;
begin result:=master end;

function fetchSlaveParams():string;
begin
result:='';
if paramsQ.Count = 0 then exit;
result:=paramsQ[0];
paramsQ.Delete(0);
end; // fetchSlaveParams

function atomToStr(atom:Tatom):string;
begin
setlength(result, 5000);
setlength(result, globalGetAtomName(atom, @result[1], length(result)));
end; // atomToStr

function Tmono.hook(var msg:TMessage):boolean;
var
  s: string;
begin
result:=FALSE;
if msg.msg <> msgID then exit;
result:=TRUE;
if master then
  case msg.wparam of
    MSG_WHEREAREYOU: sendMessage(msg.lparam, msgID, MSG_HEREIAM, myWhnd);
    MSG_PARAMS:
      begin
      s:=atomToStr(msg.lparam);
      if assigned(onSlaveParams) then onSlaveParams(s)
      else paramsQ.Add(s);
      end;
    end
else
  case msg.wparam of
    MSG_HEREIAM: peerWhnd:=msg.LParam;
    end;
end; // hook

function paramsToStr():string;
var
  i: integer;
begin
result:=paramStr(0);
for i:=1 to paramCount() do
  result:=result+#13+paramStr(i);
end; // paramsToStr

function init(id:string):boolean;
var
  i: integer;
begin
result:=FALSE;
// the mutex is auto-released when the application terminates
if createMutex(nil, True, pchar(id)) = 0 then
  begin
  setlength(error,1000);
  setlength(error, FormatMessage(
    FORMAT_MESSAGE_FROM_SYSTEM+FORMAT_MESSAGE_IGNORE_INSERTS, NIL,
    GetLastError(), 0, @error[1], length(error), NIL) );
  exit;
  end;
master:= GetLastError() <> ERROR_ALREADY_EXISTS;
msgID:=registerWindowMessage(pchar(id));
myWhnd:=application.handle;
paramsQ:=TStringList.create;
peerWhnd:=0;
mono:=Tmono.create;
application.HookMainWindow(mono.hook);
working:=TRUE;
result:=TRUE;
if master then exit;
// slave part (try three times)
for i:=1 to 3 do
  begin
  // it's like sendMessage (no queue) but it is non-blocking
  SendMessageCallback(HWND_BROADCAST, msgId, MSG_WHEREAREYOU, myWhnd, NIL, 0);
  application.ProcessMessages();
  if peerWhnd <> 0 then exit;
  sleep(10*i*i); // 10, 40, 90
  end;
working:=FALSE;
result:=FALSE;
error:='no reply';
end; // init

procedure sendParams();
var
  atom: TAtom;
begin
atom:=GlobalAddAtom(pchar(paramsToStr()));
sendMessage(peerWhnd, msgId, MSG_PARAMS, atom);
GlobalDeleteAtom(atom);
end; // sendParams

end.
