{
Copyright (C) 2002-2008  Massimo Melina (www.rejetto.com)

This file is part of HFS ~ HTTP File Server.

    HFS is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    HFS is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with HFS; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
}
{$A+,B-,C+,E-,F-,G+,H+,I-,J-,K-,L+,M-,N+,O+,P+,Q-,R-,S-,T-,U-,V+,X+,Y+,Z1}
{$MINSTACKSIZE $00004000}
{$MAXSTACKSIZE $00100000}
{$IMAGEBASE $00400000}
{$APPTYPE GUI}
{$INCLUDE defs.inc }

unit main;

interface

uses
  // delphi libs
  Windows, Messages, SysUtils, Classes, Graphics, Controls, ComCtrls, Dialogs,
  registry, ExtCtrls, shellapi, ImgList, ToolWin, StdCtrls, strutils, AppEvnts,
  types, Forms, clipbrd, Menus, shlobj, activex, Buttons, FileCtrl, dateutils, iniFiles,
  // 3rd part libs (ensure you have all of these, compare your version in file notes.txt)
  wsocket, httpProt, icsMD5, GIFimage, zlibex,
  HSlib, traylib, monoLib, progFrmLib, Mask; // rejetto libs

const
  VERSION = '2.2f';
  VERSION_BUILD = '155';
{$IFDEF STABLE }
  VERSION_STABLE = TRUE;
{$ELSE}
  VERSION_STABLE = FALSE;
{$ENDIF}
  CURRENT_VFS_FORMAT :integer = 1;
  CRLF = #13#10;
  CFG_KEY = 'Software\rejetto\HFS';
  CFG_FILE = 'hfs.ini';
  TPL_FILE = 'hfs.tpl';
  HFS_HTTP_AGENT = 'HFS/'+VERSION;
  COMMENTS_FILE = 'hfs.comments.txt';
  DIFF_TPL_FILE = 'hfs.diff.tpl';
  FILELIST_TPL_FILE = 'filelist.tpl';
  COMMENT_FILE_EXT = '.comment';
  SECONDS = 24*60*60; // Tdatetime * SECONDS = time in seconds
  ETA_FRAME = 5; // time frame for ETA (in seconds)
  DOWNLOAD_MIN_REFRESH_TIME :Tdatetime = 1/(5*SECONDS); // 5 Hz
  BYTES_GROUPING_THRESHOLD :Tdatetime = 1/SECONDS; // group bytes in log
  STATUSBAR_REFRESH = 10; // tenth of second
  MAX_RECENT_FILES = 5;
  MANY_ITEMS_THRESHOLD = 1000;
  COMPRESSION_THRESHOLD = 200; // if more than X items, VFS files are compressed
  KILO = 1024;
  MEGA = KILO*KILO;
  YESNO :array [boolean] of string=('no','yes');
  IP_SERVICES_URL = 'http://www.rejetto.com/hfs/ipservices.php';
  FACTORY_IP_SERVICES :array [0..3] of string = (
    'http://www.melauto.it/public/rejetto/ip.php|',
    'http://www.myip.dk/|Your IP: ',
    'http://www.dovedove.it/hfs/ip.php|',
    'http://www.whatismyip.org/|'
  );
  USER_ANONYMOUS = '@anonymous';
  ALWAYS_ON_WEB_SERVER = '64.233.187.99'; // guess who... google
  ADDRESS_COLOR = clGreen;
  DEFAULT_FILE_ICON = 37;

  ICONMENU_NEW = 1;

  RESTRICTMENU_ALL_ACCOUNTS = 1;
  RESTRICTMENU_NO_RESTRICTION = 2;
  RESTRICTMENU_ANONYMOUS = 3;
  RESTRICTMENU_NEW_ACCOUNT = 4;

  UPLOADMENU_ANYONE = 1;
  UPLOADMENU_NEW_ACCOUNT = 2;

  USER_ICON_MASKS_OFS = 10000;
  // messages
  MSG_UNPROTECTED_LINKS = 'Links are NOT actually protected.'
    +#13'The feature is there to be used in in conjunction with the "list protected items only..." option.'
    +#13'Continue?';
  MSG_SAME_NAME ='An item with the same name is already present in this folder.'
    +#13'Continue?';
  MSG_OPTIONS_SAVED = 'Options saved';
  MSG_SOME_LOCKED = 'Some items were not affected because locked';
  MSG_ITEM_LOCKED = 'The item is locked';
  MSG_ERROR_EXTERNAL = 'Can''t find external address';
  MSG_INVALID_VALUE = 'Invalid value';
  MSG_EMPTY_NO_LIMIT = 'Leave blank to get no limits.';
  MSG_ADDRESSES_EXCEED = 'The following addresses exceed the limit:'#13'%s';
  MSG_NO_TEMP = 'Cannot save temporary file';
  MSG_ERROR_REGISTRY = 'Can''t write to registry.'
    +#13'You may lack necessary rights.';
  MSG_MANY_ITEMS = 'You are putting many files.'
    +#13'This is just a warning, not an error.'
    +#13'You may notice slow operations.'
    +#13'Maybe real-folders could help you.'
    +#13'Read documentation or ask on the forum for help.';
  MSG_ADD_TO_HFS = '"Add to HFS" has been added to your Window''s Explorer right-click menu.';
  MSG_SINGLE_INSTANCE = 'Sorry, this feature only works with the "Only 1 instance" option enabled.'
    +#13#13'You can find this option under Menu -> Start/Exit'
    +#13'(only in expert mode)';
  MSG_ENABLED = 'Option enabled';
  MSG_DISABLED = 'Option disabled';

type
  Pboolean = ^boolean;

  TfileAttribute = (
    FA_FOLDER,       // folder kind
    FA_VIRTUAL,      // does not exist on disc
    FA_ROOT,         // only the root item has this attribute
    FA_BROWSABLE,    // permit listing of this folder (not recursive, only dir)
    FA_HIDDEN,       // hidden iterms won't be shown to browsers (not recursive)
    { no more used attributes have to stay for backward compatibility with
    { VFS files }
    FA_NO_MORE_USED1,
  	FA_NO_MORE_USED2,
    FA_TEMP,            // this is a temporary item and is not part of the VFS
    FA_HIDDENTREE,      // recursive hidden
    FA_LINK,            // redirection
    FA_UNIT,            // logical unit (drive)
    FA_VIS_ONLY_ANON,   // visible only to anonymous users [no more used]
    FA_DL_FORBIDDEN,    // forbid download (not recursive)
    FA_HIDE_EMPTY_FOLDERS,  // (recursive)
    FA_DONT_COUNT_AS_DL,    // (not recursive)
    FA_SOLVED_LNK,
    FA_HIDE_EXT,       // (recursive)
    FA_DONT_LOG,       // (recursive)
    FA_ARCHIVABLE      // (recursive)
  );
  TfileAttributes = set of TfileAttribute;

  Tfile = class;

  // returning FALSE stops recursion
  TfileCallback = function(f:Tfile; par:integer):boolean;

  Tfile = class (Tobject)
  private
    locked: boolean;
    FDLcount: integer;
    function  getVisible():boolean;
    function  getParent():Tfile;
    function  getDLcount():integer;
    procedure setDLcount(i:integer);
  public
    name, comment, user, pwd: string;
    resource: string;  // link to physical file/folder; URL for links
    flags: TfileAttributes;
    node: Ttreenode;
    size: int64; // -1 is NULL
    atime,            // when was this file added to the VFS ?
    mtime: Tdatetime; // modified time, read from disk
    icon: integer;
    accounts, uploadAccounts: TStringDynArray;
    filesFilter, foldersFilter, realm, diffTpl,
    defaultFileMask, dontCountAsDownloadMask: string;
    constructor create(fullpath:string);
    constructor createVirtualFolder(name:string);
    constructor createLink(name:string);
    property  visible:boolean read getVisible;
    property  parent:Tfile read getParent;
    property  DLcount:integer read getDLcount write setDLcount;
    function  getDLcountRecursive():integer;
    function  toggle(att:TfileAttribute):boolean;
    function  isFolder():boolean;
    function  isFile():boolean;
    function  isFileOrFolder():boolean;
    function  isRealFolder():boolean;
    function  isVirtualFolder():boolean;
    function  isEmptyFolder(conn:ThttpConn=NIL):boolean;
    function  isRoot():boolean;
    function  isLink():boolean;
    function  isTemp():boolean;
    function  isNew():boolean;
    function  isPublic():boolean;
    function  url(fullEncode:boolean=FALSE):string;
    function  relativeURL(fullEncode:boolean=FALSE):string;
    function  pathTill(root:Tfile=NIL; delim:char='\'):string;
    function  parentURL():string;
		function  fullURL(userpwd:string=''; ip:string=''):string;
    procedure setupImage(newIcon:integer); overload;
    procedure setupImage(); overload;
		function  accessList(specialUsernames:boolean=FALSE;
      outInherited:Pboolean=NIL):TstringDynArray;
    function  accessFor(username, password:string):boolean; overload;
    function  accessFor(conn:ThttpConn):boolean; overload;
    function  hasRecursive(attribute: TfileAttribute;
      outInherited:Pboolean=NIL):boolean;
    function  getImageIndex():integer;
    function  getShownRealm():string;
    function  getFolder():string;
    function  getRecursiveFileMask():string;
    function  countAsDownload():boolean;
    function  getDefaultFile():Tfile;
    procedure recursiveApply(callback:TfileCallback; par:integer=0);
    procedure getFiltersRecursively(var files,folders:string);
    function  diskfree():int64;
    function  same(f:Tfile):boolean;
    procedure setName(name:string);
    procedure setResource(res:string);
    function  getDynamicComment(skipParent:boolean=FALSE):string; 
    function  getRecursiveDiffTplAsStr(outInherited:Pboolean=NIL;
      outFromDisk:Pboolean=NIL):string;
    // locking prevents modification of all its ancestors and descendants
    procedure lock();
    procedure unlock();
    function  isLocked():boolean;
    end; // Tfile

  TfilterMethod = function(self:Tobject):boolean;

  Thelp = ( HLP_NONE, HLP_TPL );

  TdownloadingWhat = ( DW_UNK, DW_FILE, DW_FOLDERPAGE, DW_ICON, DW_ERROR,
      DW_ARCHIVE );

  TpreReply =  (PR_NONE, PR_BAN, PR_OVERLOAD);

  TconnData = class  // data associated to a client connection
    averageSpeed: real;   { calculated on disconnection as bytesSent/totalTime.
                            it is calculated also while sending and it is different from
                            conn.speed because conn.speed is average speed in the last
                            second, while averageSpeed is calculated on ETA_FRAME seconds }
    time: Tdatetime;  // connection start time
    tray: Ttrayicon;
    tray_ico: Ticon;
    lastFN: string;
    lastFile: Tfile;
    countAsDownload: boolean; // cache the value for the Tfile method
    { cache User-Agent because often retrieved by connBox.
    { this value is filled after the http request is complete (HE_REQUESTED),
    { or before, during the request as we get a file (HE_POST_FILE). }
    agent: string;
    conn: ThttpConn;
    deleting: boolean;      // don't use, this item is about to be discarded
    nextDloadScreenUpdate: Tdatetime; // avoid too fast updating during download
    disconnectReason: string;
    error: string;         // error details
    eta: record
      idx: integer;   // estimation time (seconds)
      data: array [0..ETA_FRAME-1] of real;  // accumulates speed data
      result: Tdatetime;
      end;
    downloadingWhat: TdownloadingWhat;
    preReply: TpreReply;
    banReason: string;
    lastBytesSent, lastBytesGot: int64;
    lastActivityTime, fileXferStart: Tdatetime;
    uploadingFN: string;
    uploadFailed: string; // reason (empty on success)
    uploadResults: string; // filename:reason<CRLF>filename:reason<CRLF>
    disconnectAfterReply, logLaterInApache, dontLog, fullDLlogged: boolean;
    bytesGotGrouping, bytesSentGrouping: record
      bytes: integer;
      from: Tdatetime;
      end;
    params: THashedStringList;
    workaroundForIEutf8: (toDetect, yes, no);
    { here we put just a pointer because the file type would triplicate the
    { size of this record, while it is used only for very few connections }
    f: ^file;

    constructor create(conn:ThttpConn);
    destructor Destroy; override;
    end; // Tconndata

  Paccount = ^Taccount;
	Taccount = record   // user/pass profile
    user, pwd: string;
    enabled, noLimits: boolean;
    end;
  Taccounts = array of Taccount;

  Tautosave = record
    every, minimum: integer; // in seconds
    last: Tdatetime;
    menu: Tmenuitem;
    end;

  TstringIntPairs = array of record
    str:string;
    int:integer;
    end;
    
  TmainFrm = class(TForm)
    splitH: TSplitter;
    filemenu: TPopupMenu;
    newfolder1: TMenuItem;
    images: TImageList;
    Remove1: TMenuItem;
    ToolBar1: TToolBar;
    startBtn: TToolButton;
    ToolButton1: TToolButton;
    menuBtn: TToolButton;
    menu: TPopupMenu;
    About1: TMenuItem;
    connmenu: TPopupMenu;
    Kickconnection1: TMenuItem;
    KickIPaddress1: TMenuItem;
    Kickallconnections1: TMenuItem;
    Viewhttprequest1: TMenuItem;
    Saveoptions1: TMenuItem;
    toregistrycurrentuser1: TMenuItem;
    tofile1: TMenuItem;
    toregistryallusers1: TMenuItem;
    timer: TTimer;
    ToolBar2: TToolBar;
    IPaddress1: TMenuItem;
    AutocopyURLonadditionChk: TMenuItem;
    Letbrowse1: TMenuItem;
    Hide1: TMenuItem;
    foldersbeforeChk: TMenuItem;
    Browseit1: TMenuItem;
    Openit1: TMenuItem;
    appEvents: TApplicationEvents;
    logmenu: TPopupMenu;
    DumprequestsChk: TMenuItem;
    CopyURL1: TMenuItem;
    Readonly1: TMenuItem;
    Clear1: TMenuItem;
    Copy1: TMenuItem;
    N3: TMenuItem;
    LogtimeChk: TMenuItem;
    LogdateChk: TMenuItem;
    Saveas1: TMenuItem;
    Save1: TMenuItem;
    N4: TMenuItem;
    connPnl: TPanel;
    connheader: THeaderControl;
    connBox: TListBox;
    MinimizetotrayChk: TMenuItem;
    Restore1: TMenuItem;
    Numberofcurrentconnections1: TMenuItem;
    Numberofloggeddownloads1: TMenuItem;
    Numberofloggedhits1: TMenuItem;
    Exit1: TMenuItem;
    Shellcontextmenu1: TMenuItem;
    Flashtaskbutton1: TMenuItem;
    onDownloadChk: TMenuItem;
    onconnectionChk: TMenuItem;
    never1: TMenuItem;
    N6: TMenuItem;
    startminimizedChk: TMenuItem;
    EditHTMLtemplate1: TMenuItem;
    N7: TMenuItem;
    trayicons1: TMenuItem;
    trayfordownloadChk: TMenuItem;
    N8: TMenuItem;
    leavedisconnectedconnectionsChk: TMenuItem;
    Loadfilesystem1: TMenuItem;
    Savefilesystem1: TMenuItem;
    N1: TMenuItem;
    N12: TMenuItem;
    usesystemiconsChk: TMenuItem;
    N13: TMenuItem;
    Officialwebsite1: TMenuItem;
    numbers: TImageList;
    showmaintrayiconChk: TMenuItem;
    Speedlimit1: TMenuItem;
    N10: TMenuItem;
    saveoptionsonexitChk: TMenuItem;
    Limits1: TMenuItem;
    Maxconnections1: TMenuItem;
    Maxconnectionsfromsingleaddress1: TMenuItem;
    Weblinks1: TMenuItem;
    Forum1: TMenuItem;
    FAQ1: TMenuItem;
    License1: TMenuItem;
    Paste1: TMenuItem;
    Addfiles1: TMenuItem;
    Addfolder1: TMenuItem;
    Splitter1: TSplitter;
    graphBox: TPaintBox;
    Graphrefreshrate1: TMenuItem;
    Pausestreaming1: TMenuItem;
    Comment1: TMenuItem;
    Setuserpass1: TMenuItem;
    BanIPaddress1: TMenuItem;             
    N2: TMenuItem;
    BannedIPaddresses1: TMenuItem;
    Loadrecentfiles1: TMenuItem;
    alwaysontopChk: TMenuItem;
    Checkforupdates1: TMenuItem;
    Rename1: TMenuItem;
    Otheroptions1: TMenuItem;
    Nodownloadtimeout1: TMenuItem;
    Autoclose1: TMenuItem;
    Showbandwidthgraph1: TMenuItem;
    Pause1: TMenuItem;
    reloadonstartupChk: TMenuItem;
    MIMEtypes1: TMenuItem;
    autocopyURLonstartChk: TMenuItem;
    Accounts1: TMenuItem;
    Restrictaccess1: TMenuItem;
    encodenonasciiChk: TMenuItem;
    encodeSpacesChk: TMenuItem;
    URLencoding1: TMenuItem;
    traymessage1: TMenuItem;
    DMbrowserTplChk: TMenuItem;
    Guide1: TMenuItem;
    autosaveVFSchk: TMenuItem;
    sendHFSidentifierChk: TMenuItem;
    persistentconnectionsChk: TMenuItem;
    Logfile1: TMenuItem;
    Hidetree1: TMenuItem;
    VirtualFileSystem1: TMenuItem;
    listfileswithhiddenattributeChk: TMenuItem;
    listfileswithsystemattributeChk: TMenuItem;
    hideProtectedItemsChk: TMenuItem;
    StartExit1: TMenuItem;
    Font1: TMenuItem;
    Newlink1: TMenuItem;
    SetURL1: TMenuItem;
    Icon1: TMenuItem;
    usecommentasrealmChk: TMenuItem;
    Resetuserpass1: TMenuItem;
    Switchtovirtual1: TMenuItem;
    LogiconsChk: TMenuItem;
    Customizedrealm2: TMenuItem;
    N17: TMenuItem;
    Loginrealm1: TMenuItem;
    Logwhat1: TMenuItem;
    N9: TMenuItem;
    N16: TMenuItem;
    logconnectionsChk: TMenuItem;
    logDisconnectionsChk: TMenuItem;
    logRequestsChk: TMenuItem;
    logRepliesChk: TMenuItem;
    logFulldownloadsChk: TMenuItem;
    logBytesreceivedChk: TMenuItem;
    logBytessentChk: TMenuItem;
    logServerstartChk: TMenuItem;
    logServerstopChk: TMenuItem;
    logBrowsingChk: TMenuItem;
    Help1: TMenuItem;
    Introduction1: TMenuItem;
    N18: TMenuItem;
    Resetfileshits1: TMenuItem;
    Kickidleconnections1: TMenuItem;
    Connectionsinactivitytimeout1: TMenuItem;
    sbar: TStatusBar;
    logOnVideoChk: TMenuItem;
    N19: TMenuItem;
    Clearfilesystem1: TMenuItem;
    HintsfornewcomersChk: TMenuItem;
    Upload1: TMenuItem;
    Uploadforaccounts1: TMenuItem;
    logUploadsChk: TMenuItem;
    only1instanceChk: TMenuItem;
    compressedbrowsingChk: TMenuItem;
    Numberofloggeduploads1: TMenuItem;
    logProgressChk: TMenuItem;
    Flagfilesaddedrecently1: TMenuItem;
    Flagasnew1: TMenuItem;
    confirmexitChk: TMenuItem;
    Donotlogaddress1: TMenuItem;
    N15: TMenuItem;
    Custom1: TMenuItem;
    noPortInUrlChk: TMenuItem;
    saveTotalsChk: TMenuItem;
    Findexternaladdress1: TMenuItem;
    findExtOnStartupChk: TMenuItem;
    DynamicDNSupdater1: TMenuItem;
    Custom2: TMenuItem;
    N21: TMenuItem;
    CJBtemplate1: TMenuItem;
    NoIPtemplate1: TMenuItem;
    DynDNStemplate1: TMenuItem;
    searchbetteripChk: TMenuItem;
    deletePartialUploadsChk: TMenuItem;
    Whyisuploaddisabled1: TMenuItem;
    Minimumdiskspace1: TMenuItem;
    Banthisaddress1: TMenuItem;
    modalOptionsChk: TMenuItem;
    Address2name1: TMenuItem;
    Resetnewflag1: TMenuItem;
    beepChk: TMenuItem;
    Renamepartialuploads1: TMenuItem;
    SelfTest1: TMenuItem;
    Opendirectlyinbrowser1: TMenuItem;
    MaxDLsIP1: TMenuItem;
    maxDLs1: TMenuItem;
    Editresource1: TMenuItem;
    logBannedChk: TMenuItem;
    ToolButton2: TToolButton;
    modeBtn: TToolButton;
    Addfiles2: TMenuItem;
    Addfolder2: TMenuItem;
    Clearoptionsandquit1: TMenuItem;
    numberFilesOnUploadChk: TMenuItem;
    Upload2: TMenuItem;
    UninstallHFS1: TMenuItem;
    maxIPs1: TMenuItem;
    maxIPsDLing1: TMenuItem;
    keepBakUpdatingChk: TMenuItem;
    Nodownload1: TMenuItem;
    Autohideemptyfolders1: TMenuItem;
    Autosaveevery1: TMenuItem;
    Autosaveevery2: TMenuItem;
    Apachelogfileformat1: TMenuItem;
    SwitchON1: TMenuItem;
    loadSingleCommentsChk: TMenuItem;
    Bindroottorealfolder1: TMenuItem;
    Unbindroot1: TMenuItem;
    Switchtorealfolder1: TMenuItem;
    abortBtn: TToolButton;
    Seelastserverresponse1: TMenuItem;
    N5: TMenuItem;
    logOtherEventsChk: TMenuItem;
    supportDescriptionChk: TMenuItem;
    Showcustomizedoptions1: TMenuItem;
    useISOdateChk: TMenuItem;
    Difftemplate1: TMenuItem;
    browseUsingLocalhostChk: TMenuItem;
    Addingfolder1: TMenuItem;
    askFolderKindChk: TMenuItem;
    defaultToVirtualChk: TMenuItem;
    defaultToRealChk: TMenuItem;
    enableNoDefaultChk: TMenuItem;
    RunHFSwhenWindowsstarts1: TMenuItem;
    trayInsteadOfQuitChk: TMenuItem;
    Hidefileextentioninlisting1: TMenuItem;
    Addicons1: TMenuItem;
    Iconmasks1: TMenuItem;
    CopyURLwithpassword1: TMenuItem;
    CopyURLwithdifferentaddress1: TMenuItem;
    hisIPaddressisusedforURLbuilding1: TMenuItem;
    N20: TMenuItem;
    Acceptconnectionson1: TMenuItem;
    Anyaddress1: TMenuItem;
    autoCommentChk: TMenuItem;
    fingerprintsChk: TMenuItem;
    CopyURLwithfingerprint1: TMenuItem;
    recursiveListingChk: TMenuItem;
    Disable1: TMenuItem;
    logOnlyServedChk: TMenuItem;
    Fingerprints1: TMenuItem;
    saveNewFingerprintsChk: TMenuItem;
    Createfingerprintonaddition1: TMenuItem;
    encodePwdUrlChk: TMenuItem;
    pwdInPagesChk: TMenuItem;
    deleteDontAskChk: TMenuItem;
    Updates1: TMenuItem;
    updateDailyChk: TMenuItem;
    N22: TMenuItem;
    Howto1: TMenuItem;
    testerUpdatesChk: TMenuItem;
    urlBox: TEdit;
    Defaultsorting1: TMenuItem;
    Name1: TMenuItem;
    Size1: TMenuItem;
    Time1: TMenuItem;
    Hits1: TMenuItem;
    N23: TMenuItem;
    centralPnl: TPanel;
    filesBox: TTreeView;
    logBox: TRichEdit;
    splitV: TSplitter;
    ToolBar3: TToolBar;
    browseBtn: TToolButton;
    Resettotals1: TMenuItem;
    Clearandresettotals1: TMenuItem;
    Dontlogsomefiles1: TMenuItem;
    preventLeechingChk: TMenuItem;
    NumberofdifferentIPaddresses1: TMenuItem;
    NumberofdifferentIPaddresseseverconnected1: TMenuItem;
    Addresseseverconnected1: TMenuItem;
    N24: TMenuItem;
    Allowedreferer1: TMenuItem;
    ToolButton4: TToolButton;
    Allowarchive1: TMenuItem;
    oemForIonChk: TMenuItem;
    N25: TMenuItem;
    Advanced1: TMenuItem;
    Filesfilter1: TMenuItem;
    Foldersfilter1: TMenuItem;
    Defaultfilemask1: TMenuItem;
    Dontconsiderasadownloadmask1: TMenuItem;
    Dontconsiderasadownload1: TMenuItem;
    Dontlog1: TMenuItem;
    N11: TMenuItem;
    portBtn: TToolButton;
    Donate1: TMenuItem;
    procedure FormResize(Sender: TObject);
    procedure filesBoxCollapsing(Sender: TObject; Node: TTreeNode;
      var AllowCollapse: Boolean);
    procedure newfolder1Click(Sender: TObject);
    procedure filesBoxEditing(Sender: TObject; Node: TTreeNode;
      var AllowEdit: Boolean);
    procedure filesBoxEdited(Sender: TObject; Node: TTreeNode; var S: String);
    procedure Remove1Click(Sender: TObject);
    procedure startBtnClick(Sender: TObject);
    procedure filesBoxChange(Sender: TObject; Node: TTreeNode);
    procedure Kickconnection1Click(Sender: TObject);
    procedure Kickallconnections1Click(Sender: TObject);
    procedure KickIPaddress1Click(Sender: TObject);
    procedure Viewhttprequest1Click(Sender: TObject);
    procedure connmenuPopup(Sender: TObject);
    procedure filemenuPopup(Sender: TObject);
    procedure toregistrycurrentuser1Click(Sender: TObject);
    procedure toregistryallusers1Click(Sender: TObject);
    procedure About1Click(Sender: TObject);
    procedure timerEvent(Sender: TObject);
    procedure menuPopup(Sender: TObject);
    procedure filesBoxDblClick(Sender: TObject);
    procedure Letbrowse1Click(Sender: TObject);
    procedure Hide1Click(Sender: TObject);
    procedure filesBoxMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure filesBoxCompare(Sender: TObject; Node1, Node2: TTreeNode;
      Data: Integer; var Compare: Integer);
    procedure foldersbeforeChkClick(Sender: TObject);
    procedure Browseit1Click(Sender: TObject);
    procedure Openit1Click(Sender: TObject);
    procedure ToolBar3CustomDraw(Sender: TToolBar; const ARect: TRect;
      var DefaultDraw: Boolean);
    procedure splitVMoved(Sender: TObject);
    procedure appEventsShowHint(var HintStr: String;
      var CanShow: Boolean; var HintInfo: THintInfo);
    procedure logmenuPopup(Sender: TObject);
    procedure Readonly1Click(Sender: TObject);
    procedure Clear1Click(Sender: TObject);
    procedure Copy1Click(Sender: TObject);
    procedure Saveas1Click(Sender: TObject);
    procedure Save1Click(Sender: TObject);
    procedure Clearoptionsandquit1click(Sender: TObject);
    procedure connBoxDrawItem(Control: TWinControl; Index: Integer;
      R: TRect; State: TOwnerDrawState);
    procedure connheaderResize(Sender: TObject);
    procedure connheaderSectionResize(HeaderControl: THeaderControl;
      Section: THeaderSection);
    procedure connBoxMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure appEventsMinimize(Sender: TObject);
    procedure appEventsRestore(Sender: TObject);
    procedure Restore1Click(Sender: TObject);
    procedure Numberofcurrentconnections1Click(Sender: TObject);
    procedure Numberofloggeddownloads1Click(Sender: TObject);
    procedure Numberofloggedhits1Click(Sender: TObject);
    procedure Exit1Click(Sender: TObject);
    procedure onDownloadChkClick(Sender: TObject);
    procedure onconnectionChkClick(Sender: TObject);
    procedure never1Click(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure filesBoxDragOver(Sender, Source: TObject; X, Y: Integer;
      State: TDragState; var Accept: Boolean);
    procedure filesBoxDragDrop(Sender, Source: TObject; X, Y: Integer);
    procedure EditHTMLtemplate1Click(Sender: TObject);
    procedure Savefilesystem1Click(Sender: TObject);
    procedure filesBoxDeletion(Sender: TObject; Node: TTreeNode);
    procedure Loadfilesystem1Click(Sender: TObject);
    procedure connheaderSectionEndDrag(Sender: TObject);
    procedure leavedisconnectedconnectionsChkClick(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure Officialwebsite1Click(Sender: TObject);
    procedure showmaintrayiconChkClick(Sender: TObject);
    procedure Speedlimit1Click(Sender: TObject);
    procedure tofile1Click(Sender: TObject);
    procedure Maxconnections1Click(Sender: TObject);
    procedure Maxconnectionsfromsingleaddress1Click(Sender: TObject);
    procedure Forum1Click(Sender: TObject);
    procedure FAQ1Click(Sender: TObject);
    procedure License1Click(Sender: TObject);
    procedure Paste1Click(Sender: TObject);
    procedure Addfiles1Click(Sender: TObject);
    procedure Addfolder1Click(Sender: TObject);
    procedure Splitter1Moved(Sender: TObject);
    procedure FormPaint(Sender: TObject);
    procedure Graphrefreshrate1Click(Sender: TObject);
    procedure Pausestreaming1Click(Sender: TObject);
    procedure Comment1Click(Sender: TObject);
    procedure filesBoxCustomDrawItem(Sender: TCustomTreeView;
      Node: TTreeNode; State: TCustomDrawState; var DefaultDraw: Boolean);
    procedure Setuserpass1Click(Sender: TObject);
    procedure browseBtnClick(Sender: TObject);
    procedure BanIPaddress1Click(Sender: TObject);
    procedure BannedIPaddresses1Click(Sender: TObject);
    procedure FormKeyUp(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure Checkforupdates1Click(Sender: TObject);
    procedure Rename1Click(Sender: TObject);
    procedure Nodownloadtimeout1Click(Sender: TObject);
    procedure alwaysontopChkClick(Sender: TObject);
    procedure Showbandwidthgraph1Click(Sender: TObject);
    procedure Pause1Click(Sender: TObject);
    procedure MIMEtypes1Click(Sender: TObject);
    procedure Accounts1Click(Sender: TObject);
    procedure Filesfilter1Click(Sender: TObject);
    procedure Foldersfilter1Click(Sender: TObject);
    procedure traymessage1Click(Sender: TObject);
    procedure Guide1Click(Sender: TObject);
    procedure filesBoxAddition(Sender: TObject; Node: TTreeNode);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure Logfile1Click(Sender: TObject);
    procedure Hidetree1Click(Sender: TObject);
    procedure Font1Click(Sender: TObject);
    procedure Newlink1Click(Sender: TObject);
    procedure SetURL1Click(Sender: TObject);
    procedure Resetuserpass1Click(Sender: TObject);
    procedure Switchtovirtual1Click(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure Customizedrealm2Click(Sender: TObject);
    procedure Loginrealm1Click(Sender: TObject);
    procedure Introduction1Click(Sender: TObject);
    procedure Resetfileshits1Click(Sender: TObject);
    procedure persistentconnectionsChkClick(Sender: TObject);
    procedure Kickidleconnections1Click(Sender: TObject);
    procedure Connectionsinactivitytimeout1Click(Sender: TObject);
    procedure splitHMoved(Sender: TObject);
    procedure Clearfilesystem1Click(Sender: TObject);
    procedure Defaultfilemask1Click(Sender: TObject);
    procedure Numberofloggeduploads1Click(Sender: TObject);
    procedure Flagfilesaddedrecently1Click(Sender: TObject);
    procedure Flagasnew1Click(Sender: TObject);
    procedure Donotlogaddress1Click(Sender: TObject);
    procedure Custom1Click(Sender: TObject);
    procedure Findexternaladdress1Click(Sender: TObject);
    procedure sbarDblClick(Sender: TObject);
    procedure NoIPtemplate1Click(Sender: TObject);
    procedure Custom2Click(Sender: TObject);
    procedure CJBtemplate1Click(Sender: TObject);
    procedure DynDNStemplate1Click(Sender: TObject);
    procedure Whyisuploaddisabled1Click(Sender: TObject);
    procedure Minimumdiskspace1Click(Sender: TObject);
    procedure Banthisaddress1Click(Sender: TObject);
    procedure Address2name1Click(Sender: TObject);
    procedure Resetnewflag1Click(Sender: TObject);
    procedure Renamepartialuploads1Click(Sender: TObject);
    procedure SelfTest1Click(Sender: TObject);
    procedure Opendirectlyinbrowser1Click(Sender: TObject);
    procedure noPortInUrlChkClick(Sender: TObject);
    procedure maxDLs1Click(Sender: TObject);
    procedure MaxDLsIP1Click(Sender: TObject);
    procedure Editresource1Click(Sender: TObject);
    procedure modeBtnClick(Sender: TObject);
    procedure Shellcontextmenu1Click(Sender: TObject);
    procedure Dontconsiderasdownload1Click(Sender: TObject);
    procedure UninstallHFS1Click(Sender: TObject);
    procedure maxIPs1Click(Sender: TObject);
    procedure maxIPsDLing1Click(Sender: TObject);
    procedure Nodownload1Click(Sender: TObject);
    procedure Autohideemptyfolders1Click(Sender: TObject);
    procedure Autosaveevery1Click(Sender: TObject);
    procedure Autosaveevery2Click(Sender: TObject);
    procedure CopyURL1Click(Sender: TObject);
    procedure Apachelogfileformat1Click(Sender: TObject);
    procedure Bindroottorealfolder1Click(Sender: TObject);
    procedure Unbindroot1Click(Sender: TObject);
    procedure Switchtorealfolder1Click(Sender: TObject);
    procedure abortBtnClick(Sender: TObject);
    procedure Seelastserverresponse1Click(Sender: TObject);
    procedure Showcustomizedoptions1Click(Sender: TObject);
    procedure useISOdateChkClick(Sender: TObject);
    procedure Difftemplate1Click(Sender: TObject);
    procedure RunHFSwhenWindowsstarts1Click(Sender: TObject);
    procedure askFolderKindChkClick(Sender: TObject);
    procedure defaultToVirtualChkClick(Sender: TObject);
    procedure defaultToRealChkClick(Sender: TObject);
    procedure Hidefileextentioninlisting1Click(Sender: TObject);
    procedure Addicons1Click(Sender: TObject);
    procedure Iconmasks1Click(Sender: TObject);
    procedure Anyaddress1Click(Sender: TObject);
    procedure filesBoxEndDrag(Sender, Target: TObject; X, Y: Integer);
    procedure CopyURLwithfingerprint1Click(Sender: TObject);
    procedure Disable1Click(Sender: TObject);
    procedure saveNewFingerprintsChkClick(Sender: TObject);
    procedure Createfingerprintonaddition1Click(Sender: TObject);
    procedure pwdInPagesChkClick(Sender: TObject);
    procedure Howto1Click(Sender: TObject);
    procedure Name1Click(Sender: TObject);
    procedure Size1Click(Sender: TObject);
    procedure Time1Click(Sender: TObject);
    procedure Hits1Click(Sender: TObject);
    procedure Resettotals1Click(Sender: TObject);
    procedure menuBtnClick(Sender: TObject);
    procedure Clearandresettotals1Click(Sender: TObject);
    procedure Dontlog1Click(Sender: TObject);
    procedure Dontlogsomefiles1Click(Sender: TObject);
    procedure trayicons1Click(Sender: TObject);
    procedure NumberofdifferentIPaddresses1Click(Sender: TObject);
    procedure NumberofdifferentIPaddresseseverconnected1Click(Sender: TObject);
    procedure Addresseseverconnected1Click(Sender: TObject);
    procedure Allowedreferer1Click(Sender: TObject);
    procedure filesBoxEnter(Sender: TObject);
    procedure filesBoxMouseEnter(Sender: TObject);
    procedure filesBoxMouseLeave(Sender: TObject);
    procedure filesBoxExit(Sender: TObject);
    procedure sbarMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure Allowarchive1Click(Sender: TObject);
    procedure Dontconsiderasadownloadmask1Click(Sender: TObject);
    procedure portBtnClick(Sender: TObject);
    procedure SwitchON1Click(Sender: TObject);
    procedure Donate1Click(Sender: TObject);
  private
    function  getGraphPic(conn:ThttpConn=NIL): string;
    procedure WMDropFiles(var msg:TWMDropFiles);
      message WM_DROPFILES;
    procedure WMQueryEndSession(var msg:TWMQueryEndSession);
      message WM_QUERYENDSESSION;
    procedure WMEndSession(var msg:TWMEndSession);
      message WM_ENDSESSION;
    procedure WMNCLButtonDown(var msg:TWMNCLButtonDown);
      message WM_NCLBUTTONDOWN;
    procedure trayEvent(sender:Tobject; ev:TtrayEvent);
    procedure downloadtrayEvent(sender:Tobject; ev:TtrayEvent);
    procedure httpEvent(event:ThttpEvent; conn:ThttpConn);
    function  addFileRecur(f:Tfile; parent:Ttreenode=NIL):Tfile;
    procedure remove(node:Ttreenode=NIL);
    function  findFilebyURL(url:string):Tfile;
    function  pointedFile(strict:boolean=TRUE):Tfile;
    function  pointedConnection():TconnData;
    procedure refreshSbar();
    function  getFolderPage(folder:Tfile; conn:ThttpConn; otpl:Tobject):string;
    procedure getPage(section:string; data:TconnData; f:Tfile=NIL);
    function  selectedConnection():ThttpConn;
    function  setCfg(cfg:string; alreadyStarted:boolean=FALSE):boolean;
    function  getCfg():string;
    function  loadCfg():boolean;
    function  sendPic(conn:ThttpConn; idx:integer=-1):boolean;
    procedure ipmenuclick(sender:Tobject);
    procedure acceptOnMenuclick(sender:Tobject);
    procedure restrictaccessClick(sender:Tobject);
    procedure uploadMenuClick(sender:Tobject);
    procedure iconMenuClick(sender:Tobject);
    procedure copyURLwithAddressMenuClick(sender:Tobject);
    procedure copyURLwithPasswordMenuClick(sender:Tobject);
    procedure setlastCol();
    procedure updateTrayTip();
    procedure setTrayShows(s:string);
    procedure addTray();
    procedure refreshConn(conn:ThttpConn);
    function  getVFS(node:Ttreenode=NIL):string;
    procedure setVFS(vfs:string; node:Ttreenode=NIL);
    procedure setSpeedLimit(v:real);
    function  saveCFG():boolean;
    procedure setnoDownloadTimeout(v:integer);
		procedure addDropFiles(hnd:Thandle; under:Ttreenode);
    procedure pasteFiles();
		function  addFilesFromString(files:string; under:Ttreenode=NIL):Tfile;
		procedure refreshGraphBox();
		procedure setGraphRate(v:integer);
    procedure updateRecentFilesMenu();
    procedure recentsClick(sender:Tobject);
    procedure popupMainMenu();
    procedure updateAlwaysOnTop();
    procedure initVFS();
		procedure refreshIPlist();
		procedure updateUrlBox();
    function  addFile(f:Tfile; parent:Ttreenode=NIL;
      skipComment:boolean=FALSE):Tfile;
    procedure loadVFS(fn:string);
    procedure compressReply(cd:TconnData);
    procedure purgeConnections();
    procedure setEasyMode(easy:boolean=TRUE);
    procedure hideGraph();
    procedure showGraph();
    function  fileAttributeInSelection(fa:TfileAttribute):boolean;
    procedure progFrmHttpGetUpdate(sender:TObject; buffer:pointer; Len:integer);
    procedure add2log(lines:string; conn:ThttpConn=NIL; clr:Tcolor=clBlack);
    procedure recalculateGraph();
 public
    function  ipPointedInLog():string;
    procedure saveVFS(fn:string='');
    function finalInit():boolean;
    procedure beforeQuit();
    procedure processParams_before(var params:TStringDynArray;
      allowed:string='');
    procedure processParams_after(var params:TStringDynArray);
    procedure setStatusBarText(s:string; lastFor:integer=2);
    procedure minimizeToTray();
    procedure autoCheckUpdates();
  end; // Tmainfrm

var
  mainFrm: TmainFrm;
  srv: ThttpSrv;
  rootFile: Tfile;
  rootNode: TtreeNode;
  clock: integer;       // program ticks (tenths of second)
  defaultIP: string;    // the IP address to use forming URLs
  customIPs: TStringDynArray;   // user customized IP addresses
  // workaround for splitters' bad behaviour
  lastGoodLogWidth, lastGoodConnHeight: integer;
  tray: Ttrayicon;
  tray_ico: Ticon;            // the actual icon shown in tray
  usingFreePort: boolean=TRUE;  // the actual server port set was 0
  hitsLogged, downloadsLogged, uploadsLogged: integer;
  upTime: Tdatetime;           // the server is up since...
  programStart: Tdatetime;     // execution time
  trayed: boolean;             // true if the window has been minimized to tray
  trayShows: string;           // describes the content of the tray icon
  flashOn: string;             // describes when to flash the taskbar
  addFolderDefault: string;    // how to default adding a folder (real/virtual)
  defSorting: string;          // default sorting, browsing
  toDelete: Tlist;             // connections pending for deletion
  systemimages: Timagelist;    // system icons
  speedLimit: real;            // overall limit, Kb/s
  maxConnections: integer;     // max number of connections (total)
  maxConnectionsIP: integer;   // ...from a single address
  maxContempDLs: integer; // max number of contemporaneous downloads
  maxContempDLsIP: integer; // ...from a single address
  maxIPs: integer;             // max number of different addresses connected
  maxIPsDLing: integer;        // max number of different addresses downloading
  autoFingerprint: integer;    // create fingerprint on file addition
  lastFileOpen: string;
  renamePartialUploads: string;
  allowedReferer: string;      // check over the Refer header field
  altPressedForMenu: boolean;  // used to enable the menu on ALT key
  noDownloadTimeout: integer;  // autoclose the application after (minutes)
  connectionsInactivityTimeout: integer; // autokick connection after (seconds)
  startingImagesCount: integer;
  lastUpdateCheck, lastFilelistTpl: Tdatetime;
  lastUpdateCheckFN: string;   // eventual temp file for saving lastUpdateCheck
  lastActivityTime: Tdatetime;  // used for the "no download timeout"
  recentFiles: TStringDynArray; // recently loaded files
  addingItemsCounter: integer = -1; // -1 is disabled
  stopAddingItems: boolean;
  port: string;
  defaultTpl: string;
  tpl_help: string;
  lastAddedFolder: string;
  lastWindowRect: Trect;
  iconMasks: TstringIntPairs;
  // maps system imagelist icons to internal imagelist
  sysidx2index: array of record sysidx, idx:integer; end;
  saveMode: ( SM_USER, SM_SYSTEM, SM_FILE );
  banlist: array of record ip,comment:string; end;
  loginRealm: string;
  MIMEtypes, uploadPaths, address2name, IPservices: TstringDynArray;
  accounts: Taccounts;
  trayMsg: string;
  VFSmodified: boolean; // TRUE if the VFS changes have not been saved
  VFScounterMod: boolean; // if any counter has changed
  imagescache: array of string;
  trayNL: string = #13;
  logFontName: string;
  logFontSize: integer;
  applicationFullyInitialized: boolean;
  lockTimerevent: boolean;
  filesStayFlaggedForMinutes: integer;
  selectedFile: Tfile;  // last selected file on the tree
  windowsShuttingDown: boolean = FALSE;
  dontLogAddressMask: string;
  openInBrowser: string; // to not send the "attachment" suggestion in header
  inTotalOfs, outTotalOfs: int64; // used to cumulate in/out totals
  externalIP: string;
  noReplyBan: boolean;
  scrollFilesBox: integer = -1;
  defaultCfg: string;
  customIPservice: string;
  minDiskSpace: int64; // in MB
  selfTesting: boolean;
  fakingMinimize: boolean; // user clicked the [X] but we simulate the [_]
  quitASAP: boolean;  // deferred quit
  quitting: boolean; // ladies, we're quitting
  easyMode: boolean = TRUE;
  autosaveVFS, autosaveCFG: Tautosave;
  warnManyItems: boolean = TRUE;
  exePath, cfgPath, tmpPath: string;
  runningOnRemovable: boolean;
  startupFilename: string;
  filesToAddQ: TstringDynArray;
  ipsEverConnected: TstringDynArray;
  listenOn: string;  // interfaces HFS should listen on
  sbarTextTimeout: Tdatetime;
  sbarIdxs: record  // indexes within the statusbar
    totalIn, totalOut, banStatus, customTpl, oos, out, notSaved: integer;
    end;
  graph: record
  	rate: integer;    // update speed
    lastOut, lastIn: int64; // save bytesSent and bytesReceived last values
    maxV: integer;    // max value in scale
    size: integer;    // height of the box
    samplesIn, samplesOut: array [0..3000] of integer; // 1 sample, 1 pixel
    beforeRecalcMax: integer;  // countdown
    end;
  dyndns: record
    url, lastResult, lastIP: string;
    user, pwd, host: string;
    active: boolean;
    end;
  userInteraction: record
    disabled: boolean;
    bakVisible: boolean;  // backup value for mainFrm.visible
    end;
  logFile: record
    filename: string;
    apacheFormat: string;
    apacheZoneString: string;
    end;
  loadingVFS: record
    resetLetBrowse, unkFK, disableAutosave, visOnlyAnon: boolean;
    end;
  lastDiffTpl: record
    f: Tfile;
    ofs: integer;
    end;

function getAccount(user: string):Paccount; 
procedure kickBannedOnes();
procedure repaintTray();

implementation

{$R *.dfm}
{$R data.res}

uses
  editDlg, newuserpassDlg, optionsDlg, utilLib, folderKindDlg,
  shellExtDlg, diffDlg, classesLib, ipsEverDlg;

var
  tpl, dmBrowserTpl, filelistTpl: Ttpl;    // template for generated pages
  autoupdatedFiles: TstringToIntHash;   // download counter for temp Tfile.s 
  iconsCache: TiconsCache;
  usersInVFS: TusersInVFS;    // keeps track of user/pwd in the VFS
  tplFrm: TeditFrm;
  progFrm: TprogressForm;

function isCommentFile(fn:string):boolean;
begin
result:=(fn=COMMENTS_FILE)
  or mainfrm.loadSingleCommentsChk.checked and isExtention(fn, COMMENT_FILE_EXT)
  or mainfrm.supportDescriptionChk.checked and sameText('descript.ion',fn)
end; // isCommentFile

function isFingerprintFile(fn:string):boolean;
begin
result:=mainfrm.fingerprintsChk.checked and isExtention(fn, '.md5')
end; // isFingerprintFile

function isEnabledAccountFor(conn:ThttpConn):boolean;
var
  a: Paccount;
begin
a:=getAccount(conn.request.user);
result:=assigned(a) and a.enabled and (a.pwd = conn.request.pwd);
end; // isEnabledAccountFor

function uploadAllowed(conn:ThttpConn; f:Tfile):boolean;
begin
result:=FALSE;
if (f = NIL) or not f.isRealFolder() then exit;
if f.isTemp() then f:=f.parent;
if f.uploadAccounts = NIL then exit;
result:=stringExists('*', f.uploadAccounts);
if result or not stringExists(conn.request.user, f.uploadAccounts) then exit;
result:=isEnabledAccountFor(conn);
end; // uploadAllowed

type
  TfileListing = class
  protected
  public
    dir: array of Tfile;
    constructor create();
    destructor Destroy; override;
    procedure fromFolder(folder:Tfile; conn:ThttpConn; recursive:boolean=FALSE;
      limit:integer=-1; doClear:boolean=TRUE; forArchive:boolean=FALSE);
    procedure sort(conn:ThttpConn);
    end;

constructor TfileListing.create();
begin
dir:=NIL;
end; // create

destructor TfileListing.destroy;
var
  i: integer;
begin
for i:=0 to length(dir)-1 do
  if dir[i].isTemp() then
    dir[i].Free;
inherited destroy;
end; // destroy

procedure TfileListing.sort(conn:ThttpConn);
var
  foldersBefore: boolean;
  sortBy: ( SB_NAME, SB_SIZE, SB_TIME, SB_DL );

  function compareFiles(item1,item2:pointer):integer;
  var
    f1, f2:Tfile;
  begin
  f1:=item1;
  f2:=item2;
  if foldersBefore and (f1.isFolder() <> f2.isFolder()) then
    begin
    if f1.isFolder() then result:=-1
    else result:=+1;
    exit;
    end;
  case sortby of
    SB_SIZE: result:=compare_(f1.size, f2.size);
    SB_TIME: result:=-compare_(f1.mtime, f2.mtime);
    SB_DL: result:=-compare_(f1.DLcount, f2.DLcount);
    else result:=ansiCompareText(f1.name, f2.name);
    end;
  end; // compareFiles

  procedure qsort(left, right:integer);
  var
    split, t: Tfile;
    i, j: integer;
  begin
  if left >= right then exit;
  application.processmessages();
  if conn.state = HCS_DISCONNECTED then exit;

  i:=left;
  j:=right;
  split:=dir[(i+j) div 2];
    repeat
    while compareFiles(dir[i], split) < 0 do inc(i);
    while compareFiles(split, dir[j]) < 0 do dec(j);
    if i <= j then
      begin
      t:=dir[i];
      dir[i]:=dir[j];
      dir[j]:=t;

      inc(i);
      dec(j);
      end
    until i > j;
  if left < j then qsort(left, j);
  if i < right then qsort(i, right);
  end; // qsort

var
  v: string;
begin
// caching
foldersBefore:=mainfrm.foldersbeforeChk.checked;
if defSorting = 'time' then sortBy:=SB_TIME
else if defSorting = 'size' then sortBy:=SB_SIZE
else if defSorting = 'hits' then sortBy:=SB_DL
else sortBy:=SB_NAME;
v:=TconnData(conn.data).params.values['sort'];
if v > '' then
  case v[1] of
    'n': sortBy:=SB_NAME;
    's': sortBy:=SB_SIZE;
    't': sortBy:=SB_TIME;
    'd': sortBy:=SB_DL;
    end;
qsort( 0, length(dir)-1 );
end; // sort

procedure TfileListing.fromFolder(folder:Tfile; conn:ThttpConn;
  recursive:boolean=FALSE; limit:integer=-1; doClear:boolean=TRUE;
  forArchive:boolean=FALSE);

var
  actualCount, physicalSize, toSkip: integer;
  attrFilter: TfileType;
  seeProtected, noEmptyFolders: boolean;
  filesFilter, foldersFilter: string;

  procedure addToListing(f:Tfile);
  begin
  if noEmptyFolders and f.isEmptyFolder(conn)
  and not uploadAllowed(conn, f) then exit;
  application.ProcessMessages();

  if toSkip > 0 then dec(toSkip)
  else
    begin
    if actualCount >= physicalSize then
      begin
      inc(physicalSize, 100);
      setLength(dir, physicalSize);
      end;
    dir[actualCount]:=f;
    inc(actualCount);
    end;

  if not recursive or not f.isFolder() then exit;
  setLength(dir,actualCount);
  fromFolder(f, conn, TRUE, limit, FALSE, forArchive);
  physicalSize:=length(dir);
  actualCount:=physicalSize;
  end; // addToListing

  function hasRightAttributes(a:integer):boolean;
  begin
  result:=(ftSystem in attrFilter) or (a and faSysFile = 0)
    and (ftHidden in attrFilter) or (a and faHidden = 0)
  end; // hasRightAttributes

  function allowedTo(f:Tfile):boolean;
  begin
  if conn = NIL then result:=FALSE
  else result:=(not (FA_VIS_ONLY_ANON in f.flags) or (conn.request.user = ''))
    and (seeProtected or f.accessFor(conn))
    and (not forArchive or not (FA_DL_FORBIDDEN in f.flags))
  end; // allowedTo

  procedure includeFilesFromDisk();
  var
    comments: TstringList; // we use the hashing mechanism of this class
    commentMasks: TStringDynArray;

    procedure loadDescriptionFile();
    var
      s, l, fn: string;
    begin
    if not mainfrm.supportDescriptionChk.checked then exit;
    s:=loadFile(folder.resource+'\descript.ion');
    if mainfrm.oemForIonChk.checked then
      OEMToCharBuff(@s[1], @s[1], length(s));
    while s > '' do
      begin
      l:=chopLine(s);
      if l = '' then continue;
      if l[1] <> '"' then fn:=chop(' ',l)
      else
        begin
        delete(l,1,1);
        fn:=chop('"',l);
        end;
      comments.Add(fn+'='+trim(l));
      end;
    end; // loadDescriptionFile

    procedure extractCommentsWithWildcards();
    var
      i: integer;
      s: string;
    begin
    i:=0;
    while i < comments.count do
      begin
      s:=comments.names[i];
      if ansiContainsStr(s, '?')
      or ansiContainsStr(s, '*') then
        begin
        addString(comments[i], commentMasks);
        comments.Delete(i);
        end
      else
        inc(i);
      end;
    end; // extractCommentsWithWildcards

    function getCommentMaskFor(fn:string):string;
    var
      i: integer;
      s, mask: string;
    begin
    for i:=0 to length(commentMasks)-1 do
      begin
      s:=commentMasks[i];
      mask:=chop('=', s);
      if fileMatch(mask, fn) then
        begin
        result:=s;
        exit;
        end;
      end;
    result:='';
    end; // getCommentMaskFor

  var
    f: Tfile;
    sr: TSearchRec;
    namesInVFS: TStringDynArray;
    n: TtreeNode;
  begin
  if (limit >= 0) and (actualCount >= limit) then exit;

  // collect names in the VFS at this level. supposed to be faster than existsNodeWithName(). 
  namesInVFS:=NIL;
  n:=folder.node.getFirstChild();
  while assigned(n) do
    begin
    addString(n.text, namesInVFS);
    n:=n.getNextSibling();
    end;

  comments:=TstringList.create;
  try
    comments.CaseSensitive:=FALSE;
    try comments.LoadFromFile(folder.resource+'\'+COMMENTS_FILE);
    except end;
    loadDescriptionFile();
    if findfirst(folder.resource+'\*', faAnyFile, sr) <> 0 then exit;
    extractCommentsWithWildcards();
      repeat
      // we don't list these entries
      if (sr.name = '.') or (sr.name = '..')
      or isCommentFile(sr.name) or isFingerprintFile(sr.name)
      or sameText(sr.name, DIFF_TPL_FILE)
      or not hasRightAttributes(sr.attr) or not fileMatch(
        ifThen(sr.Attr and faDirectory > 0, foldersFilter, filesFilter),
        sr.name)
      or stringExists(sr.name, namesInVFS)
      then continue;

      f:=Tfile.create(folder.resource+'\'+sr.name);
      if (FA_SOLVED_LNK in f.flags) and f.isFolder() then
        // sorry, but we currently don't support lnk to folders in real-folders
        begin
        f.Free();
        continue;
        end;
      f.comment:=comments.values[sr.name];
      if f.comment = '' then
        f.comment:=getCommentMaskFor(sr.name);
      f.size:=0;
      if f.isFile() then
        f.size:=sr.FindData.nFileSizeLow
          +int64(sr.FindData.nFileSizeHigh) shl 32;
      f.mtime:=filedateToDatetime(sr.Time);
      include(f.flags, FA_TEMP);
      f.node:=folder.node; // temporary nodes are bound to the parent's node
      addToListing(f);
      until (findNext(sr) <> 0) or (conn.state = HCS_DISCONNECTED)
        or (limit >= 0) and (actualCount >= limit);
    findClose(sr);
  finally
    comments.free;
    end
  end; // includeFilesFromDisk

  procedure includeItemsFromVFS();
  var
    f: Tfile;
    sr: TSearchRec;
    n: Ttreenode;
  begin
  { this folder has been dinamically generated, thus the node is not actually
  { its own... skip }
  if folder.isTemp() then exit;

  // include (valid) items from the VFS branch
  n:=folder.node.getFirstChild();
  while assigned(n) and (conn.state <> HCS_DISCONNECTED)
  and ((limit < 0) or (actualCount < limit)) do
    begin
    f:=n.data;
    n:=n.getNextSibling();

    // watching not allowed, to anyone
    if (FA_HIDDEN in f.flags) or (FA_HIDDENTREE in f.flags) then continue;
    // watching not allowed, to you (unless you are *)
    if not allowedTo(f) then continue;

    //**credo che le virtual folder non vengano filtrate da foldersFilter, mentre le real s
    if FA_VIRTUAL in f.flags then // links and virtual folders are virtual
      begin
      addToListing(f);
      continue;
      end;
    if FA_UNIT in f.flags then
      begin
      if directoryExists(f.resource+'\') then
        addToListing(f);
      continue;
      end;

    // filtered out
    if not fileMatch( ifThen(f.isFolder(), foldersfilter, filesfilter), f.name)
    then continue;

    // try to get more info about this item
    if findFirst(f.resource, faAnyFile, sr) = 0 then
      begin
      // update size and time
      with sr.FindData do f.size:=nFileSizeLow+int64(nFileSizeHigh) shl 32;
      try f.mtime:=filedateToDatetime(sr.Time)
      except f.mtime:=0 end;
      findClose(sr);
      if not hasRightAttributes(sr.attr) then continue;
      end
    else
      // why findFirst() failed? is it a shared folder?
      if not directoryExists(f.resource) then continue;
    addToListing(f);
    end;
  end; // includeItemsFromVFS

var
  data: TconnData;
begin
if doClear then dir:=NIL;

if not folder.isFolder()
or not folder.accessFor(conn)
or folder.hasRecursive(FA_HIDDENTREE)
or not (FA_BROWSABLE in folder.flags)
then exit;

toSkip:=0;
if assigned(conn) and assigned(conn.data) then
  begin
  data:=conn.data;
  toSkip:=StrToIntDef(data.params.Values['offset'], 0);
  if limit < 0 then
    limit:=StrToIntDef(data.params.Values['limit'], -1);
  end;

physicalSize:=length(dir);
actualCount:=physicalSize;
folder.getFiltersRecursively(filesFilter, foldersFilter);
// cache user options
seeProtected:=not mainfrm.hideProtectedItemsChk.Checked and not forArchive;
noEmptyFolders:=folder.hasRecursive(FA_HIDE_EMPTY_FOLDERS);
attrFilter:=[];
if mainfrm.listfileswithhiddenattributeChk.checked then
  include(attrFilter, ftHidden);
if mainfrm.listfileswithsystemattributeChk.checked then
  include(attrFilter, ftSystem);

try
  if folder.isRealFolder() and not (FA_HIDDENTREE in folder.flags)
  and allowedTo(folder) then includeFilesFromDisk();
  includeItemsFromVFS();
finally
  setLength(dir, actualCount)
  end;
end; // fromFolder

function getAccount(user:string):Paccount;
var
  i: integer;
begin
result:=NIL;
for i:=0 to length(accounts)-1 do
  if sameText(user, accounts[i].user) then
    begin
    result:= @accounts[i];
    exit;
    end;
end; // getAccount

function customizedTpl():boolean;
begin result:=tpl.fullText<>defaultTpl end;

function isSendingFile(data:Tconndata):boolean;
begin
with data.conn do
  result:=(state = HCS_REPLYING_BODY)
    and (reply.bodyMode in [RBM_FILE, RBM_STREAM])
end; // isSendingFile

function isDownloading(data:TconnData):boolean;
begin
result:=assigned(data) and data.countAsDownload
  and (data.conn.state in [HCS_REPLYING_BODY, HCS_REPLYING_HEADER, HCS_REPLYING])
end; // isDownloading

function isReceivingFile(data:Tconndata):boolean;
begin
with data.conn do
  result:=(state = HCS_POSTING) and (post.filename > '')
end; // isReceivingFile

function countConnectionsByIP(ip:string):integer;
var
  i: integer;
begin
result:=0;
i:=0;
while i < srv.conns.count do
  begin
  if ThttpConn(srv.conns[i]).address = ip then
  	inc(result);
  inc(i);
  end;
end; // countConnectionsByIP

function countDownloads(ip:string=''):integer;
var
  i: integer;
  c: ThttpConn;
begin
result:=0;
i:=0;
while i < srv.conns.count do
  begin
  c:=srv.conns[i];
  if isDownloading(c.data) and ((ip = '') or (c.address = ip)) then inc(result);
  inc(i);
  end;
end; // countDownloads

function countDLthreads(f:Tfile; ip:string=''):integer;
var
  i: integer;
  c: ThttpConn;
  d: TconnData;
begin
result:=0;
i:=0;
while i < srv.conns.count do
  begin
  c:=srv.conns[i];
  d:=c.data;
  if isDownloading(d) and ((ip = '') or (c.address = ip))
  and assigned(d.lastFile) and d.lastFile.same(f) then inc(result);
  inc(i);
  end;
end; // countDLthreads

function countIPs(onlyDownloading:boolean=FALSE):integer;
var
  i: integer;
  c: ThttpConn;
  ips: TStringDynArray;
begin
i:=0;
ips:=NIL;
while i < srv.conns.count do
  begin
  c:=srv.conns[i];
  if not onlyDownloading or isDownloading(c.data) then
    addArray(ips, c.address);
  inc(i);
  end;
uniqueStrings(ips);
result:=length(ips);
end; // countIPs

function bmp2str(bmp:Tbitmap):string;
var
  stream: Tstringstream;
	gif: TGIFImage;
begin
{ the gif component has a GDI object leak. it happens while reducing colors of
{ transparent images. this seems to be not a big problem since the icon
{ cache system was introduced, but a real fix would be nice. }
stream:=Tstringstream.create('');
gif:=TGIFImage.Create();

gif.ColorReduction:=rmQuantize;
gif.Compression:=gcLZW;
gif.Assign(bmp);
gif.SaveToStream(stream);
result:=stream.DataString;

gif.free;
stream.free;
end; // bmp2str

function pic2str(idx:integer):string;
var
  pic, pic2: Tbitmap;
begin
result:='';
if idx < 0 then exit;
if length(imagescache) < idx+1 then setlength(imagescache, idx+1);
result:=imagescache[idx];
if result > '' then exit;
pic:=Tbitmap.create;
mainfrm.images.GetBitmap(idx,pic);
// pic2 is the transparent version of pic
pic2:=Tbitmap.create;
pic2.Width:=mainfrm.images.Width;
pic2.height:=mainfrm.images.height;
pic2.TransparentMode:=tmFixed;
pic2.TransparentColor:=$2FFFFFF;
pic2.Transparent:=TRUE;
BitBlt(pic2.Canvas.Handle, 0,0,16,16, pic.Canvas.Handle, 0,0, SRCAND);
BitBlt(pic2.Canvas.Handle, 0,0,16,16, pic.Canvas.Handle, 0,0, SRCPAINT);

result:=bmp2str(pic2);
pic2.free;
pic.free;
imagescache[idx]:=result;
end; // pic2str

function str2pic(s:string):integer;
var
  stream: Tstringstream;
	gif: TGIFImage;
begin
for result:=0 to mainfrm.images.count-1 do
  if pic2str(result) = s then exit;
stream:=Tstringstream.create(s);
gif:=TGIFImage.Create();
gif.LoadFromStream(stream);
result:=mainfrm.images.AddMasked(gif.bitmap, gif.Bitmap.TransparentColor);
gif.Free;
stream.free;
end; // str2pic

function getImageIndexForFile(fn:string):integer;
var
  i, n: integer;
  ico: Ticon;
  shfi: TShFileInfo;
  s: string;
begin
fillChar(shfi, SizeOf(TShFileInfo), 0);
// documentation reports shGetFileInfo() to be working with relative paths too,
// but it does not actually work without the expandFileName()
shGetFileInfo( pchar(expandFileName(fn)), 0, shfi, SizeOf(shfi), SHGFI_SYSICONINDEX);
if shfi.iIcon = 0 then
  begin
  result:=DEFAULT_FILE_ICON;
  exit;
  end;
// as reported by official docs
if shfi.hIcon <> 0 then
  destroyIcon(shfi.hIcon);

// have we already met this sysidx before?
for i:=0 to length(sysidx2index)-1 do
  if sysidx2index[i].sysidx = shfi.iIcon then
  	begin
    result:=sysidx2index[i].idx;
    exit;
    end;
// found not, let's check deeper: byte comparison.
// we first add the ico to the list, so we can use pic2str()
ico:=Ticon.create;
systemimages.GetIcon(shfi.iIcon, ico);
i:=mainfrm.images.AddIcon(ico);
ico.free;

s:=pic2str(i);
n:=0;
while n < length(sysidx2index) do
  begin
  if pic2str(sysidx2index[n].idx) = s then
    begin // found, delete the duplicate
    mainfrm.images.Delete(i);
    setlength(imagescache, i);
    i:=sysidx2index[n].idx;
    break;
    end;
  inc(n);
  end;

n:=length(sysidx2index);
setlength(sysidx2index, n+1);
sysidx2index[n].sysidx:=shfi.iIcon;
sysidx2index[n].idx:=i;
result:=i;
end; // getImageIndexForFile

function getBaseTrayIcon(perc:real=0):Tbitmap;
var
  x: integer;
begin
result:=Tbitmap.create;
result.Width:=16;
result.Height:=16;
mainfrm.images.GetBitmap( ifThen(assigned(srv) and srv.active,24,30), result);
if perc > 0 then
  begin
  x:=round(14*perc);
  result.canvas.Brush.color:=clYellow;
  result.Canvas.FillRect(rect(1,7,x+1,15));
  result.canvas.Brush.color:=clGreen;
  result.Canvas.FillRect(rect(x+1,7,15,15));
  end;
end; // getBaseTrayIcon

procedure drawTrayIconString(cnv:Tcanvas; s:string);
var
  x, i, idx: integer;
begin
x:=10;
for i:=length(s) downto 1 do
	begin
  if s[i] = '%' then idx:=10
  else idx:=ord(s[i])-ord('0');
  mainfrm.numbers.Draw(cnv, x, 8, idx);
  dec(x,mainfrm.numbers.Width);
  end;
end; // drawTrayIconString

procedure repaintTray();
var
  bmp: Tbitmap;
  s: string;
begin
if quitting or (mainfrm = NIL) then exit;
bmp:=getBaseTrayIcon();
s:=trayShows;
if s = 'connections' then s:=intTostr(srv.conns.Count);
if s = 'downloads' then s:=intToStr(downloadsLogged);
if s = 'uploads' then s:=intToStr(uploadsLogged);
if s = 'hits' then s:=intToStr(hitsLogged);
if s = 'ips' then s:=intToStr(countIPs());
if s = 'ips-ever' then s:=intToStr(length(ipsEverConnected));

drawTrayIconString(bmp.canvas, s);
tray_ico.Handle:=bmpToHico(bmp);
tray_ico.Transparent:=FALSE;
bmp.free;
tray.setIcon(tray_ico);
end; // repaintTray

procedure resetTotals();
begin
hitsLogged:=0;
downloadsLogged:=0;
uploadsLogged:=0;
repainttray();
end; // resetTotals

procedure flash();
begin
FlashWindow(application.handle, TRUE);
if mainFrm.beepChk.checked then MessageBeep(MB_OK);
end; // flash

function localDNSget(ip:string):string;
var
  i: integer;
begin
for i:=0 to length(address2name) div 2-1 do
  if addressmatch(address2name[i*2+1], ip) then
    begin
    result:=address2name[i*2];
    exit;
    end;
result:='';
end; // localDNSget

function existsNodeWithName(name:string; parent:Ttreenode):boolean;
var
  n: Ttreenode;
begin
result:=FALSE;
if parent = NIL then parent:=rootNode;
if parent = NIL then exit;
while assigned(parent.data) and not Tfile(parent.data).isFolder() do
  parent:=parent.parent;
n:=parent.getFirstChild();
while assigned(n) do
  begin
  result:=sameText(n.text, name);
  if result then exit;
  n:=n.getNextSibling();
  end;
end; // existsNodeWithName

function getUniqueNodeName(start:string; parent:Ttreenode):string;
var
  i: integer;
begin
result:=start;
if not existsNodeWithName(result, parent) then exit;
i:=1;
  repeat
  result:=format('%s (%d)', [start,i]);
  inc(i);
  until not existsNodeWithName(result, parent);
end; // getUniqueNodeName

procedure updateDynDNS();
const
  ERRORS: array [1..10] of record code,msg:string; end = (
    (code:'badauth'; msg:'invalid user/password'),
    (code:'notfqdn'; msg:'incomplete hostname, required form aaa.bbb.com'),
    (code:'nohost'; msg:'specified hostname does not exist'),
    (code:'!yours'; msg:'specified hostname belongs to another username'),
    (code:'numhost'; msg:'too many or too few hosts found'),
    (code:'abuse'; msg:'specified hostname is blocked for update abuse'),
    (code:'dnserr'; msg:'server error'),
    (code:'911'; msg:'server error'),
    (code:'!donator'; msg:'an option specified requires payment'),
    (code:'badagent'; msg:'banned client')
  );

  function interpretResponse(s:string):string;
  var
    i: integer;
    code: string;
  begin
  code:='';
  result:='successful';
  if ansiContainsText(dyndns.url, 'no-ip.com/dns')
  or ansiContainsText(dyndns.url, 'no-ip.com/update') then
    begin
    // no-ip sometimes uses ':', sometimes '='
    delete(s, 1, poss([':','='], s));
    s:=trim(chopLine(s));
    case StrToIntDef(s, -1) of
      0, 1, 11, 12: exit;
      2: result:='specified hostname does not exist';
      3: result:='invalid username';
      4: result:='invalid password';
      5: result:='too many updates sent, you are blocked for 1 hour since now';
      6: result:='account disabled due to violation of No-IP terms of service';
      7: result:='invalid ip';
      8: result:='disable/locked hostname';
      9: result:='hostname configured as web redirect';
      13: result:='update not available for this hostname';
      14: result:='offline not configured';
      99: result:='client disabled';
      else result:='';
      end;
    if result > '' then
      begin
      result:='error: '+result;
      exit;
      end;
    end
  else
    begin
    code:=trim(lowercase(getTill(' ',s)));
    if stringExists(code, ['good','nochg']) then exit;
    for i:=1 to length(ERRORS) do
      if code = ERRORS[i].code then
        begin
        result:='error: '+ERRORS[i].msg;
        dyndns.active:=FALSE;
        exit;
        end;
    end;
  result:='unknown reply: '+trim(s);
  end; // interpretResponse

var
  s: string;
begin
if externalIP = '' then exit;
dyndns.lastIP:=externalIP;
mainfrm.setStatusBarText('Updating dynamic DNS...', 5);
s:=httpGet(AnsiReplaceStr(dyndns.url, '%ip%', externalIP));
if s > '' then dyndns.lastResult:=s;
if not mainfrm.logOtherEventsChk.checked then exit;
if length(s) > 30 then s:=intToStr(length(s))+' bytes reply'
else s:=interpretResponse(s);
mainfrm.add2log('DNS update requested for '+dyndns.lastIP+': '+s);
if not dyndns.active then
  msgDlg('DNS update failed.'#13+s+'.'#13'User intervention is required.',
    MB_ICONERROR);
mainfrm.setStatusBarText('');
end; // updateDynDNS

procedure disableUserInteraction();
begin
if userInteraction.disabled then exit;
userInteraction.disabled:=TRUE;
if mainFrm = NIL then userInteraction.bakVisible:=FALSE
else
  begin
  userInteraction.bakVisible:=mainfrm.visible;
  mainfrm.visible:=FALSE;
  end;
end; // disableUserInteraction

procedure reenableUserInteraction();
begin
if not userInteraction.disabled then exit;
userInteraction.disabled:=FALSE;
if assigned(mainFrm) then
  mainfrm.visible:=userInteraction.bakVisible;
end; // reenableUserInteraction

constructor TconnData.create(conn:ThttpConn);
begin
conn.data:=self;
self.conn:=conn;
time:=now();
lastActivityTime:=time;
downloadingWhat:=DW_UNK;
params:=THashedStringList.create;
end; // constructor

destructor TconnData.destroy;
begin
freeAndNIL(params);
if assigned(f) then
  begin
  closeFile(f^);
  freeAndNIL(f);
  end;
inherited destroy;
end; // destructor

constructor Tfile.create(fullpath:string);
begin
fullpath:=ExcludeTrailingPathDelimiter(fullpath);
icon:=-1;
size:=-1;
atime:=now();
mtime:=atime;
flags:=[];
setResource(fullpath);
if (resource > '') and directoryExists(resource) then flags:=flags+[FA_FOLDER, FA_BROWSABLE];
end; // create

constructor Tfile.createVirtualFolder(name:string);
begin
icon:=-1;
setResource('');
flags:=[FA_FOLDER, FA_VIRTUAL, FA_BROWSABLE];
self.name:=name;
atime:=now();
mtime:=atime;
end; // createVirtualFolder

constructor Tfile.createLink(name:string);
begin
icon:=-1;
setName(name);
atime:=now();
mtime:=atime;
flags:=[FA_LINK, FA_VIRTUAL];
end; // createLink

procedure Tfile.setResource(res:string);

  function sameDrive(f1,f2:string):boolean;
  begin
  result:=(length(f1) >= 2) and (length(f2) >= 2) and (f1[2] = ':')
    and (f2[2] = ':') and (upcase(f1[1]) = upcase(f2[1]));
  end; // sameDrive

begin
if isExtention(res, '.lnk') then
  begin
  setName(extractFileName(copy(res, 1, length(res)-4)));
  res:=resolveLnk(res);
  include(flags, FA_SOLVED_LNK);
  end;
res:=ExcludeTrailingPathDelimiter(res);

// in this case, drive letter may change. useful with pendrives.
if runningOnRemovable and sameDrive(exePath, res) then
  delete(res, 1,2);

resource:=res;
if (length(res) = 2) and (res[2] = ':') then // logical unit
  begin
  if not isRoot() then setName(res);
  include(flags, FA_UNIT);
  end
else
  if not isRoot() then setName(ExtractFileName(res));
size:=-1;
end; // setResource

procedure Tfile.setName(name:string);
begin
self.name:=name;
if node = NIL then exit;
node.Text:=name;
end; // setName

function Tfile.same(f:Tfile):boolean;
begin result:=(self = f) or (resource = f.resource) end;

function Tfile.toggle(att:TfileAttribute):boolean;
begin
if att in flags then exclude(flags, att)
else include(flags, att);
result:=att in flags
end;

function Tfile.isRoot():boolean;
begin result:=FA_ROOT in flags end;

function Tfile.isFolder():boolean;
begin result:=FA_FOLDER in flags end;

function Tfile.isLink():boolean;
begin result:=FA_LINK in flags end;

function Tfile.isTemp():boolean;
begin result:=FA_TEMP in flags end;

function Tfile.isFile():boolean;
begin result:=not ((FA_FOLDER in flags) or (FA_LINK in flags)) end;

function Tfile.isFileOrFolder():boolean;
begin result:=not (FA_LINK in flags) end;

function Tfile.isRealFolder():boolean;
begin result:=(FA_FOLDER in flags) and not (FA_VIRTUAL in flags) end;

function Tfile.isVirtualFolder():boolean;
begin result:=(FA_FOLDER in flags) and (FA_VIRTUAL in flags) end;

function Tfile.isEmptyFolder(conn:ThttpConn=NIL):boolean;
var
  listing: TfileListing;
begin
result:=FALSE;
if not isFolder() then exit;
listing:=TfileListing.create();
{ i fear it is not ok to use fromFolder() to know if the folder is empty,
{ because it gives empty also for unallowed folders. } 
listing.fromFolder( self, conn, FALSE, 1 );
result:= length(listing.dir) = 0;
listing.free;
end; // isEmptyFolder

// uses comments file
function Tfile.getDynamicComment(skipParent:boolean=FALSE):string;
var
  comments: TstringList;
begin
result:=comment;
if result > '' then exit;
if mainfrm.loadSingleCommentsChk.checked then
  result:=loadFile(resource+COMMENT_FILE_EXT);
if (result > '') or skipParent then exit;
comments:=TstringList.create;
try
  try
    comments.CaseSensitive:=FALSE;
    comments.LoadFromFile(resource+'\..\'+COMMENTS_FILE);
    result:=comments.values[name];
  except end
finally
  comments.free;
  end;
end; // getDynamicComment

function Tfile.getParent():Tfile;
begin
if isTemp() then result:=Tfile(node.data)
else if node.parent = NIL then result:=NIL
else result:=node.parent.data
end; // getParent

function Tfile.getDLcount():integer;
begin
if isTemp() then result:=autoupdatedFiles.getInt(resource)
else result:=FDLcount;
end; // getDLcount

procedure Tfile.setDLcount(i:integer);
begin
if isTemp() then autoupdatedFiles.setInt(resource, i)
else FDLcount:=i;
end; // setDLcount

function Tfile.getDLcountRecursive():integer;
var
  n: Ttreenode;
  i: integer;
  f: Tfile;
begin
if not isFolder() then
  begin
  result:=DLcount;
  exit;
  end;
result:=0;
if node = NIL then exit;
n:=node.getFirstChild();
if not isTemp() then
  while assigned(n) do
    begin
    f:=Tfile(n.data);
    if f.isFolder() then inc(result, f.getDLcountRecursive())
    else inc(result, f.FDLcount);
    n:=n.getNextSibling();
    end;
if isRealFolder() then
  for i:=0 to autoupdatedFiles.count-1 do
    if ansiStartsText(resource, autoupdatedFiles[i]) then
      inc(result, autoupdatedFiles.getIntByIdx(i));
end; // getDLcountRecursive

function Tfile.diskfree():int64;
var
  tmp: int64;
begin
if (FA_VIRTUAL in flags)
or not getDiskFreeSpaceEx(pchar(resource), result, tmp, NIL) then
  result:=0;
end; // diskfree

procedure Tfile.setupImage(newIcon:integer);
begin
icon:=newIcon;
setupImage();
end; // setupImage

procedure Tfile.setupImage();
begin
if icon >= 0 then node.Imageindex:=icon
else if FA_UNIT in flags then node.imageIndex:=31
else if FA_ROOT in flags then node.ImageIndex:=1
else if FA_LINK in flags then node.ImageIndex:=4
else
  if FA_FOLDER in flags then
    if FA_VIRTUAL in flags then node.ImageIndex:=6
    else node.ImageIndex:=19
  else
    if mainfrm.useSystemIconsChk.checked and (resource > '') then
      node.ImageIndex:=getImageIndexForFile(resource) // skip iconsCache
    else
      node.ImageIndex:=7;
node.SelectedIndex:=node.imageindex;
end; // setupImage

function encodeURL(s:string; fullEncode:boolean=FALSE):string;
begin
if fullEncode or mainFrm.encodenonasciiChk.checked then
  s:=ansiToUTF8(s);
result:=HSlib.encodeURL(s, mainFrm.encodeNonasciiChk.checked,
  fullEncode or mainFrm.encodeSpacesChk.checked)
end; // encodeURL

function totallyEncoded(s:string):string;
var
  i: integer;
begin
result:='';
for i:=1 to length(s) do
  result:=result+'%'+intToHex(ord(s[i]),2)
end; // totallyEncoded

function Tfile.relativeURL(fullEncode:boolean=FALSE):string;
begin
if isLink() then result:=x_template(resource, ['%ip%', defaultIP])
else if isRoot() then result:=''
else result:=encodeURL(name, fullEncode)+ifThen(isFolder(),'/')
end;

function Tfile.pathTill(root:Tfile=NIL; delim:char='\'):string;
var
  f: Tfile;
begin
result:='';
if self = root then exit;
result:=name;
f:=parent;
if isTemp() then
  begin
  result:=copy(resource, length(f.resource)+2, length(resource));
  if delim <> '\' then result:=ansiReplaceStr(result,'\', delim);
  end;
while assigned(f) and (f <> root) and (f <> rootFile) do
  begin
  result:=f.name+delim+result;
  f:=f.parent;
  end;
end; // pathTill

function Tfile.url(fullEncode:boolean=FALSE):string;
begin
assert(node<>NIL, 'node can''t be NIL');
if isLink() then result:=relativeURL(fullEncode)
else result:='/'+encodeURL(pathTill(rootFile,'/'), fullEncode)
  +ifThen(isFolder() and not isRoot(), '/');
end; // url

function Tfile.getFolder():string;
var
  f: Tfile;
  s: string;
begin
result:='/';
f:=parent;
while assigned(f) and assigned(f.parent) do
  begin
  result:='/'+f.name+result;
  f:=f.parent;
  end;
if not isTemp() then exit;
f:=parent; // f now points to the non-temporary ancestor item
s:=extractFilePath(resource);
s:=copy( s, length(f.resource)+2, length(s) );
result:=result+ansiReplaceStr(s,'\','/');
end; // getFolder

function Tfile.fullURL(userpwd:string=''; ip:string=''):string;
begin
result:=url();
if isLink() then exit;
if assigned(srv) and srv.active and (srv.port <> '80') and (pos(':',ip) = 0)
and not mainfrm.noPortInUrlChk.checked then
  result:=':'+srv.port+result;
if ip = '' then ip:=defaultIP;
result:='http://'+nonEmptyConcat('',userpwd,'@')+ip+result
end; // fullURL

function Tfile.isNew():boolean;
var
  t: Tdatetime;
begin
if FA_TEMP in flags then t:=mtime
else t:=atime;
result:=(filesStayFlaggedForMinutes > 0)
  and (trunc(abs(now()-t)*24*60) <= filesStayFlaggedForMinutes)
end; // isNew

function Tfile.isPublic():boolean;
var
  f: Tfile;
begin
result:=FALSE;
f:=self;
while assigned(f) do
  begin
  if (f.user > '') or (length(f.accounts) > 1)
  or (length(f.accounts) = 1) and not stringExists(USER_ANONYMOUS, f.accounts)
  then exit;
  f:=f.parent;
  end;
result:=TRUE;
end;

function Tfile.getRecursiveDiffTplAsStr(outInherited:Pboolean=NIL;
  outFromDisk:Pboolean=NIL):string;
var
  incDiff, basePath, runPath, s: string;
  f: Tfile;
  first: boolean;
begin
result:='';
runPath:='';
f:=self;
if assigned(outInherited) then outInherited^:=FALSE;
if assigned(outFromDisk) then outFromDisk^:=FALSE;
first:=TRUE;
while assigned(f) do
  begin
  incDiff:='';
  if f.isRealFolder() then
    if f.isTemp() then
      begin
      basePath:=excludeTrailingPathDelimiter( extractFilePath(f.parent.resource) );
      runPath:=copy(f.resource, length(basePath)+2, length(f.resource));
      f:=f.parent;
      end
    else
      begin
      basePath:=excludeTrailingPathDelimiter(extractFilePath(f.resource));
      runPath:=extractFileName(f.resource);
      end;
  // temp realFolder will cycle more than once, while non-temp only once
  while runPath > '' do
    begin
    incDiff:=loadFile(basePath+'\'+runPath+'\'+DIFF_TPL_FILE);
    if incDiff > '' then
      begin
      if assigned(outFromDisk) then outFromDisk^:=TRUE;
      incDiff:=incDiff+CRLF;
      end;
    runPath:=excludeTrailingPathDelimiter(ExtractFilePath(runPath));
    end;
  // consider the diffTpl in node
  s:=f.diffTpl;
  if singleLine(s) then
    begin
    // maybe it refers to a file 
    s:=trim(s);
    if fileExists(s) then s:=loadfile(s)
    else if fileExists(f.resource+'\'+s) then s:=loadfile(f.resource+'\'+s);
    end;
  if s > '' then
    incDiff:=incDiff+s+CRLF;

  result:=result+incDiff;
  if not first and (incDiff > '') and assigned(outInherited) then
    outInherited^:=TRUE;
  f:=f.parent;
  first:=FALSE;
  end;
end; // getRecursiveDiffTplAsStr

function Tfile.getDefaultFile():Tfile;
var
  f: Tfile;
  mask, s: string;
  sr: TsearchRec;
  n: Ttreenode;
begin
result:=NIL;
mask:=getRecursiveFileMask();
if mask = '' then exit;

n:=node.getFirstChild();
{ if this folder has been dinamically generated, the treenode is not actually
{ its own, and we won't care about subitems }
if not isTemp() then
  while assigned(n) do
    begin
    f:=n.data;
    n:=n.getNextSibling();
    if (FA_LINK in f.flags) or f.isFolder()
    or not fileMatch(mask, f.name) or not FileExists(f.resource) then continue;
    result:=f;
    exit;
    end;

if not isRealFolder() then exit;
while mask > '' do
  begin
  s:=chop(';', mask);
  if findFirst(resource+'\'+s, faAnyFile-faDirectory, sr) <> 0 then continue;
  // encapsulate for returning
  result:=Tfile.create(resource+'\'+sr.name);
  include(result.flags, FA_TEMP);
  result.node:=node; // temporary nodes are bound to the parent's node
  findClose(sr);
  exit;
  end;
end; // getDefaultFile

function Tfile.countAsDownload():boolean;
var
  f: Tfile;
  mask: string;
begin
result:=not (FA_DONT_COUNT_AS_DL in flags);
if not result then exit;
f:=self;
  repeat
  mask:=f.dontCountAsDownloadMask;
  f:=f.parent;
  until (f = NIL) or (mask > '');
if mask > '' then result:=not fileMatch(mask, name)
end; // countAsDownload

function Tfile.getShownRealm():string;
var
  f: Tfile;
begin
f:=self;
  repeat
  result:=f.realm;
  if result > '' then exit;
  f:=f.parent;
  until f = NIL;
if mainfrm.useCommentAsRealmChk.checked then
  result:=getDynamicComment();
end; // getShownRealm

function Tfile.parentURL():string;
var
  i: integer;
begin
result:=url(TRUE);
i:=length(result)-1;
while (i > 1) and (result[i] <> '/') do dec(i);
setlength(result,i);
end; // parentURL

function Tfile.getImageIndex():integer;
var
  ic: PcachedIcon;
  i: integer;
begin
result:=icon;
if result >= 0 then exit;
if isFile() then
  for i:=0 to length(iconMasks)-1 do
    if fileMatch(iconMasks[i].str, name) then
      begin
      result:=iconMasks[i].int;
      exit;
      end;
ic:=iconsCache.get(resource);
if ic = NIL then
  begin
  result:=getImageIndexForFile(resource);
  iconsCache.put(resource, result, mtime);
  exit;
  end;
if mtime <= ic.time then result:=ic.idx
else
  begin
  result:=getImageIndexForFile(resource);
  ic.time:=mtime;
  ic.idx:=result;
  end;
end; // getImageIndex

procedure Tfile.lock();
begin locked:=TRUE end;

procedure Tfile.unlock();
begin locked:=FALSE end;

function Tfile.isLocked():boolean;
var
  f: Tfile;
  n: Ttreenode;
begin
// check ancestors (first, because it is always fast)
f:=self;
  repeat
  result:=f.locked;
  f:=f.parent;
  until (f = NIL) or result;
// check descendants
n:=node.getFirstChild();
while assigned(n) and not result do
  begin
  result:=Tfile(n.Data).isLocked();
  n:=n.getNextSibling();
  end;
end; // isLocked

procedure Tfile.recursiveApply(callback:TfileCallback; par:integer=0);
var
  n: Ttreenode;
begin
if not callback(self, par) then exit;
n:=node.getFirstChild();
while assigned(n) do
  begin
  if assigned(n.data) then Tfile(n.data).recursiveApply(callback, par);
  n:=n.getNextSibling();
  end;
end; // recursiveApply

function Tfile.getVisible():boolean;
var
  f: Tfile;
begin
result:=TRUE;
f:=self;
while result and assigned(f) do
  begin
  result:=not (FA_HIDDEN in f.flags) and not (FA_HIDDENTREE in f.flags);
  f:=f.parent;
  end;
end; // getVisible

function Tfile.hasRecursive(attribute: TfileAttribute;
  outInherited:Pboolean=NIL):boolean;
var
  f: Tfile;
begin
result:=FALSE;
f:=self;
if assigned(outInherited) then outInherited^:=FALSE;
while assigned(f) do
  begin
  result:=attribute in f.flags;
  if result then exit;
  f:=f.parent;
  if assigned(outInherited) then outInherited^:=TRUE;
  end;
if assigned(outInherited) then outInherited^:=FALSE; // grant it is set only if result=TRUE
end; // hasRecursive

function Tfile.accessFor(conn:ThttpConn):boolean;
begin
if conn = NIL then result:=accessFor('', '')
else result:=accessFor(conn.request.user, conn.request.pwd)
end; // accessFor

function Tfile.accessFor(username, password:string):boolean;
var
  a: Paccount;
  f: Tfile;
begin
result:=FALSE;
if isFile() and (FA_DL_FORBIDDEN in flags) then exit;
result:=TRUE;
f:=self;
while assigned(f) and not assigned(f.accounts) and (f.user = '') do
  f:=f.parent;
if (f = NIL) then exit; // no restriction found
if (username = '') and stringExists(USER_ANONYMOUS, f.accounts) then exit;
// first check in user/pass
if (f.user > '') and sameText(f.user, username)
and (f.pwd = password) then exit;
// then in accounts
result:=stringExists(username, f.accounts);
if not result then exit;
a:=getAccount(username);
result:=assigned(a) and a.enabled and (password = a.pwd);
end; // accessFor

function Tfile.getRecursiveFileMask():string;
var
  f: Tfile;
begin
f:=self;
  repeat
  result:=f.defaultFileMask;
  if result > '' then exit;
  f:=f.parent;
  until f = NIL;
end; // getRecursiveFileMask

function Tfile.accessList(specialUsernames:boolean=FALSE;
  outInherited:Pboolean=NIL):TstringDynArray;
var
  i: integer;
  f: Tfile;
  a: Paccount;
  s: string;
begin
result:=NIL;
f:=self;
if assigned(outInherited) then outInherited^:=FALSE;
while assigned(f) do
  begin
  for i:=0 to length(f.accounts)-1 do
  	begin
    s:=f.accounts[i];
    a:=getAccount(s);
    if assigned(a) and a.enabled
    or specialUsernames and (s>'') and (s[1] = '@') then addString(s, result);
    end;
  if f.user > '' then addString(f.user, result);
  if assigned(result) then exit;
  if assigned(outInherited) then outInherited^:=TRUE;
  f:=f.parent;
  end;
end; // accessList

procedure Tfile.getFiltersRecursively(var files,folders:string);
var
  f: Tfile;
begin
files:='';
folders:='';
f:=self;
while assigned(f) do
  begin
  if (files = '') and (f.filesfilter > '') then files:=f.filesFilter;
  if (folders = '') and (f.foldersfilter > '') then folders:=f.foldersFilter;
  if (files > '') and (folders > '') then break;
  f:=f.parent;
  end;
end; // getFiltersRecursively

function getSafeHost(conn:ThttpConn):string;
begin
if conn = NIL then result:=''
else result:=stripChars(conn.getHeader('host'), ['0'..'9','a'..'z','A'..'Z','.',':','-','_'], FALSE)
end; // getSafeHost

procedure kickByIP(ip:string);
var
  i: integer;
  d: TconnData;
begin
i:=0;
while i < srv.conns.count do
  begin
  with ThttpConn(srv.conns[i]) do
    if (address = ip) or (ip = '*') then
      begin
      d:=data;
      if assigned(d) and (d.disconnectReason = '') then
        d.disconnectReason:='kicked';
      disconnect();
      end;
  inc(i);
  end;
end; // kickByIP

function nodeIsLocked(n:Ttreenode):boolean;
begin
result:=FALSE;
if (n = NIL) or (n.data = NIL) then exit;
result:=Tfile(n.data).isLocked();
end; // nodeIsLocked

function banAddress(ip:string):boolean;
const
  MSG = 'There are %d open connections from this address.'#13+
        'Do you want to kick them all now?';
  MSG2 = 'You can edit the address.'#13'Masks and ranages are allowed.';
var
  i: integer;
  comm: string;
begin
result:=FALSE;
mainfrm.setFocus();
if not InputQuery('IP mask',MSG2,ip) then exit;

for i:=0 to length(banlist)-1 do
  if banlist[i].ip = ip then
    begin
    msgDlg('This IP address is already banned', MB_ICONWARNING);
    exit;
    end;

comm:='';
if not InputQuery('Ban comment','A comment for this ban...',comm) then exit;

i:=length(banlist);
setlength(banlist, i+1);
banlist[i].ip:=ip;
banlist[i].comment:=comm;

i:=countConnectionsByIP(ip);
if (i > 0) and (msgDlg(format(MSG,[i]), MB_ICONQUESTION+MB_YESNO) = IDYES) then
  kickByIP(ip);
result:=TRUE;
end; // banAddress

function uptimestr():string;
var
  t: Tdatetime;
begin
result:='server down';
if not srv.active then exit;
t:=now()-uptime;
result:=ifThen(t>1, format('(%d days) ',[trunc(t)]) )
  +formatDateTime('hh:nn:ss',t)
end; // uptimeStr

function getBaseSymbolsTable(folder:Tfile; conn:ThttpConn;
  tpl:Ttpl):TStringDynArray;
var
  comment, fdurl: string;
  bytes: int64;
  i: integer;
  a: TStringDynArray;
begin
result:=NIL;
addArray(result, [
  '%connections%', intToStr(srv.conns.Count),
  '%style%', tpl['style'],
  '\n', CRLF,
  '%timestamp%', optUTF8(tpl.utf8, dateTimeToStr(now())),
  '%version%', VERSION,
  '%build%', VERSION_BUILD,
  '%uptime%', optUTF8(tpl.utf8, uptimestr()),
  '%speed-out%', floatToStrF(srv.speedOut/1000, ffFixed, 7,2),
  '%speed-in%', floatToStrF(srv.speedIn/1000, ffFixed, 7,2),
  '%total-out%', smartSize(outTotalOfs+srv.bytesSent),
  '%total-in%', smartSize(inTotalOfs+srv.bytesReceived),
  '%total-downloads%', intToStr(downloadsLogged),
  '%total-hits%', intToStr(hitsLogged),
  '%total-uploads%', intToStr(uploadsLogged),
  '%number-addresses%', intToStr(countIPs()),
  '%number-addresses-ever%', intToStr(length(ipsEverConnected)),
  '%number-addresses-downloading%', intToStr(countIPs(TRUE))
]);
a:=tpl.getSyms();
for i:=0 to length(a)- 1 do
  addArray(result, ['%'+a[i]+'%', tpl[a[i]]]);

if assigned(conn) then
  addArray(result, [
    '%host%', getSafeHost(conn),
    '%ip%', conn.address
  ]);

if folder = NIL then exit;

comment:=folder.getDynamicComment();
if comment > '' then comment:=x_template(tpl['folder-comment'], [
  '%item-comment%', comment
]);

bytes:=folder.diskfree()-minDiskSpace*MEGA;

if folder.isRoot() then fdurl:='/'
else fdurl:=folder.getFolder()+folder.name+'/';

addArray(result, [
  '%up%', ifThen(not folder.isRoot(), tpl['up']),
  '%diskfree%', ifThen(bytes>0, smartSize(bytes), '0'),
  '%encoded-folder%', folder.url(TRUE),
  '%parent-folder%', folder.parentURL(),
  '%folder-comment%', optUTF8(tpl.utf8, comment),
  '%folder%', optUTF8(tpl.utf8, fdurl),
  '%folder-name%', optUTF8(tpl.utf8, folder.name)
]);
end; // getBaseSymbolsTable

function getBaseSymbolsTable_user(user:string; tpl:Ttpl):TStringDynArray;
begin
result:=NIL;
addArray(result, ['%user%', user]);
if user = '' then addArray(result, [
  '%loggedin%', '',
  '%login-link%', tpl['login-link']
]) else addArray(result, [
  '%loggedin%', tpl['loggedin'],
  '%login-link%', ''
])
end; // getBaseSymbolsTable_user

function loadMD5for(fn:string):string;
begin
if getMtimeUTC(fn+'.md5') < getMtimeUTC(fn) then result:=''
else result:=trim(getTill(' ',loadfile(fn+'.md5')))
end; // loadMD5for

function shouldRecur(data:TconnData):boolean;
begin
result:=mainFrm.recursiveListingChk.checked
  and (data.params.IndexOf('recursive') >= 0)
end; // shouldRecur

function Tmainfrm.getFolderPage(folder:Tfile; conn:ThttpConn;
  // we pass the Tpl as Tobject because symbol Ttpl is not defined yet
  otpl:Tobject):string;
var
  s, list, fileTpl, folderTpl, linkTpl, comment, name, fingerprint: string;
  table: TStringDynArray;
  i, ofsRelUrl, numberFiles, numberFolders: integer;
  counterForFolders, hideExt: boolean; // cache the value
  totalBytes: int64;
  fast: TfastStringAppend;
  buildTime: Tdatetime;
  listing: TfileListing;
  f: Tfile;
  diffTpl: Ttpl;
  data: TconnData;
  isDMbrowser: boolean;
  hasher: Thasher;
  fullEncode, recur: boolean;

  procedure applySequential();
  const
    PATTERN = '%sequential%';
  var
    idx, p: integer;
    idxS: string;
  begin
  idx:=0;
    repeat
    p:=pos(PATTERN, result);
    if p = 0 then exit;
    inc(idx);
    idxS:=intToStr(idx);
    delete(result, p, length(PATTERN)-length(idxS));
    move(idxS[1], result[p], length(idxS));
    until false;
  end; // applySequential

begin
result:='';
if (folder = NIL) or not folder.isFolder() then exit;

buildTime:=now();
data:=conn.data;
conn.addHeader('Cache-Control: no-cache');
recur:=shouldRecur(data);

listing:=TfileListing.create();
listing.fromFolder( folder, conn, recur );
listing.sort(conn);

diffTpl:=Ttpl.create;
diffTpl.over:=otpl as Ttpl;
if otpl <> filelistTpl then
  diffTpl.fullText:=optUTF8(diffTpl.over.utf8, folder.getRecursiveDiffTplAsStr());

isDMbrowser:= otpl = dmBrowserTpl;
fullEncode:=not isDMbrowser;
ofsRelUrl:=length(folder.url(fullEncode))+1;
hasher:=Thasher.create;
hasher.loadFrom(folder.resource);

// cache these values
fileTpl:=diffTpl['file'];
folderTpl:=diffTpl['folder'];
linkTpl:=diffTpl['link'];
counterForFolders:= pos('%item-dl-count%', folderTpl) > 0;
hideExt:=folder.hasRecursive(FA_HIDE_EXT);

// build %list% based on dir[]
numberFolders:=0;
numberFiles:=0;
totalBytes:=0;
fast:=TfastStringAppend.Create();
for i:=0 to length(listing.dir)-1 do
  begin
  application.ProcessMessages();
  if conn.state = HCS_DISCONNECTED then break;
  
  f:=listing.dir[i];
  if f.size > 0 then inc(totalBytes, f.size);

  // build up the table
  table:=NIL;
  if f.icon >= 0 then
    begin
    s:='~img'+intToStr(f.icon);
    addArray(table, ['~img_folder', s, '~img_link', s]);
    end;
  comment:=f.getDynamicComment(TRUE);
  if comment > '' then
    comment:=x_template(diffTpl['comment'],
      ['%item-comment%', comment] );
  name:=f.name;
  if f.isFile() then
    begin
    if useSystemIconsChk.checked or (f.icon >= 0) then
      addArray(table, ['~img_file', '~img'+intToStr(f.getImageIndex())]);
    if hideExt then
      setLength(name, length(name)-length(ExtractFileExt(name)) );
    addArray(table, [
      '%item-dl-count%', intToStr(f.DLcount),
      '%item-size-b%', intToStr(f.size),
      '%item-size-kb%', intToStr(f.size div KILO),
      '%item-size%', smartsize(f.size)
    ]);
    end;
  if recur then s:=optUTF8(diffTpl.utf8, f.getFolder())
  else s:='%folder%';
  addArray(table, ['%item-folder%', s]);
  if f.accessFor(conn) then s:=''
  else s:=diffTpl['protected'];
  addArray(table, ['%protected%', s]);

  // this may be heavy to calculate, only do it upon request
  if counterForFolders and f.isFolder() then
    addArray(table, [ '%item-dl-count%', intToStr(f.getDLcountRecursive()) ]);

  if f.isNew() then s:=diffTpl['newfile']
  else s:='';
  addArray(table, [
    '%new%', s,
    '%item-name%', optUTF8(diffTpl.utf8, name),
    '%item-ext%', optUTF8(diffTpl.utf8, copy(extractFileExt(f.name), 2, MAXINT)),
    '%item-added%', optUTF8(diffTpl.utf8, datetimeToStr(f.atime)),
    '%item-modified%', ifThen(f.mtime=0, 'error',
      optUTF8(diffTpl.utf8, datetimeToStr(f.mtime)))
  ]);

  // url building
  fingerprint:='';
  if fingerprintsChk.checked and f.isFile() then
    begin
    s:=loadMD5for(f.resource);
    if s = '' then s:=hasher.getHashFor(f.resource);
    if s > '' then fingerprint:='#!md5!'+s;
    end;
  if f.isLink() then
    s:=f.resource
  else if pwdInPagesChk.Checked and (conn.request.user > '') then
    s:=f.fullURL( encodeURL(conn.request.user)+':'+totallyEncoded(conn.request.pwd), getSafeHost(conn) )+fingerprint
  else if recur then
    s:=copy(f.url(fullEncode), ofsRelUrl, MAXINT)+fingerprint
  else
    s:=f.relativeURL(fullEncode)+fingerprint;
  addArray(table, [ '%item-url%', s ]);

  // so we can use symbols in comment
  addArray(table, [ '%comment%', x_template(optUTF8(diffTpl.utf8, comment), table) ]);

  // select appropriate template
  if FA_LINK in f.flags then s:=linkTpl
  else
    if f.isFolder() then
      begin
      s:=folderTpl;
      inc(numberFolders);
      end
    else
      begin
      s:=diffTpl.getTxtByExt(ExtractFileExt(f.name));
      if s = '' then s:=diffTpl.getTxtByExt(ExtractFileExt(f.name));
      if s = '' then s:=fileTpl;
      inc(numberFiles);
      end;

  fast.append( x_template(s, table) );
  end;
list:=fast.reset();
fast.free;
hasher.free;

// build final page
if list = '' then list:=diffTpl['nofiles']
else list:=x_template(diffTpl['files'], ['%list%',list]);

table:=getBaseSymbolsTable(folder, conn, diffTpl);
addArray(table, getBaseSymbolsTable_user(conn.request.user, diffTpl));
addArray(table, [
  '%upload-link%',
    ifThen(uploadAllowed(conn, folder), diffTpl['upload-link']),
  '%files%', list,
  '%number%', intToStr(numberFiles+NumberFolders),
  '%archive%', ifThen(folder.hasRecursive(FA_ARCHIVABLE) and (numberFiles+numberFolders > 0), diffTpl['archive']),
  '%number-files%', intToStr(numberFiles),
  '%number-folders%', intToStr(NumberFolders),
  '%total-bytes%', intToStr(totalBytes),
  '%total-kbytes%', intToStr(totalBytes div KILO),
  '%total-size%', smartsize(totalBytes)
]);

result:=diffTpl[''];
// apply 3 times the table because symbols can generate symbols
for i:=1 to 3 do
  if conn.state <> HCS_DISCONNECTED then
    result:=x_template(result, table);
applySequential();
result:=x_template(result, [
  '%build-time%', floatToStrF((now()-buildTime)*SECONDS, ffFixed, 7,3)
]);
listing.free;
diffTpl.free;
end; // getFolderPage

function getETA(data:TconnData):string;
begin
if (isSendingFile(data) or isReceivingFile(data))
and (data.eta.idx > ETA_FRAME) then result:=elapsedToStr(data.eta.result)
else result:='-'
end; // getETA

procedure setDefaultIP(v:string);
var
  old: string;
begin
old:=defaultIP;
if v > '' then defaultIP:=v
else if externalIP > '' then defaultIP:=externalIP
else defaultIP:=getIP();
if mainfrm = NIL then exit;
mainfrm.updateUrlBox();
v:=clipboard.AsText;
if (old = defaultIP) or (pos(old, v) = 0) then exit;
setClip( ansiReplaceText(v, old, defaultIP) );
end; // setDefaultIP

procedure Tmainfrm.getPage(section:string; data:TconnData; f:Tfile=NIL);
var
  table: TStringDynArray;
  diffTpl: Ttpl;

  procedure addProgressSymbols();
  var
    t, files, fn: string;
    i: integer;
    c: ThttpConn;
    d: TconnData;
    perc, speed: real;
    bytes, total: int64;
  begin
  if section <> 'progress' then exit;

  bytes:=0; speed:=0; total:=0; // shut up compiler
  files:='';
  i:=-1;
    repeat // a while-loop would look better but would lead to heavy indentation
    inc(i);
    if i >= srv.conns.count then break;
    c:=ThttpConn(srv.conns[i]);
    if c.address <> data.conn.address then continue;
    d:=c.data;
    fn:='';
    // fill fields
    if isReceivingFile(d) then
      begin
      t:=diffTpl['progress-upload-file'];
      fn:=ExtractFileName(d.uploadingFN); // already encoded by the browser 
      bytes:=c.bytesPosted;
      total:=c.post.length;
      speed:=c.speedIn;
      end;
    if isSendingFile(d) then
      begin
      if c.reply.bodymode <> RBM_FILE then continue;
      t:=diffTpl['progress-download-file'];
      fn:=optUTF8(diffTpl.utf8, d.lastFN);
      bytes:=c.bytesSentLastItem;
      total:=c.bytesPartial;
      speed:=c.speedOut;
      end;
    perc:=safeDiv(0.0+bytes, total);
    // no file exchange
    if fn = '' then continue;
    // apply fields
    files:=files+x_template(t, [
      '%perc%',intToStr( trunc(perc*100) ),
      '%filename%', fn,
      '%filename-js%', jsEncode(fn, '''"'),
      '%done-bytes%', intToStr(bytes),
      '%total-bytes%', intToStr(total),
      '%done%', smartsize(bytes),
      '%total%', smartsize(total),
      '%time-left%', getETA(d),
      '%speed-kb%', floatToStrF(speed/1000, ffFixed, 7,1)
    ]);
    until false;
  if files = '' then files:=diffTpl['progress-nofiles'];
  addArray(table, ['%progress-files%', files]);
  end; // addProgressSymbols

  procedure addUploadSymbols();
  var
    i: integer;
    files: string;
  begin
  if section <> 'upload' then exit;
  files:='';
  for i:=1 to 10 do
    files:=files+ x_template(diffTpl['upload-file'], ['%idx%',intToStr(i)]);
  addArray(table, ['%upload-files%', files]);
  end; // addUploadSymbols

  procedure addUploadResultsSymbols();
  var
    s, l, fn, reason, speed, files: string;
    size: int64;
  begin
  if section <> 'upload-results' then exit;
  s:=data.uploadResults;
  files:='';
  while s > '' do
    begin
    l:=chopLine(s);
    fn:=chop('|', l);
    try size:=strToInt64(chop('|', l))
    except size:=-1 end;
    reason:=chop('|', l);
    speed:=chop('|', l);
    l:=diffTpl[ ifThen(reason='','upload-success','upload-failed') ];
    files:=files+x_template(l,[
      '%item-name%', optUTF8(diffTpl.utf8, fn),
      '%item-url%', encodeURL(fn),
      '%item-size%', smartsize(size),
      '%reason%', optUTF8(diffTpl.utf8, reason),
      '%speed%', speed
    ]);
    end;
  addArray(table, ['%uploaded-files%', files]);
  end; // addUploadResultsSymbols

var
  s: string;
begin
if f = NIL then s:=''
else s:=f.getRecursiveDiffTplAsStr();
if trim(s) = '' then diffTpl:=tpl
else
  begin
  diffTpl:=Ttpl.create;
  diffTpl.fullText:=s;
  diffTpl.over:=tpl;
  end;

try
  table:=getBaseSymbolsTable(f, data.conn, diffTpl);
  addUploadSymbols();
  addProgressSymbols();
  addUploadResultsSymbols();
  if data = NIL then s:=''
  else
    begin
    s:=data.banReason;
    addArray(table, getBaseSymbolsTable_user(data.conn.request.user, diffTpl) );
    end;
  addArray(table, ['%reason%', optUTF8(diffTpl.utf8, s)]);

  data.conn.reply.bodyMode:=RBM_STRING;
  data.conn.reply.contentType:='text/html';

  data.conn.reply.mode:=HRM_REPLY;
  if section = 'ban' then data.conn.reply.mode:=HRM_DENY;
  if section = 'deny' then data.conn.reply.mode:=HRM_DENY;
  if section = 'not found' then data.conn.reply.mode:=HRM_NOT_FOUND;
  if section = 'unauthorized' then data.conn.reply.mode:=HRM_UNAUTHORIZED;
  if section = 'overload' then data.conn.reply.mode:=HRM_OVERLOAD;
  if section = 'max contemp downloads' then data.conn.reply.mode:=HRM_OVERLOAD;

  s:=diffTpl[section];
  if data.conn.reply.mode <> HRM_REPLY then
    s:=x_template(diffTpl['error-page'], ['%content%', s]);
  data.conn.reply.body:=x_template(x_template(s, table), table);

  compressReply(data);
finally
  if diffTpl <> tpl then diffTpl.free
  end
end; // getPage

function dirCrossing(s:string):boolean;
var
  i, n, l: integer;
begin
result:=TRUE;
i:=1;
l:=length(s);
while i <= l do
  begin
  // go forth counting dots
  n:=i;
  while (i <= l) and (s[i] = '.') do
    inc(i);
  n:=i-n;
  // condition for resulting true: 2 dots or more, and must be the whole chunk between path delimiters
  if (n >= 2) and ((i > l) or (s[i] = '\')) then
    exit;
  // move to next path delimiter
  i:=posEx('\', s, i);
  if i = 0 then break; // not found, the rest was just part of the name
  inc(i);
  end;
result:=FALSE; // we tried to the end of the string
end; // dirCrossing

function Tmainfrm.findFilebyURL(url:string):Tfile;
var
  s: string;
  cur, n: Ttreenode;
  found: boolean;
  f: Tfile;
begin
result:=NIL;
if (url = '') or anycharIn(#0, url) then exit;
url:=stringReplace(url, '//', '/', [rfReplaceAll,rfIgnoreCase]);
delete(url, 1,1);  // remove initial "/"
cur:=rootNode;   // lets place on the root
while url > '' do
  begin
  s:=chop('/', url);
  if s = '' then exit; // no support for null filenames
  found:=FALSE;
  // search inside the VFS
  n:=cur.getFirstChild();
  while assigned(n) do
    begin
    found:= sameText(n.Text, s) or sameText(n.Text, UTF8toAnsi(s));
    if found then break;
    n:=n.getNextSibling();
    end;
  // search on the disk if not found in the VFS
  f:=cur.data;
  if not found and f.isRealFolder() then
  	begin
    s:='\'+s+nonEmptyConcat('\', ansiReplaceStr(url,'/','\'));
    if dirCrossing(s) then exit;
    s:=f.resource+s; // we made the ".." test before, so ".." relative paths are allowed in the VFS
    if not fileOrDirExists(s) then
      if fileOrDirExists(s+'.lnk') then s:=s+'.lnk'
      else s:=UTF8ToAnsi(s);
    if fileOrDirExists(s) then
      begin
      // found on disk, we need to build a temporary Tfile to return it
      result:=Tfile.create(s);
      include(result.flags, FA_TEMP);
      // the temp file inherits flags from the real folder
      if FA_DL_FORBIDDEN in f.flags then include(result.flags, FA_DL_FORBIDDEN);
      if FA_DONT_LOG in f.flags then include(result.flags, FA_DONT_LOG);
      if not (FA_BROWSABLE in f.flags) then exclude(result.flags, FA_BROWSABLE);
      // temp nodes are bound to parent's node      
      result.node:=f.node;
      exit;
      end;
    end;
  if not found then exit;
  cur:=n;
  end;
result:=cur.data;
end; // findFileByURL

function Tmainfrm.sendPic(conn:ThttpConn; idx:integer=-1):boolean;
var
  s: string;
  data: TconnData;
  special: (no, graph);
begin
result:=FALSE;
data:=conn.data;
special:=no;
if idx < 0 then
  begin
  s:=conn.request.url;
  if not AnsiStartsText('/~img', s) then exit;
  delete(s,1,5);
  // converts special symbols
  if AnsiStartsText('_graph', s) then special:=graph else
  if AnsiStartsText('_link', s) then idx:=4 else
  if AnsiStartsText('_file', s) then idx:=7 else
  if AnsiStartsText('_folder', s) then idx:=6 else
  if AnsiStartsText('_lock', s) then idx:=12 else
    try idx:=strToInt(s) except exit end;
  end;

if (special = no) and ((idx < 0) or (idx >= images.count)) then exit;

case special of
  no: conn.reply.body:=pic2str(idx);
  graph: conn.reply.body:=getGraphPic(conn);
  end;
conn.reply.mode:=HRM_REPLY;
conn.reply.contentType:='image/gif';
conn.reply.bodyMode:=RBM_STRING;
data.downloadingWhat:=DW_ICON;
data.lastFN:=copy(conn.request.url,2,1000);

result:=TRUE;
end; // sendPic

function getAgentID(s:string):string; overload;
var
  res: string;

  function test(id:string):boolean;
  var
    i: integer;
  begin
  result:=FALSE;
  i:=pos(id,s);
  case i of
    0: exit;
    1: res:=chop(' ',s);
    else
      begin
      delete(s,1,i-1);
      res:=chop(';',s);
      end;
    end;
  result:=TRUE;
  end; // its

begin
result:=s;
if test('Crazy Browser')
or test('Opera')
or test('MSIE')
or test('Mozilla') then
  result:=res;
end; // getAgentID

function getAgentID(conn:ThttpConn):string; overload;
begin result:=getAgentID(conn.getHeader('User-Agent')) end;

procedure setupDownloadIcon(data:TconnData);

  procedure painticon();
  var
    bmp: Tbitmap;
    s: string;
    perc: real;
  begin
  perc:=safeDiv(0.0+data.conn.bytesSentLastItem, data.conn.bytesPartial);
  s:=intToStr( trunc(perc*100) )+'%';
  bmp:=getBaseTrayIcon(perc);
  drawTrayIconString(bmp.canvas, s);
  data.tray_ico.Handle:=bmpToHico(bmp);
  bmp.free;
  data.tray.setIcon(data.tray_ico);
  data.tray.setTip(
    ifThen( data.conn.reply.bodyMode=RBM_STRING, data.conn.request.url,
      data.lastFN )
    +trayNL+format('%.1f KB/s', [data.conn.speedOut/1000])
    +trayNL+dotted(data.conn.bytesSentLastItem)+' bytes sent'
    +trayNL+data.conn.address
  );
  data.tray.show();
  end; // paintIcon

begin
if (data = NIL) or (data.conn = NIL) then exit;
if assigned(data.tray)
and ((data.conn.state <> HCS_REPLYING_BODY) or
  (data.conn.bytesSentLastItem = data.conn.bytesPartial)) then
  begin
  data.tray.hide();
  freeAndNIL(data.tray);
  data.tray_ico.free;
  exit;
  end;
if not isSendingFile(data) then exit;

if not data.countAsDownload then exit;

if data.tray = NIL then
  begin
  data.tray:=TtrayIcon.create(mainfrm);
  data.tray.data:=data;
  data.tray_ico:=Ticon.create;
  data.tray.onEvent:=mainfrm.downloadTrayEvent;
  end;
if mainfrm.trayfordownloadChk.checked and isSendingFile(data) then
  paintIcon()
else data.tray.hide();
end; // setupDownloadIcon

function mimetypeByFile(f:Tfile):string;
var
  i: integer;
begin
for i:=0 to length(mimetypes) div 2-1 do
  if fileMatch(mimetypes[i*2], f.name) then
    begin
    result:=mimetypes[i*2+1];
    exit;
    end;
result:='application/octet-stream';
end; // mimetypeByFile

function getDynLogFilename(conn:ThttpConn=NIL):string;
var
  d, m, y, w: word;
  u: string;
begin
DecodeDateFully(now(), y,m,d,w);
if conn = NIL then u:=''
else u:='('+conn.request.user+')';
result:=x_template(logFile.filename, [
  '%d%', int0(d,2),
  '%m%', int0(m,2),
  '%y%', int0(y,4),
  '%dow%', int0(w-1,2),
  '%w%', int0(weekOf(now()),2),
  '%user%', u
]);
end; // getDynLogFilename

procedure Tmainfrm.add2log(lines:string; conn:ThttpConn=NIL; clr:Tcolor=clBlack);
var
  ts, first, rest, addr: string;
begin
if not logOnVideoChk.checked
and ((logFile.filename = '') or (logFile.apacheFormat > '')) then
  exit;

if logTimeChk.checked then
  if logDateChk.checked then ts:=datetimeToStr(now())
  else ts:=timeToStr(now())
else
  if logDateChk.checked then ts:=dateToStr(now())
  else ts:='';

addr:='';
if assigned(conn) then
  addr:=nonEmptyConcat('', conn.request.user, '@')
    +conn.address+':'+conn.port
    +nonEmptyConcat('{', localDNSget(conn.address), '}');

first:=chopLine(lines);
rest:='';
while lines > '' do
  rest:=rest+'> '+chopline(lines)+CRLF;
if rest > '' then includeTrailingString(rest, CRLF);

if (logFile.filename > '') and (logFile.apacheFormat = '') then
  appendFile(getDynLogFilename(conn),
    ts+nonEmptyConcat(' ',addr)+' '+first+CRLF+rest);

if not logOnVideoChk.checked then exit;
logbox.selstart:=length(logbox.Text);
logBox.SelAttributes.name:=logFontName;
logBox.SelAttributes.size:=logFontSize;
logBox.SelAttributes.Color:=clRed;
logBox.SelText:=ts+' ';
if addr > '' then
  begin
  logBox.SelAttributes.Color:=ADDRESS_COLOR;
  logBox.SelText:=addr+' ';
  end;
logBox.SelAttributes.Color:=clr;
logBox.SelText:=first+CRLF;
logBox.selAttributes.color:=clBlue;
logBox.SelText:=rest;
end; // add2log

function isBanned(conn:ThttpConn):boolean;
var
  i:integer;
begin
result:=FALSE;
for i:=0 to length(banlist)-1 do
  if addressMatch(banlist[i].ip, conn.address) then
    begin
    if assigned(conn.data) then
      TconnData(conn.data).banReason:=banlist[i].comment;
    result:=TRUE;
    exit;
    end;
end; // isBanned

procedure kickBannedOnes();
var
  i: integer;
  c: ThttpConn;
  d: TconnData;
begin
i:=0;
while i < srv.conns.count do
  begin
  c:=srv.conns[i];
  if isBanned(c) then
    begin
    d:=c.data;
    if assigned(d) and (d.disconnectReason = '') then
      d.disconnectReason:='kick banned';
    c.disconnect();
    end;
  inc(i);
  end;
end; // kickBannedOnes

function startServer():boolean;

  procedure tryPorts(list:array of string);
  var
    i: integer;
  begin
  for i:=0 to length(list)-1 do
    begin
    srv.port:=trim(list[i]);
    if srv.start(listenOn) then exit;
    end;
  end; // tryPorts

begin
result:=FALSE;
if srv.active then exit; // fail if already active

if (localIPlist.IndexOf(listenOn) < 0) and (listenOn <> '127.0.0.1') then
  listenOn:='';

if port > '' then
  tryPorts([port])
else
  tryPorts(['80','8080','280','10080','0']);
if not srv.active then exit; // failed
upTime:=now();
result:=TRUE;
end; // startServer

procedure stopServer();
begin if assigned(srv) then srv.stop() end;

procedure sayPortBusy();
var
  fn: string;
begin
try fn:=ExtractFileName(pid2file(port2pid(strToInt(srv.port))));
except fn:=''
  end;
msgDlg('Cannot open port.'#13'It is already used used by '+firstNonEmpty(fn, 'another software'), MB_ICONERROR);
end; // sayPortBusy

procedure toggleServer();
const
  MSG2 = 'There are %d connections open.'#13'Do you want to close them now?';
begin
if srv.active then stopServer()
else
  if not startServer() then
    sayPortBusy();
if (srv.conns.count = 0) or srv.active then exit;
if msgDLg(format(MSG2,[srv.conns.count]), MB_ICONQUESTION+MB_YESNO) = IDYES then
  kickByIP('*');
end; // toggleServer

function restartServer():boolean;
var
  port: string;
begin
result:=FALSE;
if not srv.active then exit;
port:=srv.port;
srv.stop();
srv.port:=port;
result:=srv.start(listenOn);
end; // restartServer

procedure updatePortBtn();
begin
if assigned(srv) then
  mainfrm.portBtn.Caption:=format('Port: %s', [
    ifThen(srv.active, srv.port, firstNonEmpty(port,'any'))]);
end;

procedure Tmainfrm.httpEvent(event:ThttpEvent; conn:ThttpConn);
var
  data: TconnData;
  f: Tfile;
  url, httpDate: string;
  acc: Paccount;

  procedure switchToDefaultFile();
  var
    default: Tfile;
  begin
  if (f = NIL) or not f.isFolder() then exit;
  default:=f.getDefaultFile();
  if default = NIL then exit;
  try if f.isTemp then f.Free;
  finally f:=default end;
  end; // switchToDefaultFile

  function averageUpSpeedFile():integer;
  begin
  result:=safeDiv(data.conn.bytesPostedLastItem,
    round(1000*(now()-data.fileXferStart)*SECONDS))
  end; // averageUpSpeedFile

  function averageDownSpeedFile():integer;
  begin
  result:=safeDiv(data.conn.bytesSentLastItem,
    round(1000*(now()-data.fileXferStart)*SECONDS))
  end; // averageDownSpeedFile

  procedure doLog();

    procedure apacheFormatLog();
    const
      APACHE_TIMESTAMP_FORMAT = 'dd"/"mmm"/"yyyy":"hh":"nn":"ss';
    var
      isPublic: boolean;
    begin
    if (logFile.filename = '') or (logFile.apacheFormat = '')
    or (data = NIL) or not data.logLaterInApache
    or not (event in [HE_LAST_BYTE_DONE, HE_DISCONNECTED]) then exit;

    isPublic:=assigned(data.lastFile) and data.lastFile.isPublic();
    data.logLaterInApache:=FALSE;
    appendFile(getDynLogFilename(conn), x_template(logfile.apacheFormat+CRLF, [
      '%h', conn.address,
      '%l', '-',
      '%u', ifThen((conn.request.user='') or isPublic, '-', conn.request.user),
      '%t', '['+formatDatetime(APACHE_TIMESTAMP_FORMAT,now())+' '
        +logfile.apacheZoneString+']',
      '%r', getTill(CRLF, conn.request.full),
      '%>s', intToStr(HRM2CODE[conn.reply.mode]),
      '%b', intToStr(conn.bytesSentLastItem),
      '%{Referer}i', conn.getHeader('Referer'),
      '%{User-agent}i', conn.getHeader('User-agent'),
      '\t', #9,
      '\n', #10,
      '\\', '\'
    ]));
    end; // apacheFormatLog

  var
    i: integer;
    url_: string; // an alias, final '_' is to not confuse with the other var
  begin
  if assigned(data) and data.dontLog then exit;

  if assigned(data) and (data.preReply = PR_BAN)
  and not logBannedChk.checked then exit;

  if conn = NIL then url_:=''
  else url_:=conn.request.url;
  if not (event in [HE_OPEN, HE_CLOSE, HE_CONNECTED, HE_DISCONNECTED]) then
    if not logIconsChk.checked
      and (ansiStartsText('/~img', url_) or ansiEndsStr('/favicon.ico', url_))
    or not logBrowsingChk.checked
      and assigned(conn) and (conn.reply.bodyMode = RBM_STRING)
    or not logProgressChk.checked and (url_ = '/~progress') then
      exit;

  if not (event in [HE_OPEN, HE_CLOSE])
  and addressMatch(dontLogAddressMask, conn.address) then
    exit;

  case event of
    HE_OPEN: if logServerstartChk.Checked then add2log('Server start');
    HE_CLOSE: if logServerstopChk.checked then add2log('Server stop');
    HE_CONNECTED: if logconnectionsChk.Checked then add2log('Connected', conn);
    HE_DISCONNECTED: if logDisconnectionsChk.checked then
      add2log('Disconnected'+ifThen(conn.disconnectedByServer, ' by server')
      +nonEmptyConcat(': ', data.disconnectReason)
      +ifThen(conn.bytesSent>0, ' - '+intToStr(conn.bytesSent)+' bytes sent'),
      conn
    );
    HE_GOT:
      begin
      i:=conn.bytesGot-data.lastBytesGot;
      if i <= 0 then exit;
      if logBytesreceivedChk.Checked then
        if now()-data.bytesGotGrouping.from > BYTES_GROUPING_THRESHOLD then
          begin
          add2log(format('Got %d bytes',[i+data.bytesGotGrouping.bytes]),conn);
          data.bytesGotGrouping.from:=now();
          data.bytesGotGrouping.bytes:=0;
          end
        else
          inc(data.bytesGotGrouping.bytes, i);
      inc(data.lastBytesGot, i);
      end;
    HE_SENT:
      begin
      i:=conn.bytesSent-data.lastBytesSent;
      if i <= 0 then exit;
      if logBytessentChk.checked then
        if now()-data.bytesSentGrouping.from > BYTES_GROUPING_THRESHOLD then
          begin
          add2log(format('Sent %d bytes',[i+data.bytesSentGrouping.bytes]),
            conn);
          data.bytesSentGrouping.from:=now();
          data.bytesSentGrouping.bytes:=0;
          end
        else
          inc(data.bytesSentGrouping.bytes, i);
      inc(data.lastBytesSent, i);
      end;
    HE_REQUESTED:
      if not logOnlyServedChk.checked
      or (conn.reply.mode in [HRM_REPLY, HRM_REPLY_HEADER, HRM_REDIRECT]) then
        begin
        data.logLaterInApache:=TRUE;
        if logRequestsChk.Checked then
          add2log(format('Requested %s %s', [ METHOD2STR[conn.request.method],
            decodeURL(conn.request.url) ]), conn);
        if dumprequestsChk.checked then
          add2log('Request dump'+CRLF+conn.request.full, conn);
        end;
    HE_REPLIED:
      begin
      if logOnlyServedChk.Checked
      and not (conn.reply.mode in [HRM_REPLY, HRM_REPLY_HEADER, HRM_REDIRECT])
      then exit;

      if logRepliesChk.checked then
        case conn.reply.mode of
          HRM_REPLY: if not data.fullDLlogged then add2log(format('Served %s',
            [smartSize(conn.bytesSentLastItem)]), conn);
          HRM_REPLY_HEADER: add2log('Served head', conn);
          HRM_NOT_MODIFIED: add2log('Not modified, use cache', conn);
          HRM_REDIRECT: add2log(format('Redirected to %s', [conn.reply.url]),
            conn);
          else
            add2log(format('Not served: %d - %s', [
              HRM2CODE[conn.reply.mode], HRM2STR[conn.reply.mode] ])
              +nonEmptyConcat(': ', data.error), conn);
          end;
      end;
    HE_POST_FILE:
      if logUploadsChk.checked then
        add2log(format('Uploading %s', [optAnsi(tpl.utf8, conn.post.filename)]), conn);
    HE_POST_END_FILE:
      if logUploadsChk.checked then
        add2log(format('Fully uploaded -  %s @ %d KB/s', [
          smartSize(data.conn.bytesPostedLastItem),
          averageUpSpeedFile() ]), conn);
    HE_LAST_BYTE_DONE:
      if logFulldownloadsChk.checked
      and (data.downloadingWhat in [DW_FILE, DW_ARCHIVE]) then
        begin
        data.fullDLlogged:=TRUE;
        add2log(format('Fully downloaded - %s @ %d KB/s - %s', [
          smartSize(data.conn.bytesSentLastItem), averageDownSpeedFile(),
          decodeURL(url_)]), conn);
        end;
    end;

  { apache format log is only related to http events, that's why it resides
  { inside httpEvent(). moreover, it needs to access to some variables. }
  apacheFormatLog();
  end; // doLog

  function limitsExceededOnConnection():boolean;
  begin
  if assigned(acc) and acc.noLimits then result:=FALSE
  else
    result:=(maxConnections>0) and (srv.conns.count > maxConnections)
      or (maxConnectionsIP>0)
        and (countConnectionsByIP(conn.address) > maxConnectionsIP)
      or (maxIPs>0) and (countIPs() > maxIPs)
  end; // limitsExceededOnConnection

  function limitsExceededOnDownload():boolean;
  begin
  result:=FALSE;
  if assigned(acc) and acc.noLimits then exit;

  data.disconnectReason:='';
  if (maxContempDLs > 0) and (countDownloads() > maxContempDLs)
  or (maxContempDLsIP > 0) and (countDownloads(conn.address) > maxContempDLsIP) then
    data.disconnectReason:='Max simultaneous downloads'
  else if (maxIPsDLing > 0) and (countIPs(TRUE) > maxIPsDLing) then
    data.disconnectReason:='Max simultaneous addresses downloading'
  else if preventLeechingChk.checked and (countDLthreads(f, conn.address) > 1) then
    data.disconnectReason:='Leeching';

  result:=data.disconnectReason > '';
  end; // limitsExceededOnDownload

  procedure extractParams();
  var
    s: string;
  begin
  s:=url;
  url:=chop('?',s);
  data.params.Clear();
  if s > '' then extractStrings(['&'], [], @s[1], data.params);
  end; // extractParams

  procedure closeUploadingFile();
  begin
  if data.f = NIL then exit;
  closeFile(data.f^);
  dispose(data.f);
  data.f:=NIL;
  end; // closeUploadingFile

  // close and eventually delete/rename
  procedure closeUploadingFile_partial();
  var
    fn: string;
  begin
  if (data = NIL) or (data.f = NIL) then exit;
  fn:=getFilename(data.f^);
  closeUploadingFile();
  if deletePartialUploadsChk.checked then deleteFile(fn)
  else if renamePartialUploads = '' then exit;
  if pos('%name%', renamePartialUploads) = 0 then
    renameFile(fn, fn+renamePartialUploads)
  else
    renameFile(fn, ExtractFilePath(fn)
      +x_template(renamePartialUploads,['%name%',ExtractFileName(fn)]) );
  end; // closeUploadingFile_partial

  function isDownloadManagerBrowser():boolean;
  begin
  result:=(pos('GetRight',data.agent)>0)
    or (pos('FDM',data.agent)>0)
    or (pos('FlashGet',data.agent)>0)
  end; // isDownloadManagerBrowser

  procedure logUploadFailed();
  begin
  if not logUploadsChk.checked then exit;
  add2log(format('Upload failed, %s: %s',
    [data.uploadFailed, optAnsi(tpl.utf8, conn.post.filename)]), conn);
  end; // logUploadFile

  function canWriteFile():boolean;
  var
    fn: string;
  begin
  result:=FALSE;
  if data.f = NIL then exit;
  fn:=getFilename(data.f^);
  if fn = '' then exit;
  result:= minDiskSpace <= DiskFree(filenameToDriveByte(fn)) div MEGA;
  if result then exit;
  closeUploadingFile_partial();
  data.uploadFailed:='Minimum disk space reached';
  logUploadFailed();
  end; // canWriteFile

  function getUploadDestinationFileName():string;
  var
    i: integer;
    fn, ext: string;
  begin
  fn:=optAnsi(tpl.utf8, conn.post.filename);
  result:=f.resource+'\'+fn;
  if not numberFilesOnUploadChk.checked then exit;
  ext:=extractFileExt(fn);
  setLength(fn, length(fn)-length(ext));
  i:=0;
  while fileExists(result) do
    begin
    inc(i);
    result:=format('%s\%s (%d)%s', [f.resource,fn,i,ext]);
    end;
  end; // getUploadDestinationFileName

  procedure serveTar();
  var
    listing: TfileListing;
    i, ofs: integer;
    tar: TtarStream;
    fi: Tfile;
    fIsTemp: boolean;
  begin
  if not f.hasRecursive(FA_ARCHIVABLE) then
    begin
    getPage('deny', data);
    exit;
    end;

  listing:=TfileListing.create();
  try
    tar:=TtarStream.create; // this is freed upon HE_DISCONNECTED event
    listing.fromFolder( f, conn, shouldRecur(data), -1, TRUE, TRUE );
    fIsTemp:=f.isTemp();
    ofs:=length(f.resource)-length(f.name)+1;
    for i:=0 to length(listing.dir)-1 do
      begin
      fi:=listing.dir[i];
      if conn.state = HCS_DISCONNECTED then break;
      if fi.isFile() then
        if not fIsTemp then
          tar.addFile(fi.resource, fi.pathTill(f.parent), ifThen(not fi.isTemp(), fi))
        else // pathTill won't work this case, because f.parent is an ancestor but not the parent
          tar.addFile(fi.resource, copy(fi.resource, ofs, MAXINT));
      end
  finally
    listing.free
    end;

  data.downloadingWhat:=DW_ARCHIVE;
  data.countAsDownload:=TRUE;
  data.fileXferStart:=now();
  conn.reply.mode:=HRM_REPLY;
  conn.reply.contentType:='application/x-tar';
  conn.reply.bodyMode:=RBM_STREAM;
  conn.reply.bodyStream:=tar;
  if f.name > '' then
    begin
    data.lastFN:=ifThen(f.name='/', 'home', f.name)+'.folder.tar';
    conn.addHeader('Content-Disposition: filename="'+data.lastFN+'";');
    end;
  end; // serveTar

  procedure handleRequest();

    function accessGranted(forceFile:Tfile=NIL):boolean;
    begin
    result:=FALSE;
    if assigned(forceFile) then f:=forceFile;
    if f = NIL then exit;
    if f.isFile() and (FA_DL_FORBIDDEN in f.flags) then
      begin
      getPage('deny', data);
      exit;
      end;
    result:=f.accessFor(conn);
    if result then exit;
    conn.reply.realm:=f.getShownRealm();
    getPage('unauthorized', data);
    if (forceFile = NIL) and stringExists(conn.request.user, f.accessList())
    and logOtherEventsChk.checked then
      add2log('Login failed', conn);
    end; // accessGranted

    function hostFromURL(url:string):string;
    begin
    chop('://', url);
    result:=chop('/', url);
    if pos('@', result) = 0 then exit;
    chop('@', result);
    end; // hostFromURL

    function isAllowedReferer():boolean;
    var
      r: string;
    begin
    result:=TRUE;
    if allowedReferer = '' then exit;
    r:=hostFromURL(conn.getHeader('Referer'));
    if r = '' then exit;
    result:=fileMatch(allowedReferer, r);
    end; // isAllowedReferer

  var
    b: boolean;
    s, urlCmd: string;
    i: integer;
  begin
  data.downloadingWhat:=DW_UNK;
  data.fullDLlogged:=FALSE;
  data.countAsDownload:=FALSE;

  if (length(conn.request.user) > 100) or anycharIn('/\:?*<>|', conn.request.user) then
    begin
    conn.reply.mode:=HRM_BAD_REQUEST;
    exit;
    end;

  acc:=getAccount(conn.request.user);
  if assigned(acc) and acc.noLimits then
    conn.ignoreSpeedLimit:=TRUE;

  conn.addHeader('Accept-Ranges: bytes');
  if sendHFSidentifierChk.checked then
    conn.addHeader('Server: HFS '+VERSION);

  case data.preReply of
    PR_OVERLOAD:
      begin
      data.disconnectReason:='limits exceeded';
      getPage('overload', data);
      end;
    PR_BAN:
      begin
      getPage('ban', data);
      conn.reply.reason:='Banned: '+data.banReason;
      end;
    end;

  if not (conn.request.method in [HM_GET,HM_HEAD,HM_POST]) then
    begin
    conn.reply.mode:=HRM_METHOD_NOT_ALLOWED;
    exit;
    end;
  inc(hitsLogged);

  if data.preReply <> PR_NONE then exit;

  url:=conn.request.url;
  extractParams();
  url:=decodeURL(url);

  // stay clean
  if assigned(data.lastFile) and data.lastFile.isTemp() then
    data.lastFile.Free;
  data.lastFile:=NIL; // new request...new file

  data.lastFN:=extractFileName( ansiReplaceStr(url,'/','\') );
  data.agent:=getAgentID(conn);

  if selfTesting and (url = 'test') then
    begin
    conn.reply.mode:=HRM_REPLY;
    conn.reply.bodyMode:=RBM_STRING;
    conn.reply.body:='HFS OK';
    exit;
    end;

  // all URIs must begin with /
  if (url = '') or (url[1] <> '/') then
    begin
    conn.reply.mode:=HRM_BAD_REQUEST;
    exit;
    end;

  if ansiStartsStr('/~img', url) then
    begin
    if not sendPic(conn) then
      getPage('not found', data);
    exit;
    end;

  f:=findFileByURL(url);
  if f = NIL then
    begin
    // maybe the file doesn't exist because the URL has a final command in it
    // move last url part from 'url' into 'urlCmd'
    i:=length(url);
    while (i > 1) and (url[i] <> '/') do dec(i);
    inc(i);
    urlCmd:=lowerCase(copy(url, i, MAXINT));
    delete(url, i, MAXINT);
    // we know an urlCmd must begin with ~
    // favicon is handled as an urlCmd: we furnish HFS icon.
    // an unexistant ~file will be detected a hundred lines below.
    if ansiStartsStr('~', urlCmd) or (urlCmd = 'favicon.ico') then
      f:=findFileByURL(url);
    end;
  if f = NIL then
    begin
    getPage('not found', data);
    exit;
    end;
  if f.isFolder() and not ansiEndsStr('/',url) then
    begin
    conn.reply.mode:=HRM_MOVED;
    conn.reply.url:=url+'/';
    exit;
    end;
  if f.isFolder() and (urlCmd = '') then
    switchToDefaultFile();
  if enableNoDefaultChk.checked and (urlCmd = '~nodefault') then
    urlCmd:='';

  if f.isRealFolder() and not directoryExists(f.resource)
  or f.isFile() and not fileExists(f.resource) then
    begin
    getPage('not found', data);
    exit;
    end;
  if not accessGranted() then
    exit;
  if urlCmd = 'favicon.ico' then
    begin
    sendPic(conn, 23);
    exit;
    end;

  if urlCmd = '~login' then
    begin
    if assigned(acc) and (acc.pwd = conn.request.pwd)
    or usersInVFS.match(conn) then
      begin
      conn.reply.mode:=HRM_REDIRECT;
      conn.reply.url:=url;
      end
    else
      begin
      getPage('unauthorized', data);
      if loginRealm > '' then conn.reply.realm:=loginRealm;
      end;
    exit;
    end;

  b:=urlCmd = '~upload+progress';
  if (b or (urlCmd = '~upload') or (urlCmd = '~upload-no-progress')) then
    begin
    if not f.isRealFolder() then
      getPage('deny', data)
    else if not uploadAllowed(conn, f) then
      getPage('unauthorized', data)
    else
      getPage( ifThen(b,'upload+progress','upload'), data, f);
    if b then data.disconnectAfterReply:=TRUE; // fix for IE6
    exit;
    end;

  if (conn.request.method = HM_POST) and (data.uploadResults > '') then
    begin
    getPage('upload-results', data, f);
    data.uploadResults:='';
    exit;
    end;

  if f.isFolder() and not (FA_BROWSABLE in f.flags) then
    begin
    getPage('deny', data);
    exit;
    end;

  if not isAllowedReferer()
  or f.isFile() and (FA_DL_FORBIDDEN in f.flags) then
    begin
    getPage('deny', data);
    exit;
    end;

  if urlCmd = '~folder.tar' then
    begin
    serveTar();
    exit;
    end;

  // we accept also ~files.lst.m3u
  if ansiStartsStr('~files.lst', urlCmd) then
    begin
    s:=cfgPath+FILELIST_TPL_FILE;
    if fileExists(s) and (lastFilelistTpl < getMtimeUTC(s)) then
      begin
      lastFilelistTpl:=getMtimeUTC(s);
      filelistTpl.fullText:=loadfile(s);
      end;
    if not fileExists(s) and (lastFilelistTpl > 0) then
      begin
      lastFilelistTpl:=0;
      filelistTpl.fullText:=getRes('filelistTpl');
      end;

    data.downloadingWhat:=DW_FOLDERPAGE;
    data.disconnectAfterReply:=TRUE; // needed for IE6... ugh...
    conn.reply.mode:=HRM_REPLY;
    conn.reply.bodyMode:=RBM_STRING;
    conn.reply.body:=trim(getFolderPage(f, conn, filelistTpl));
    conn.reply.contentType:='text/plain';
    compressReply(data);
    exit;
    end;

  // provide access to any [section] in the tpl, included [progress]
  if (f = rootFile) and (urlCmd > '') then
    begin
    s:=copy(urlCmd,2,MAXINT);
    if (s > '') and tpl.sectionExist(s) then
      begin
      getPage(s, data);
      exit;
      end;
    end;

  // from here on, we manage only services with no urlCmd.
  // a non empty urlCmd means the url resource was not found.
  if urlCmd > '' then
    begin
    getPage('not found', data);
    exit;
    end;

  if f.isFolder() and not ansiEndsStr('/',url) then
    begin
    conn.reply.mode:=HRM_MOVED;
    conn.reply.url:=url+'/';
    exit;
    end;

  case conn.request.method of
    HM_GET, HM_POST:
      begin
      conn.reply.mode:=HRM_REPLY;
      lastActivityTime:=now();
      end;
    HM_HEAD: conn.reply.mode:=HRM_REPLY_HEADER;
    end;

  data.lastFile:=f;

  if f.isFolder() then
    begin
    data.downloadingWhat:=DW_FOLDERPAGE;
    conn.reply.contentType:='text/html';
    conn.reply.bodyMode:=RBM_STRING;
    if DMbrowserTplChk.Checked and isDownloadManagerBrowser() then
      conn.reply.body:=getFolderPage(f, conn, dmBrowserTpl)
    else conn.reply.body:=getFolderPage(f, conn, tpl);
    compressReply(data);
    exit;
    end;

  httpDate:=dateToHTTP(getMtimeUTC(f.resource));

  if httpDate = conn.getHeader('If-Modified-Since') then
    begin
    conn.reply.mode:=HRM_NOT_MODIFIED;
    exit;
    end;

  data.countAsDownload:=f.countAsDownload();

  if data.countAsDownload and limitsExceededOnDownload() then
    begin
    data.countAsDownload:=FALSE;
    getPage('max contemp downloads', data);
    exit;
    end;

  setupDownloadIcon(data);
  data.eta.idx:=0;
  conn.reply.contentType:=mimetypeByFile(f);
  conn.reply.bodyMode:=RBM_FILE;
  conn.reply.body:=f.resource;
  data.downloadingWhat:=DW_FILE;
  data.fileXferStart:=now();
  if data.countAsDownload and (flashOn = 'download') then flash();

  s:=f.name;
  b:=(openInBrowser <> '') and fileMatch(openInBrowser, s);
  // N-th workaround for IE. The 'accept' check should let us know if the save-dialog is displayed. More information at www.rejetto.com/forum/?topic=6275
  if (data.agent = 'MSIE') and (conn.getHeader('Accept') = '*/*') then
    s:=stringReplace(s, ' ','%20', [rfReplaceAll]);
  conn.addHeader( 'Content-Disposition: '+ifThen(not b, 'attachment; ')+'filename="'+s+'";' );

  conn.addHeader( 'Last-Modified: '+httpDate );
  end; // handleRequest

  procedure lastByte();

    procedure incDLcount(f:Tfile; res:string);
    begin
    if (f = NIL) or f.isTemp() then autoupdatedFiles.incInt(res)
    else f.DLcount:=1+f.DLcount
    end;

  var
    archive: TarchiveStream;
    i: integer;
  begin
  if data.countAsDownload then
    inc(downloadsLogged);
  // workaround for a bug that was fixed in Wget/1.10
  if stringExists(data.agent, ['Wget/1.8.2', 'Wget/1.9', 'Wget/1.9.1']) then
    begin
    data.disconnectReason:='wget bug workaround (consider updating)';
    conn.disconnect();
    end;
  VFScounterMod:=TRUE;
  case data.downloadingWhat of
    DW_FILE:
      if assigned(data) then
        incDLcount(data.lastFile, data.lastFile.resource);
    DW_ARCHIVE:
      begin
      archive:=conn.reply.bodyStream as TarchiveStream;
      for i:=0 to length(archive.flist)-1 do
        incDLcount(Tfile(archive.flist[i].data), archive.flist[i].src);
      end;
    end;
  end; // lastByte

var
  s: string;
begin
f:=NIL;
data:=NIL;
if assigned(conn) then data:=conn.data;
if assigned(data) then data.lastActivityTime:=now();

// this situation can happen when there is a call to processMessage() before this function ends
if (data = NIL) and (event in [HE_REQUESTED, HE_GOT]) then
  exit;

case event of
  HE_CANT_OPEN_FILE: data.error:='Can''t open file';
  HE_OPEN:
    begin
    startBtn.Hide();
    updateUrlBox();
    // this happens when the server is switched on programmatically
    usingFreePort:= port='';
    updatePortBtn();
    end;
  HE_CLOSE:
    begin
    startBtn.show();
    updatePortBtn();
    updateUrlBox();
    end;
  HE_REQUESTED:
    begin
    handleRequest();
    // we save the value because we need it also in HE_REPLY
    // and temp files are not avaliable in HE_REPLY
    data.dontLog:=assigned(f) and f.hasRecursive(FA_DONT_LOG);
    end;
  HE_REPLIED:
    begin
    setupDownloadIcon(data); // remove the icon
    data.lastBytesGot:=0;
    if data.disconnectAfterReply then conn.disconnect();
    end;
  HE_LAST_BYTE_DONE:
    if (conn.reply.mode = HRM_REPLY) and (data.downloadingWhat in [DW_FILE, DW_ARCHIVE]) then
      lastByte();
  HE_CONNECTED:
    begin
    data:=TconnData.create(conn);
    data.downloadingWhat:=DW_ERROR;
    if limitsExceededOnConnection() then data.preReply:=PR_OVERLOAD;
    if isBanned(conn) then
      begin
      data.disconnectReason:='banned';
      data.preReply:=PR_BAN;
      if noReplyBan then conn.reply.mode:=HRM_CLOSE;
      end;
    if flashOn = 'connection' then flash();
    connBox.Items.AddObject('', pointer(data));
    if (data.preReply = PR_NONE)
    and addUniqueString(conn.address, ipsEverConnected)
    and ipsEverFrm.Visible then
      ipsEverFrm.refreshData();
    end;
  HE_DISCONNECTED:
    begin
    closeUploadingFile_partial();
    if leavedisconnectedconnectionsChk.checked then
      begin
      if assigned(data) then
        data.averageSpeed:=safeDiv(data.conn.bytesSent,SECONDS*(now()-data.time));
      setupDownloadIcon(data); // remove the tray icon anyway
      end
    else
      begin
      data.deleting:=TRUE;
      toDelete.add(data);
      end;
    refreshConn(conn);
    end;
  HE_GOT: lastActivityTime:=now();
  HE_SENT:
    begin
    if data.nextDloadScreenUpdate <= now() then
      begin
      data.nextDloadScreenUpdate:= now()+DOWNLOAD_MIN_REFRESH_TIME;
      refreshConn(conn);
      setupDownloadIcon(data);
      end;
    lastActivityTime:=now();
    end;
  HE_POST_FILE:
    begin
    data.agent:=getAgentID(conn);
    data.fileXferStart:=now();
    f:=findFileByURL(decodeURL(conn.request.url));
    data.uploadingFN:=optAnsi(tpl.utf8, conn.post.filename);
    data.uploadFailed:='';
    if (f = NIL) or not uploadAllowed(conn, f) or not f.accessFor(conn) then
      begin
      data.uploadFailed:=ifThen(f=NIL, 'Folder not found', 'Not allowed');
      logUploadFailed();
      end
    else
      begin
      closeUploadingFile();
      new(data.f);
      assignFile(data.f^, getUploadDestinationFileName() );
      if canWriteFile() then
        begin
        rewrite(data.f^, 1);
        saveFile(data.f^, conn.post.data);
        end;
      repaintTray();
      end;
    end;
  HE_POST_MORE_FILE:
    if canWriteFile() then
      saveFile(data.f^, conn.post.data);
  HE_POST_END_FILE:
    begin
    if data.f = NIL then s:=''
    else s:=IntToStr(filesize(data.f^));
    with data do
      uploadResults:=uploadResults+extractFilename(getFilename(f^))+'|'+s+'|'+uploadFailed+'|'+
        intToStr(averageUpSpeedFile())+CRLF;
    data.uploadingFN:='';
    if assigned(data.f) then inc(uploadsLogged);
    closeUploadingFile();
    end;
  // default case
  else refreshConn(conn);
  end;//case

if event in [HE_CONNECTED, HE_DISCONNECTED, HE_OPEN, HE_CLOSE, HE_REQUESTED,
  HE_POST_END, HE_LAST_BYTE_DONE] then
  begin
  repaintTray();
  updateTrayTip();
  end;
doLog();
end; // httpEvent

procedure findSimilarIP(fromIP:string);

  function howManySameNumbers(ip1,ip2:string):integer;
  var
    n1, n2: string;
  begin
  result:=0;
  while ip1 > '' do
    begin
    n1:=chop('.',ip1);
    n2:=chop('.',ip2);
    if n1 <> n2 then exit;
    inc(result);
    end;
  end; // howManySameNumbers

var
  s, l, chosen: string;
begin
if fromIP = '' then exit;
if stringExists(fromIP, customIPs) then
  begin
  setDefaultIP(fromIP);
  exit;
  end;
chosen:=getIP();
s:=getIPs();
while s >'' do
  begin
  l:=chopLine(s);
  if howManySameNumbers(chosen, fromIP) < howManySameNumbers(l, fromIP) then
    chosen:=l;
  end;
setDefaultIP(chosen);
end; // findSimilarIP

procedure setLimitOption(var variable:integer; newValue:integer;
  menuItem:TmenuItem; menuLabel:string);
begin
if newValue < 0 then newValue:=0;
variable:=newValue;
menuItem.caption:=format(menuLabel,
  [ifThen(newValue=0,'disabled',intToStr(newValue))]);
end; // setLimitOption

procedure setMaxIPs(v:integer);
begin
setLimitOption(maxIPs,v, mainfrm.maxIPs1,
  'Max simultaneous addresses: %s ...')
end;

procedure setMaxIPsDLing(v:integer);
begin
setLimitOption(maxIPsDLing,v, mainfrm.maxIPsDLing1,
  'Max simultaneous addresses downloading: %s ...')
end;

procedure setMaxConnections(v:integer);
begin
setLimitOption(maxConnections,v, mainfrm.maxConnections1,
  'Max connections: %s ...')
end;

procedure setMaxConnectionsIP(v:integer);
begin
setLimitOption(maxConnectionsIP, v, mainfrm.MaxconnectionsfromSingleaddress1,
  'Max connections from single address: %s ...')
end;

procedure setMaxDLs(v:integer);
begin
setLimitOption(maxContempDLs, v, mainfrm.maxDLs1,
  'Max simultaneous downloads: %s ...')
end;

procedure setMaxDLsIP(v:integer);
begin
setLimitOption(maxContempDLsIP, v, mainfrm.maxDLsIP1,
  'Max simultaneous downloads from single address: %s ...')
end;

procedure setAutoFingerprint(v:integer);
begin
autoFingerprint:=v;
mainfrm.Createfingerprintonaddition1.caption:='Create fingerprint on addition'
  +ifThen(v=0, ': disabled', format(' under %d KB',[v]) );
end;

procedure applyISOdateFormat();
begin
if mainfrm.useISOdateChk.checked then ShortDateFormat:='yyyy-mm-dd'
else ShortDateFormat:=GetLocaleStr(LOCALE_USER_DEFAULT, LOCALE_SSHORTDATE,'');
end;

function loadFingerprint(fn:string):string;
var
  hasher: Thasher;
begin
result:=loadMD5for(fn);
if result > '' then exit;

hasher:=Thasher.Create;
hasher.loadFrom(ExtractFilePath(fn));
result:=hasher.getHashFor(fn);
hasher.Free;
end; // loadFingerprint

function createFingerprint(fn:string):string;
var
  fs: Tfilestream;
  digest: TMD5Digest;
  context: TMD5Context;
  buf: array [1..32*1024] of byte;
  i: integer;
begin
result:='';
fs:=TfileStream.create(fn, fmOpenRead+fmShareDenyWrite);
for i:=0 to 15 do byte(digest[i]):=succ(i);
MD5init(context);
try
  repeat
  i:=fs.Read(buf, sizeof(buf));
  MD5updateBuffer(context, @buf, i);
  if not progFrm.visible then continue;
  progFrm.progress:=safeDiv(0.0+fs.position, fs.size);
  application.processMessages();
  if progFrm.cancelRequested then exit;
  until i < sizeof(buf);
finally
  fs.free;
  MD5final(digest, context);
  for i:=0 to 15 do
    result:=result+intToHex(byte(digest[i]), 2);
  end;
end; // createFingerprint

procedure TmainFrm.FormResize(Sender: TObject);
begin
urlBox.Width:=toolbar2.ClientWidth-browseBtn.Width;
setLastCol();
toolbar3.Repaint();
end;

procedure checkIfOnlyCountersChanged();
begin if not VFSmodified and VFScounterMod then mainfrm.saveVFS(lastFileOpen) end;

procedure checkVfsOnQuit();
begin
if loadingVFS.disableAutosave then exit;
checkIfOnlyCountersChanged();
if not VFSmodified then exit;
if mainfrm.autosaveVFSchk.checked
or (msgDlg('Save file system?', MB_ICONQUESTION+MB_YESNO) = IDYES) then
  mainfrm.saveVFS(lastFileOpen);
end; // checkVfsOnQuit

procedure inputComment(f:Tfile);
const
  MSG= 'Please insert a comment for "%s".'
    +#13'You should use HTML: <br> for break line.';
begin
VFSmodified:=inputqueryLong('Comment', format(MSG, [f.name]), f.comment);
end; // inputComment

function Tmainfrm.addFile(f:Tfile; parent:Ttreenode=NIL;
  skipComment:boolean=FALSE):Tfile;
begin
abortBtn.show();
stopAddingItems:=FALSE;
try result:=addFileRecur(f,parent);
finally abortBtn.hide();
  end;
if stopAddingItems then
  msgDlg('File addition was aborted.'#13'The list of files is incomplete.', MB_ICONWARNING);
if assigned(parent) then parent.expanded:=TRUE;
filesbox.Selected:=result.node;

if skipComment or not autoCommentChk.checked then exit;
application.restore();
application.bringToFront();
inputComment(f);
end; // addFile

function Tmainfrm.addFileRecur(f:Tfile; parent:Ttreenode=NIL):Tfile;

  function sameDrive(f1,f2:string):boolean;
  begin
  result:=(length(f1) >= 2) and (length(f2) >= 2) and (f1[2] = ':')
    and (f2[2] = ':') and (upcase(f1[1]) = upcase(f2[1]));
  end; // sameDrive

var
  n: Ttreenode;
  sr: TsearchRec;
  newF: Tfile;
  s: string;
begin
result:=f;
if stopAddingItems then exit;

if parent = NIL then parent:=rootNode;

if addingItemsCounter >= 0 then // counter enabled
  begin
  inc(addingItemsCounter);
  if addingItemsCounter and 7 = 0 then // step 8
    begin
    application.ProcessMessages();
    setStatusBarText(format('Adding item #%d', [addingItemsCounter]));
    end;
  end;

// ensure the parent is a folder
while assigned(parent) and assigned(parent.data)
and not Tfile(parent.data).isFolder() do
  parent:=parent.parent;
// test for duplicate. it often happens when you have a shortcut to a file.
if existsNodeWithName(f.name, parent) then exit;

if stopAddingItems then exit;
// in this case, drive letter may change. useful with pen drives.
if runningOnRemovable and sameDrive(exePath, f.resource) then
  delete(f.resource, 1,2);

n:=filesBox.Items.AddChild(parent, f.name);
// stateIndex assignments are a workaround to a delphi bug
n.StateIndex:=0;
f.node:=n;
n.StateIndex:=-1;
n.Data:=f;
f.setupImage();
// autocreate fingerprint
if f.isFile() and fingerprintsChk.checked and (autoFingerprint > 0) then
  try
    f.size:=sizeofFile(f.resource);
    if (autoFingerprint >= f.size div 1024)
    and (loadFingerprint(f.resource) = '') then
      begin
      s:=createFingerprint(f.resource);
      if s > '' then saveFile(f.resource+'.md5', s);
      end;
  except
  end;
  
if (f.resource = '') or not f.isVirtualFolder() then exit;
// virtual folders must be run at addition-time
if findFirst(f.resource+'\*',faAnyfile, sr) <> 0 then exit;
  repeat
  if stopAddingItems then break;
  if (sr.name[1] = '.')
  or isFingerprintFile(sr.name) or isCommentFile(sr.name) then continue;
  newF:=Tfile.create(f.resource+'\'+sr.name);
  if newF.isFolder() then include(newF.flags, FA_VIRTUAL);
  addfileRecur(newF, n);
  until findnext(sr) <> 0;
FindClose(sr);
end; // addFileRecur

procedure TmainFrm.filesBoxCollapsing(Sender: TObject; Node: TTreeNode;
  var AllowCollapse: Boolean);
begin
AllowCollapse:=node.parent<>NIL;
end;

procedure TmainFrm.Newlink1Click(Sender: TObject);
var
  name: string;
begin
name:=getUniqueNodeName('New link', filesBox.Selected);
addfile(Tfile.createLink(name), filesBox.Selected).node.Selected:=TRUE;
setURL1click(sender);
end;

procedure TmainFrm.newfolder1Click(Sender: TObject);
var
  name: string;
begin
name:=getUniqueNodeName('New folder', filesBox.selected);
addFile(Tfile.createVirtualFolder(name), filesBox.Selected).node.Selected:=TRUE;
end;

procedure TmainFrm.filesBoxEditing(Sender: TObject; Node: TTreeNode;
  var AllowEdit: Boolean);
begin
if node = NIL then exit;
{ disable shortcuts, to be used in editbox. Shortcuts need to be re-activated,
{ but when the node text is left unchanged, no event is notified, so we got to
{ use timerEvent to do the work. }
remove1.ShortCut:=0;
Paste1.ShortCut:=0;

allowedit:=allowedit and not Tfile(node.data).isRoot()
end;

procedure TmainFrm.filesBoxEdited(Sender:TObject; Node:TTreeNode; var S:String);
var
  f: Tfile;
begin
f:=node.data;
if f.name = s then exit;
if existsNodeWithName(s, node.parent)
and (msgDlg(MSG_SAME_NAME, MB_ICONWARNING+MB_YESNO) <> IDYES) then exit;
if f.isFile() and not validFilename(s) or (s = '') or  (pos('/',s) > 0) then
  begin
  s:=node.text;
  msgDlg('Invalid filename', MB_ICONERROR);
  exit;
  end;
f.name:=s;
VFSmodified:=TRUE;
end;

procedure Tmainfrm.remove(node:Ttreenode=NIL);
var
  i, n: integer;
  list: array of Ttreenode;
  warn: boolean;
begin
if assigned(node) then
  begin
  if node.parent = NIL then exit;
  if nodeIsLocked(node) then
    begin
    msgDlg(MSG_ITEM_LOCKED, MB_ICONERROR);
    exit;
    end;
  node.Delete();
  exit;
  end;

n:=filesbox.SelectionCount;
if (n = 0) or (n = 1) and selectedFile.isRoot() then exit;
if not deleteDontAskChk.checked
and (msgDlg('Delete?', MB_ICONQUESTION+MB_YESNO) = IDNO) then
  exit;
// we have to save the selection, be  cause deletion modifies it
setlength(list,n);
for i:=0 to n-1 do list[i]:=filesbox.selections[i];
// now proceed
warn:=FALSE;
for i:=0 to n-1 do
  if assigned(list[i].data) and assigned(list[i].parent) then
    if nodeIsLocked(list[i]) then warn:=TRUE
    else list[i].Delete();

if warn then
  msgDlg(MSG_SOME_LOCKED, MB_ICONWARNING);
end; // remove

procedure TmainFrm.Remove1Click(Sender: TObject);
begin
// this method is bound to the DEL key also while a renaming is ongoing
if not filesBox.IsEditing() then remove()
end;

procedure TmainFrm.startBtnClick(Sender: TObject);
begin toggleServer() end;

function Tmainfrm.pointedConnection():TconnData;
var
  i: integer;
begin
result:=NIL;
i:=connbox.ItemAtPos(connbox.ScreenToClient(mouse.CursorPos), TRUE);
if i < 0 then exit;
result:=TconnData(connbox.Items.Objects[i]);
end; // pointedConnection

function Tmainfrm.pointedFile(strict:boolean=TRUE):Tfile;
var
  n: Ttreenode;
  p: Tpoint;
begin
result:=NIL;
GetCursorPos(p);
p:=filesbox.ScreenToClient(p);
if strict and not (htOnItem in filesBox.GetHitTestInfoAt(p.x,p.y)) then exit;
n:=filesbox.GetNodeAt(p.x,p.y);
if (n = NIL) or (n.data = NIL) then exit;
result:=n.data;
end; // pointedFile

procedure Tmainfrm.updateUrlBox();
var
  f: Tfile;
begin
if quitting then exit;
if selectedFile = NIL then f:=rootFile
else f:=selectedFile;

if f = NIL then urlBox.Text:=''
else urlBox.text:=f.fullURL()
end; // updateUrlBox

procedure TmainFrm.filesBoxChange(Sender: TObject; Node: TTreeNode);
begin
if filesBox.SelectionCount = 0 then selectedFile:=NIL
else selectedFile:=filesBox.selections[0].data;
updateUrlBox()
end;

function Tmainfrm.selectedConnection():ThttpConn;
var
  i: integer;
begin
result:=NIL;
i:=connBox.ItemIndex;
if i < 0 then exit;
result:=TconnData(connBox.items.objects[i]).conn;
end; // selectedConnection

procedure TmainFrm.Kickconnection1Click(Sender: TObject);
var
  conn: ThttpConn;
begin
conn:=selectedConnection();
if assigned(conn) then conn.disconnect;
end;

procedure TmainFrm.Kickallconnections1Click(Sender: TObject);
begin kickByIP('*') end;

procedure TmainFrm.KickIPaddress1Click(Sender: TObject);
var
  conn: ThttpConn;
begin
conn:=selectedConnection();
kickByIP(conn.address);
end;

procedure setAutosave(var rec:Tautosave; v:integer);
begin
rec.every:=v;
if assigned(rec.menu) then
  rec.menu.caption:=
    'Auto save every: '+ifThen(v=0,'disabled',intToStr(v)+' seconds');
end; // setAutosave

procedure autosaveClick(var rec:Tautosave; name:string);
const
  MSG = 'Auto-save %s.'
    +#13'Specify in seconds.'
    +#13'Leave blank to disable.';
  MSG_MIN = 'We don''t accept less than %d';
var
  s: string;
  v: integer;
begin
if rec.every <= 0 then s:=''
else s:=intToStr(rec.every);
  repeat
  if not inputquery('Auto-save '+name, format(MSG,[name]), s) then exit;
  s:=trim(s);
  if s = '' then
    begin
    setAutosave(rec, 0);
    break;
    end
  else
    try
      v:=strToInt(s);
      if v >= rec.minimum then
        begin
        setAutosave(rec, v);
        break;
        end;
      msgDlg(format(MSG_MIN,[rec.minimum]), MB_ICONERROR);
    except msgDlg(MSG_INVALID_VALUE, MB_ICONERROR) end;
  until false;
end; // autosaveClick

function TmainFrm.getCfg():string;

  function accountsToStr():string;
  var
    i: integer;
    a: Paccount;
  begin
  result:='';
  for i:=0 to length(accounts)-1 do
  	begin
    a:=@accounts[i];
    result:=result+'login:'+base64encode(a.user+':'+a.pwd)
      +'|enabled:'+yesno[a.enabled]+'|no-limits:'+yesno[a.noLimits]+';';
    end;
  end; // accountsToStr

  function banlistToStr():string;
  var
    i: integer;
  begin
  result:='';
  for i:=0 to length(banlist)-1 do
    result:=result+banlist[i].ip+'#'
      +AnsiReplaceStr(banlist[i].comment,'|','\$pipe')+'|';
  end;

  function connColumnsToStr():string;
  var
    i: integer;
  begin
  result:='';
  for i:=0 to connheader.Sections.Count-1 do
    with connheader.Sections[i] do
      result:=result+format('%s;%d|', [text, width]);
  end; // connColumnsToStr

var
  iconMasksStr, userIconMasks: string;

  function iconMasksToStr():string;
  var
    i, j: integer;
    s: string;
  begin
  result:='';
  for i:=0 to length(iconMasks)-1 do
    begin
    j:=iconMasks[i].int;
    if j >= startingImagesCount then
      begin
      s:=base64encode(zCompressStr(pic2str(j), zcMax));
      j:=USER_ICON_MASKS_OFS+j-startingImagesCount;
      userIconMasks:=userIconMasks+intToStr(j)+':'+s+'|';
      end;
    result:=result+format('%s|%d||', [iconMasks[i].str, j]);
    end;
  end; // iconMasksToStr

  function fontToStr(f:Tfont):string;
  begin
  result:=ifThen(fsBold in f.Style, 'B')+ifThen(fsItalic in f.Style, 'I')
    +ifThen(fsUnderline in f.Style, 'U')+ifThen(fsStrikeOut in f.Style, 'S');
  result:=format('%s|%d|%s|%s', [f.Name,f.size,colorToString(f.Color),result]);
  end; // fontToStr

begin
userIconMasks:='';
iconMasksStr:=iconMasksToStr();
result:='HFS '+VERSION+' - Build #'+VERSION_BUILD+CRLF
+'active='+yesno[srv.active]+CRLF
+'window='+rectToStr(lastWindowRect)+CRLF
+'window-max='+yesno[windowState = wsMaximized]+CRLF
+'easy='+yesno[easyMode]+CRLF
+'port='+port+CRLF
+'tpl-font='+fontToStr(tplFrm.box.font)+CRLF
+'log-read-only='+yesno[logbox.readonly]+CRLF
+'log-width='+intToStr(lastGoodLogWidth)+CRLF
+'log-file-name='+logFile.filename+CRLF
+'log-font-name='+logFontName+CRLF
+'log-font-size='+intToStr(logFontSize)+CRLF
+'log-date='+yesno[LogdateChk.checked]+CRLF
+'log-time='+yesno[LogtimeChk.checked]+CRLF
+'log-only-served='+yesno[logOnlyServedChk.checked]+CRLF
+'log-server-start='+yesno[logServerstartChk.checked]+CRLF
+'log-server-stop='+yesno[logServerstopChk.checked]+CRLF
+'log-connections='+yesno[logConnectionsChk.checked]+CRLF
+'log-disconnections='+yesno[logDisconnectionsChk.checked]+CRLF
+'log-bytes-sent='+yesno[logBytessentChk.checked]+CRLF
+'log-bytes-received='+yesno[logBytesreceivedChk.checked]+CRLF
+'log-replies='+yesno[logRepliesChk.checked]+CRLF
+'log-requests='+yesno[logRequestsChk.checked]+CRLF
+'log-full-downloads='+yesno[logFulldownloadsChk.checked]+CRLF
+'log-dump-request='+yesno[DumprequestsChk.checked]+CRLF
+'log-browsing='+yesno[logBrowsingChk.checked]+CRLF
+'log-icons='+yesno[logIconsChk.checked]+CRLF
+'log-progress='+yesno[logProgressChk.checked]+CRLF
+'log-banned='+yesno[logBannedChk.checked]+CRLF
+'log-others='+yesno[logOtherEventsChk.checked]+CRLF
+'log-apache-format='+logfile.apacheFormat+CRLF
+'delete-dont-ask='+yesno[deleteDontAskChk.checked]+CRLF
+'confirm-exit='+yesno[confirmexitChk.checked]+CRLF
+'keep-bak-updating='+yesno[keepBakUpdatingChk.checked]+CRLF
+'include-pwd-in-pages='+yesno[pwdInPagesChk.Checked]+CRLF
+'ip='+defaultIP+CRLF
+'custom-ip='+join(';',customIPs)+CRLF
+'listen-on='+listenOn+CRLF
+'dynamic-dns-updater='+base64encode(dyndns.url)+CRLF
+'dynamic-dns-user='+dyndns.user+CRLF
+'dynamic-dns-host='+dyndns.host+CRLF
+'search-better-ip='+yesno[searchbetteripChk.checked]+CRLF
+'start-minimized='+yesno[startMinimizedChk.checked]+CRLF
+'connections-height='+intToStr(lastGoodConnHeight)+CRLF
+'files-stay-flagged-for-minutes='+intToStr(filesStayFlaggedForMinutes)+CRLF
+'auto-save-vfs='+yesno[autosaveVFSchk.checked]+CRLF
+'folders-before='+yesno[foldersbeforeChk.checked]+CRLF
+'use-comment-as-realm='+yesno[usecommentasrealmChk.checked]+CRLF
+'getright-template='+yesno[DMbrowserTplChk.checked]+CRLF
+'save-options-on-exit='+yesno[saveoptionsonexitChk.checked]+CRLF
+'dont-include-port-in-url='+yesno[noPortInUrlChk.checked]+CRLF
+'persistent-connections='+yesno[persistentconnectionsChk.checked]+CRLF
+'modal-options='+yesno[modalOptionsChk.checked]+CRLF
+'beep-on-flash='+yesno[beepChk.checked]+CRLF
+'prevent-leeching='+yesno[preventLeechingChk.checked]+CRLF
+'delete-partial-uploads='+yesno[deletePartialUploadsChk.checked]+CRLF
+'rename-partial-uploads='+renamePartialUploads+CRLF
+'use-system-icons='+yesno[usesystemiconsChk.checked]+CRLF
+'minimize-to-tray='+yesno[MinimizetotrayChk.checked]+CRLF
+'tray-icon-for-each-download='+yesno[trayfordownloadChk.checked]+CRLF
+'show-main-tray-icon='+yesno[showmaintrayiconChk.checked]+CRLF
+'always-on-top='+yesno[alwaysontopChk.checked]+CRLF
+'support-descript.ion='+yesno[supportDescriptionChk.checked]+CRLF
+'oem-descript.ion='+yesno[oemForIonChk.checked]+CRLF
+'enable-fingerprints='+yesno[fingerprintsChk.checked]+CRLF
+'save-fingerprints='+yesno[saveNewFingerprintsChk.checked]+CRLF
+'auto-fingerprint='+intToStr(autoFingerprint)+CRLF
+'encode-pwd-url='+yesno[encodePwdUrlChk.checked]+CRLF
+'tpl-wrap='+yesno[tplFrm.wrapChk.checked]+CRLF
+'recursive-listing='+yesno[recursiveListingChk.checked]+CRLF
+'send-hfs-identifier='+yesno[sendHFSidentifierChk.checked]+CRLF
+'list-hidden-files='+yesno[listfileswithhiddenattributeChk.checked]+CRLF
+'list-system-files='+yesno[listfileswithsystemattributeChk.checked]+CRLF
+'list-protected-items='+yesno[hideProtectedItemsChk.checked]+CRLF
+'enable-no-default='+yesno[enableNoDefaultChk.checked]+CRLF
+'browse-localhost='+yesno[browseUsingLocalhostChk.checked]+CRLF
+'add-folder-default='+addFolderDefault+CRLF
+'default-sorting='+defSorting+CRLF
+'auto-save-vfs-every='+intToStr(autosaveVFS.every)+CRLF
+'auto-save-cfg-every='+intToStr(autosaveCFG.every)+CRLF
+'last-update-check='+floatToStr(lastUpdateCheck)+CRLF
+'allowed-referer='+allowedReferer+CRLF
+'tray-shows='+trayShows+CRLF
+'tray-message='+x_template(trayMsg,[#13,'\n',#10,''])+CRLF
+'speed-limit='+floatToStr(speedLimit)+CRLF
+'max-ips='+intToStr(maxIPs)+CRLF
+'max-ips-downloading='+intToStr(maxIPsDLing)+CRLF
+'max-connections='+intToStr(maxConnections)+CRLF
+'max-connections-by-ip='+intToStr(maxConnectionsIP)+CRLF
+'max-contemporary-dls='+intToStr(maxContempDLs)+CRLF
+'max-contemporary-dls-ip='+intToStr(maxContempDLsIP)+CRLF
+'login-realm='+loginRealm+CRLF
+'open-in-browser='+openInBrowser+CRLF
+'flash-on='+flashOn+CRLF
+'graph-rate='+intToStr(graph.rate)+CRLF
+'graph-size='+intToStr(graph.size)+CRLF
+'graph-visible='+yesno[graphBox.visible]+CRLF
+'no-download-timeout='+intToStr(noDownloadTimeout)+CRLF
+'connections-timeout='+intToStr(connectionsInactivityTimeout)+CRLF
+'no-reply-ban='+yesno[noReplyBan]+CRLF
+'ban-list='+banlistToStr()+CRLF
+'last-file-open='+lastFileOpen+CRLF
+'reload-on-startup='+yesno[reloadonstartupChk.checked]+CRLF
+'find-external-on-startup='+yesno[findExtOnStartupChk.checked]+CRLF
+'encode-non-ascii='+yesno[encodenonasciiChk.checked]+CRLF
+'encode-spaces='+yesno[encodespacesChk.checked]+CRLF
+'only-1-instance='+yesno[only1instanceChk.checked]+CRLF
+'mime-types='+join('|',MIMEtypes)+CRLF
+'icon-masks='+iconMasksStr+CRLF
+'icon-masks-user-images='+userIconMasks+CRLF
+'address2name='+join('|',address2name)+CRLF
+'recent-files='+join('|',recentFiles)+CRLF
+'ips-ever-connected='+join(';', ipsEverConnected)+CRLF
+'leave-disconnected-connections='
  +yesno[leavedisconnectedconnectionsChk.checked]+CRLF
+'accounts='+accountsToStr()+CRLF
+'tray-instead-of-quit='+yesno[trayInsteadOfQuitChk.checked]+CRLF
+'compressed-browsing='+yesno[compressedbrowsingChk.checked]+CRLF
+'use-iso-date-format='+yesno[useISOdateChk.Checked]+CRLF
+'hints4newcomers='+yesno[HintsfornewcomersChk.checked]+CRLF
+'save-totals='+yesno[saveTotalsChk.checked]+CRLF
+'number-files-on-upload='+yesno[numberFilesOnUploadChk.checked]+CRLF
+'do-not-log-address='+dontLogAddressMask+CRLF
+'last-external-address='+dyndns.lastIP+CRLF
+'min-disk-space='+intToStr(minDiskSpace)+CRLF
+'out-total='+intToStr(outTotalOfs+srv.bytesSent)+CRLF
+'in-total='+intToStr(inTotalOfs+srv.bytesReceived)+CRLF
+'hits-total='+intToStr(hitsLogged)+CRLF
+'downloads-total='+intToStr(downloadsLogged)+CRLF
+'upload-total='+intToStr(uploadsLogged)+CRLF
+'many-items-warning='+yesno[warnManyItems]+CRLF
+'load-single-comment-files='+yesno[loadSingleCommentsChk.checked]+CRLF
+'copy-url-on-start='+yesno[autocopyURLonstartChk.checked]+CRLF
+'connections-columns='+connColumnsToStr()+CRLF
+'auto-comment='+yesno[autoCommentChk.checked]+CRLF
+'update-daily='+yesno[updateDailyChk.checked]+CRLF
+'tester-updates='+yesno[testerUpdatesChk.checked]+CRLF
+'copy-url-on-addition='+yesno[AutocopyURLonadditionChk.checked]+CRLF
end; // getCfg

function TmainFrm.setcfg(cfg:string; alreadyStarted:boolean=FALSE):boolean;
const
  MSG_BAN = 'Your ban configuration may have been screwed up.'
    +#13'Please verify it.';
var
  l, savedip, build: string;
  warnings: TStringDynArray;
  userIconOfs: integer;

  function yes(s:string=''):boolean;
  begin result:= ifThen(s>'',s,l)='yes' end;

  function int():int64;
  begin if not tryStrToInt64(l, result) then result:=0 end;

  function real():double;
  begin if not tryStrToFloat(l, result) then result:=0 end;

  procedure loadBanlist(s:string);
  var
    p: string;
    i: integer;
  begin
  { old versions wrongly used ; as ban-record separator, while it was already
  { used as address separator }
  if (build < '018') and (pos(';',s) > 0) then
    begin
    s:=AnsiReplaceStr(s,';','|');
    addString(MSG_BAN, warnings);
    end;
  setLength(banlist, 0);
  i:=0;
  while s > '' do
    begin
    p:=chop('|',s);
    if p = '' then continue;
    setLength(banlist, i+1);
    banlist[i].comment:=AnsiReplaceStr(p,'\$pipe','|'); // unescape
    banlist[i].ip:=chop('#',banlist[i].comment);
    inc(i);
    end;
  end; // loadBanlist

  procedure strToAccounts();
  var
  	s, t, p: string;
    i: integer;
    a: Paccount;
  begin
  accounts:=NIL;
  while l > '' do
  	begin
    // accounts are separated by semicolons
    s:=chop(';',l);
    if s = '' then continue;
    i:=length(accounts);
    setLength(accounts, i+1);
    a:=@accounts[i];
    while s > '' do
      begin
      // account properties are separated by pipes
      t:=chop('|',s);
      p:=chop(':',t); // get property name
      if p = '' then continue;
      if p = 'login' then
      	begin
	      t:=base64decode(t);
  	    a.user:=chop(':',t);
	      a.pwd:=t;
        end;
      if p = 'enabled' then a.enabled:=yes(t);
      if p = 'no-limits' then a.noLimits:=yes(t);
      end;
    end;
  end; // strToAccounts

  procedure strToConnSections();
  var
    s, labl: string;
    hs: THeaderSection;
    i, old: integer;
  begin
  while l > '' do
    with connheader do
      begin
      s:=chop('|',l);
      if s = '' then continue;
      labl:=chop(';',s);
      old:=-1;
      for i:=0 to sections.Count-1 do
        if sections[i].text = labl then
          old:=i;
      if old < 0 then continue;
      hs:=sections.add();
      hs.assign(sections[old]);
      hs.width:=StrToIntDef(s, width);
      sections.delete(old);
      end;
  end; // strToConnSections

  procedure strToIconmasks();
  var
    i: integer;
  begin
  while l > '' do
    begin
    i:=length(iconMasks);
    setLength(iconMasks, i+1);
    iconMasks[i].str:=chop('|',l);
    iconMasks[i].int:=StrToIntDef(chop('||',l),0);
    end;
  end; // strToIconmasks

  procedure readUserIconMasks();
  var
    i, iFrom, iTo: integer;
  begin
  userIconOfs:=images.Count;
  while l > '' do
    begin
    iFrom:=strTointDef(chop(':', l), -1);
    iTo:=str2pic(ZDecompressStr(base64decode(chop('|', l))));
    for i:=0 to length(iconMasks)-1 do
      if iconMasks[i].int = iFrom then
        iconMasks[i].int:=iTo;
    end;
  end; // readUserIconmasks

  procedure strToFont(f:Tfont);
  begin
  f.Name:=chop('|', l);
  f.Size:=strToIntDef(chop('|', l), f.size);
  f.Color:=StringToColor(chop('|',l));
  f.Style:=[];
  if pos('B', l) > 0 then f.Style:=f.Style+[fsBold];
  if pos('U', l) > 0 then f.Style:=f.Style+[fsUnderline];
  if pos('I', l) > 0 then f.Style:=f.Style+[fsItalic];
  if pos('S', l) > 0 then f.Style:=f.Style+[fsStrikeout];
  end; // strToFont

var
  i: integer;
  h, connColumns: string;
  activateServer: boolean;
begin
result:=FALSE;
if cfg = '' then exit;
if chop(' ',cfg) <> 'HFS' then exit;
l:=chop(CRLF,cfg);
chop(' - Build #',l);
build:=l;

warnings:=NIL;
if alreadyStarted then activateServer:=srv.active
else activateServer:=TRUE;

while cfg > '' do
  begin
  l:=chop(CRLF,cfg);
  h:=chop('=',l);
  try
    if h = 'active' then activateServer:=yes;
    if h = 'window' then
    	begin
      lastWindowRect:=strToRect(l);
      left:=lastWindowRect.Left;
      top:=lastWindowRect.top;
      height:=lastWindowRect.Bottom-lastWindowRect.top;
      width:=lastWindowRect.right-lastWindowRect.left;
      end;
    if h = 'window-max' then if yes then windowstate:=wsMaximized;
    if h = 'port' then port:=l;
    if h = 'ip' then savedip:=l;
    if h = 'custom-ip' then customIPs:=split(';',l);
    if h = 'listen-on' then listenOn:=l;
    if h = 'dynamic-dns-updater' then dyndns.url:=base64decode(l);
    if h = 'dynamic-dns-user' then dyndns.user:=l;
    if h = 'dynamic-dns-host' then dyndns.host:=l;
    if h = 'login-realm' then loginRealm:=l;
    if h = 'easy' then setEasyMode(yes);
    if h = 'keep-bak-updating' then keepBakUpdatingChk.checked:=yes;
		if h = 'encode-non-ascii' then encodenonasciiChk.checked:=yes;
		if h = 'encode-spaces' then encodespacesChk.checked:=yes;
		if h = 'search-better-ip' then searchbetteripChk.checked:=yes;
    if h = 'start-minimized' then startMinimizedChk.checked:=yes;
    if h = 'tpl-font' then strToFont(tplFrm.box.Font);
    if h = 'log-width' then lastGoodLogWidth:=int;
    if h = 'log-file-name' then logFile.filename:=l;
    if h = 'log-font-name' then logFontName:=l;
    if h = 'log-font-size' then logFontSize:=int;
    if h = 'log-date' then LogdateChk.checked:=yes;
    if h = 'log-time' then LogtimeChk.checked:=yes;
    if h = 'log-read-only' then logbox.readonly:=yes;
    if h = 'log-browsing' then logBrowsingChk.checked:=yes;
    if h = 'log-icons' then logIconsChk.checked:=yes;
    if h = 'log-progress' then logProgressChk.checked:=yes;
    if h = 'log-banned' then logBannedChk.checked:=yes;
    if h = 'log-others' then logOtherEventsChk.checked:=yes;
    if h = 'log-dump-request' then DumprequestsChk.checked:=yes;
    if h = 'log-server-start' then logServerstartChk.checked:=yes;
    if h = 'log-server-stop' then logServerstopChk.checked:=yes;
    if h = 'log-connections' then logConnectionsChk.checked:=yes;
    if h = 'log-disconnections' then logDisconnectionsChk.checked:=yes;
    if h = 'log-bytes-sent' then logBytessentChk.checked:=yes;
    if h = 'log-bytes-received' then logBytesreceivedChk.checked:=yes;
    if h = 'log-replies' then logRepliesChk.checked:=yes;
    if h = 'log-requests' then logRequestsChk.checked:=yes;
    if h = 'log-full-downloads' then logFulldownloadsChk.checked:=yes;
    if h = 'log-apache-format' then logfile.apacheFormat:=l;
    if h = 'log-only-served' then logOnlyServedChk.checked:=yes;
    if h = 'confirm-exit' then confirmexitChk.checked:=yes;
    if h = 'connections-height' then lastGoodConnHeight:=int;
    if h = 'files-stay-flagged-for-minutes'then filesStayFlaggedForMinutes:=int;
    if h = 'folders-before' then foldersbeforeChk.checked:=yes;
    if h = 'include-pwd-in-pages' then pwdInPagesChk.Checked:=yes;
    if h = 'tpl-wrap' then tplFrm.wrapChk.checked:=yes;
    if h = 'minimize-to-tray' then MinimizetotrayChk.checked:=yes;
		if h = 'use-system-icons' then usesystemiconsChk.checked:=yes;
		if h = 'save-options-on-exit' then saveoptionsonexitChk.checked:=yes;
    if h = 'use-comment-as-realm' then usecommentasrealmChk.checked:=yes;
    if h = 'persistent-connections'then persistentconnectionsChk.checked:=yes;
		if h = 'show-main-tray-icon' then showmaintrayiconChk.checked:=yes;
    if h = 'delete-dont-ask' then deleteDontAskChk.checked:=yes;
	  if h = 'tray-icon-for-each-download' then trayfordownloadChk.checked:=yes;
    if h = 'copy-url-on-addition' then AutocopyURLonadditionChk.checked:=yes;
    if h = 'copy-url-on-start' then autocopyURLonstartChk.checked:=yes;
    if h = 'update-daily' then updateDailyChk.checked:=yes;
    if h = 'tray-instead-of-quit' then trayInsteadOfQuitChk.checked:=yes;
    if h = 'modal-options' then modalOptionsChk.checked:=yes;
    if h = 'beep-on-flash' then beepChk.checked:=yes;
    if h = 'prevent-leeching' then preventLeechingChk.checked:=yes;
    if h ='list-hidden-files' then listfileswithhiddenattributeChk.checked:=yes;
    if h ='list-system-files' then listfileswithsystemattributeChk.checked:=yes;
    if h = 'list-protected-items' then hideProtectedItemsChk.checked:=yes;
    if h = 'always-on-top' then alwaysontopChk.checked:=yes;
    if h = 'support-descript.ion' then supportDescriptionChk.Checked:=yes;
    if h = 'oem-descript.ion' then oemForIonChk.checked:=yes;
    if h = 'enable-fingerprints' then fingerprintsChk.checked:=yes;
    if h = 'save-fingerprints' then saveNewFingerprintsChk.checked:=yes;
    if h = 'auto-fingerprint' then setAutoFingerprint(int);
    if h = 'encode-pwd-url' then encodePwdUrlChk.checked:=yes;
    if h = 'last-update-check' then lastUpdateCheck:=strToFloat(l);
    if h = 'recursive-listing' then recursiveListingChk.checked:=yes;
    if h = 'enable-no-default' then enableNoDefaultChk.checked:=yes;
    if h = 'browse-localhost' then browseUsingLocalhostChk.checked:=yes;
    if h = 'add-folder-default' then addFolderDefault:=l;
    if h = 'default-sorting' then defSorting:=l;
    if h = 'send-hfs-identifier' then sendHFSidentifierChk.checked:=yes;
		if h = 'auto-save-vfs' then autosaveVFSchk.checked:=yes;
    if h = 'getright-template' then DMbrowserTplChk.checked:=yes;
    if h = 'leave-disconnected-connections' then
      leavedisconnectedconnectionsChk.checked:=yes;
		if h = 'speed-limit' then setSpeedLimit(real);
    if h = 'no-download-timeout' then setNoDownloadTimeout(int);
    if h = 'connections-timeout' then connectionsInactivityTimeout:=int;
    if h = 'max-ips' then setMaxIPs(int);
    if h = 'max-ips-downloading' then setMaxIPsDLing(int);
    if h = 'max-connections' then setMaxConnections(int);
    if h = 'max-connections-by-ip' then setMaxConnectionsIP(int);
    if h = 'max-contemporary-dls' then setMaxDLs(int);
    if h = 'max-contemporary-dls-ip' then setMaxDLsIP(int);
		if h = 'tray-message' then trayMsg:=AnsiReplaceStr(l,'\n',trayNL);
    if (h = 'ban-list') or (h = 'banned-ips') then loadBanlist(l);
    if h = 'no-reply-ban' then noReplyBan:=yes;
    if (h = 'save-in-out-totals')
    or (h = 'save-totals') then saveTotalsChk.checked:=yes;
    if h = 'allowed-referer' then allowedReferer:=l;
    if h = 'open-in-browser' then openInBrowser:=l;
		if h = 'last-file-open' then lastFileOpen:=l;
		if h = 'reload-on-startup' then reloadonstartupChk.checked:=yes;
    if h = 'find-external-on-startup' then findExtOnStartupChk.checked:=yes;
    if h = 'dont-include-port-in-url' then noPortInUrlChk.checked:=yes;
    if h = 'tray-shows' then trayshows:=l;
    if h = 'auto-save-cfg-every' then setAutosave(autosaveCFG, int);
    if h = 'auto-save-vfs-every' then setAutosave(autosaveVFS, int);
    if h = 'external-ip-server' then customIPservice:=l;
    if h = 'only-1-instance' then only1instanceChk.checked:=yes;
		if h = 'graph-rate' then setGraphRate(int);
		if h = 'graph-size' then graph.size:=int;
    if h = 'delete-partial-uploads' then deletePartialUploadsChk.checked:=yes;
    if h = 'rename-partial-uploads' then renamePartialUploads:=l;
    if h = 'do-not-log-address' then dontLogAddressMask:=l;
    if h = 'out-total' then outTotalOfs:=int;
    if h = 'in-total' then inTotalOfs:=int;
    if h = 'hits-total' then hitsLogged:=int;
    if h = 'downloads-total' then downloadsLogged:=int;
    if h = 'upload-total' then uploadsLogged:=int;
    if h = 'min-disk-space' then minDiskSpace:=int;
    if h = 'flash-on' then flashOn:=l;
    if h = 'last-external-address' then dyndns.lastIP:=l;
    if h = 'recents' then recentFiles:=split(';',l);   // legacy (the split symbol changed in #111)
    if h = 'recent-files' then recentFiles:=split('|',l);
    if h = 'ips-ever-connected' then ipsEverConnected:=split(';',l);
    if h = 'mime-types' then MIMEtypes:=split('|',l);
    if h = 'address2name' then address2name:=split('|',l);
    if h = 'compressed-browsing' then compressedbrowsingChk.checked:=yes;
    if h = 'hints4newcomers' then HintsfornewcomersChk.checked:=yes;
    if h = 'tester-updates' then testerUpdatesChk.checked:=yes;
    if h = 'number-files-on-upload' then numberFilesOnUploadChk.checked:=yes;
    if h = 'many-items-warning' then warnManyItems:=yes;
    if h = 'load-single-comment-files' then loadSingleCommentsChk.checked:=yes;
    if h = 'accounts' then strToAccounts();
    if h = 'use-iso-date-format' then useISOdateChk.Checked:=yes;
    if h = 'auto-comment' then autoCommentChk.checked:=yes;
    if h = 'icon-masks-user-images' then readUserIconMasks();
    if h = 'icon-masks' then strToIconmasks();
    if h = 'connections-columns' then strToConnSections();
		if h = 'graph-visible' then
      if yes then showGraph()
      else hideGraph();
    // extra commands for external use
    if h = 'load-tpl-from' then tpl.fullText:=loadFile(l);
    // backward compatibility
    for i:=0 to connheader.Sections.Count-1 do
      if h = 'section-'+connheader.sections[i].text+'-width' then
        connColumns:=connColumns+connheader.sections[i].text+';'+l+'|';
  except end;
  end;

if not alreadyStarted then
  // i was already seeing all the stuff, so please don't hide it
  if (build > '') and (build < '006') then easyMode:=FALSE;

l:=connColumns;
strToConnSections();

if not alreadyStarted then
  if not saveTotalsChk.checked then
    begin
    outTotalOfs:=0;
    inTotalOfs:=0;
    hitsLogged:=0;
    downloadsLogged:=0;
    uploadsLogged:=0;    
    end;
findSimilarIP(savedIP);
if lastGoodLogWidth > 0 then logBox.Width:=lastGoodLogWidth;
if lastGoodConnHeight > 0 then connPnl.Height:=lastGoodConnHeight;
srv.persistentConnections:=persistentconnectionsChk.Checked;
updateRecentFilesMenu();
updateAlwaysOnTop();
applyISOdateFormat();
for i:=0 to length(MIMEtypes)-1 do  // the filematch() would be fooled by spaces 
  MIMEtypes[i]:=trim(MIMEtypes[i]);
for i:=0 to length(warnings)-1 do
  msgDlg(warnings[i], MB_ICONWARNING);
if alreadyStarted then
  if activateServer <> srv.active then toggleServer()
  else
else
  if activateServer then startServer();
result:=TRUE;
end; // setcfg

function TmainFrm.loadcfg():boolean;
var
  tplS: string;
begin
tplS:='';
if FileExists(cfgPath+CFG_FILE) then
  begin
  result:=setcfg(loadFile(cfgPath+CFG_FILE));
  tplS:=loadFile(cfgPath+TPL_FILE);
  if result then saveMode:=SM_FILE;
  end
else
  begin
  result:=setcfg(loadregistry(CFG_KEY,''));
  tplS:=loadregistry(CFG_KEY,TPL_FILE);
  if result then saveMode:=SM_USER
  else
    begin
    result:=setcfg(loadregistry(CFG_KEY,'',HKEY_LOCAL_MACHINE));
    tplS:=loadregistry(CFG_KEY,TPL_FILE,HKEY_LOCAL_MACHINE);
	  if result then saveMode:=SM_SYSTEM;
    end;
  end;
if tplS > '' then tpl.fullText:=tplS;
if lastUpdateCheck = 0 then
  lastUpdateCheck:=getMtime(lastUpdateCheckFN);
end; // loadcfg

procedure TmainFrm.Viewhttprequest1Click(Sender: TObject);
begin
if selectedConnection()<>NIl then
  msgDlg(selectedConnection().request.full);
end;

procedure TmainFrm.connmenuPopup(Sender: TObject);
var
  bs, ba: boolean;
  i: integer;
  conn: ThttpConn;
begin
conn:=selectedConnection();
bs:=conn<>NIL;
ba:=FALSE;
for i:=0 to connBox.Count-1 do
  if TconnData(connBox.items.objects[i]).conn.state <> HCS_DISCONNECTED then
    begin
    ba:=TRUE;
    break;
    end;
Viewhttprequest1.enabled:=bs;
BanIPaddress1.enabled:=bs;
Kickconnection1.Enabled:=bs and (conn.state <> HCS_DISCONNECTED);
KickIPaddress1.Enabled:=bs and ba;
Kickallconnections1.Enabled:=ba;
Kickidleconnections1.Enabled:=ba;
pause1.Visible:=bs and isDownloading(conn.data);
Pause1.Checked:=bs and conn.paused;
end;

procedure TmainFrm.filemenuPopup(Sender: TObject);
const
  ONLY_ANY = 0;
  ONLY_EASY = 1;
  ONLY_EXPERT = 2;
var
  anyFileSelected, anyone: boolean;
  i: integer;
  mi: Tmenuitem;
  f: Tfile;
  a, all_accounts, all_uploadAccounts: TStringDynArray;

  function onlySatisfied(only:integer):boolean;
  begin
  result:=(only=ONLY_ANY)
    or (only=ONLY_EASY) and easyMode
    or (only=ONLY_EXPERT) and not easyMode
  end; // onlySatisfied

  procedure visibleAs(mi:Tmenuitem; other:Tmenuitem; only:integer=ONLY_ANY);
  begin mi.visible:=other.visible and onlySatisfied(only) end;

  procedure visibleIf(mi:Tmenuitem; should:boolean; only:integer=ONLY_ANY);
  begin if should then
    mi.visible:=TRUE and onlySatisfied(only) end;

  procedure checkedIf(mi:Tmenuitem; should:boolean);
  begin if should then mi.checked:=TRUE end;

  procedure enabledIf(mi:Tmenuitem; should:boolean);
  begin if should then mi.enabled:=TRUE end;

  procedure setDefaultValues(mi:TmenuItem);
  var
    i: integer;
  begin
  for i:=0 to mi.count-1 do
    begin
    mi[i].visible:=FALSE;
    mi[i].enabled:=TRUE;
    mi[i].checked:=FALSE;
    end;
  end; // setDefaultValues

  function itemsVisible(mi:TmenuItem):integer;
  var
    i: integer;
  begin
  result:=0;
  for i:=0 to mi.count-1 do
    if mi.Items[i].Visible then
      inc(result);
  end; // itemsVisible

begin
// default values
setDefaultValues(filemenu.items);
setDefaultValues(advanced1);
Upload1.Enabled:=FALSE;
Addfiles1.Visible:=TRUE;
Addfolder1.Visible:=TRUE;

anyFileSelected:=selectedFile<>NIL;
newfolder1.visible:=not anyFileSelected
  or ((filesBox.SelectionCount=1) and selectedFile.isFolder());
visibleIf(Comment1, anyFileSelected, ONLY_EXPERT);
Setuserpass1.visible:=anyFileSelected;
CopyURL1.visible:=anyFileSelected;

all_accounts:=NIL;
all_uploadAccounts:=NIL;
visibleIf(Customizedrealm2, anyFileSelected, ONLY_EXPERT);
visibleIf(Bindroottorealfolder1, (filesBox.SelectionCount=1)
  and selectedFile.isRoot() and selectedFile.isVirtualFolder(), ONLY_EXPERT);
visibleIf(Unbindroot1, (filesBox.SelectionCount=1)
  and selectedFile.isRoot() and selectedFile.isRealFolder(), ONLY_EXPERT);

for i:=0 to filesBox.SelectionCount-1 do
  begin
  f:=filesBox.selections[i].data;
  addArray(all_accounts, f.accounts);
  addArray(all_uploadAccounts, f.uploadAccounts);
  visibleIf(setURL1, FA_LINK in f.flags);
  visibleIf(Remove1, not f.isRoot());
  visibleIf(Letbrowse1, f.isFolder(), ONLY_EXPERT);
  visibleIf(upload1, f.isFolder());
  visibleIf(NoDownload1, f.isFile() or f.isRealFolder());

  enabledIf(upload1, f.isRealFolder());
  checkedIf(letbrowse1, FA_BROWSABLE in f.flags);
  checkedIf(NoDownload1, FA_DL_FORBIDDEN in f.flags);
  checkedIf(hide1, FA_HIDDEN in f.flags);
  checkedIf(Hidetree1, FA_HIDDENTREE in f.flags);
  checkedIf(Allowarchive1, FA_ARCHIVABLE in f.flags);
  checkedIf(Autohideemptyfolders1, FA_HIDE_EMPTY_FOLDERS in f.flags);
  visibleIf(Flagasnew1, not f.isNew() and (filesStayFlaggedForMinutes>0));
  visibleIf(Resetnewflag1, f.isNew() and (filesStayFlaggedForMinutes>0));
  visibleIf(SwitchToVirtual1, f.isRealFolder() and not f.isRoot(), ONLY_EXPERT);
  visibleIf(SwitchToRealfolder1, f.isVirtualFolder() and not f.isRoot() and (f.resource > ''), ONLY_EXPERT);
  visibleIf(Resetuserpass1, f.user>'');
  visibleIf(Dontconsiderasadownload1, f.isFile(), ONLY_EXPERT);
  checkedIf(Dontconsiderasadownload1, FA_DONT_COUNT_AS_DL in f.flags);
  visibleIf(Dontlog1, f.isFileOrFolder(), ONLY_EXPERT);
  checkedIf(Dontlog1, FA_DONT_LOG in f.flags);
  visibleIf(Hidefileextentioninlisting1, f.isFolder(), ONLY_EXPERT);
  checkedIf(Hidefileextentioninlisting1, FA_HIDE_EXT in f.flags);
  end;
visibleAs(CopyURLwithfingerprint1, Dontconsiderasadownload1);
visibleAs(Allowarchive1, Letbrowse1);
visibleAs(Difftemplate1, Letbrowse1);
visibleAs(Autohideemptyfolders1, Letbrowse1);
visibleAs(Hidetree1, Letbrowse1);
visibleAs(Filesfilter1, Letbrowse1);
visibleAs(Foldersfilter1, Letbrowse1);
visibleAs(Defaultfilemask1, Letbrowse1);
visibleAs(Dontconsiderasadownloadmask1, letbrowse1);
visibleAs(hide1, remove1);
visibleAs(Icon1, remove1, ONLY_EXPERT);
visibleAs(newlink1, newfolder1, ONLY_EXPERT);
Advanced1.visible:=not easyMode and (itemsVisible(Advanced1) > 0);

if filesBox.SelectionCount = 1 then
  begin
  f:=selectedFile;
  visibleIf(Editresource1, not (FA_VIRTUAL in f.flags), ONLY_EXPERT);
  visibleAs(rename1, remove1);
  visibleIf(openit1, not f.isVirtualFolder());
  visibleIf(browseIt1, TRUE, ONLY_EXPERT);
  paste1.visible:=clipboard.HasFormat(CF_HDROP);

  a:=NIL;
  if anyFileSelected then a:=selectedFile.accessList();
  visibleIf(CopyURLwithpassword1, assigned(a), ONLY_EXPERT);
  CopyURLwithpassword1.Clear();
  for i:=0 to length(a)-1 do
    CopyURLwithpassword1.Add(
      newItem( a[i], 0, FALSE, TRUE, copyURLwithPasswordMenuClick, 0, '')
    );
  end;

a:=getPossibleAddresses();
if length(a) = 1 then a:=NIL;
visibleIf(CopyURLwithdifferentaddress1, anyFileSelected and assigned(a), ONLY_EXPERT);
CopyURLwithdifferentaddress1.clear();
for i:=0 to length(a)-1 do
  CopyURLwithdifferentaddress1.add(
    newItem( a[i], 0, FALSE, TRUE, copyURLwithAddressMenuclick, 0, '')
  );

visibleIf(Restrictaccess1, anyFileSelected, ONLY_EXPERT);
if Restrictaccess1.visible then
  begin
  Restrictaccess1.Clear();
  if length(accounts) > 1 then
    begin
    Restrictaccess1.add( newItem( 'All existing accounts', 0, FALSE, TRUE,
      restrictaccessClick, RESTRICTMENU_ALL_ACCOUNTS, '') );
    Restrictaccess1.add( newItem( 'No account', 0, FALSE, TRUE,
      restrictaccessClick, RESTRICTMENU_NO_RESTRICTION, '') );
    end;
  Restrictaccess1.add( newItem( 'New account...', 0, FALSE, TRUE,
    restrictaccessClick, RESTRICTMENU_NEW_ACCOUNT, '') );

  Restrictaccess1.add( newItem('-',0,FALSE,TRUE,NIL,0,'') );
  Restrictaccess1.add( newItem(USER_ANONYMOUS,0,
    stringExists(USER_ANONYMOUS,all_accounts),TRUE,restrictaccessClick,
    RESTRICTMENU_ANONYMOUS,'') );

  for i:=0 to length(accounts)-1 do
  	with accounts[i] do
  		Restrictaccess1.add( newItem(user, 0, stringExists(user,all_accounts),
        enabled, restrictaccessClick, 0, '') )
  end;

Whyisuploaddisabled1.Visible:=upload1.Visible and not upload1.enabled;
if upload1.enabled then
  begin
  Uploadforaccounts1.clear();
  anyone:=stringExists('*',all_uploadAccounts);
  Uploadforaccounts1.add( newItem('Anyone', 0, anyone, TRUE,
    uploadmenuclick, UPLOADMENU_ANYONE, '') );
  Uploadforaccounts1.add( newItem( 'New account...', 0, FALSE, TRUE,
    uploadmenuclick, UPLOADMENU_NEW_ACCOUNT, '') );
  Uploadforaccounts1.add( newItem('-',0,FALSE,TRUE,NIL,0,'') );
  for i:=0 to length(accounts)-1 do
  	with accounts[i] do
  		Uploadforaccounts1.add( newItem(user, 0,
        not anyone and stringExists(user,all_uploadAccounts),
        not anyone and enabled, uploadmenuclick, 0, '') )
  end;

icon1.clear();
icon1.Add( newItem('Default',0,FALSE,TRUE,iconmenuclick,0,'') );
icon1.Add( newItem('New icon...',0,FALSE,TRUE,iconmenuclick,ICONMENU_NEW,'') );
icon1.Add( newItem('-',0,FALSE,TRUE,NIL,0,'') );
for i:=0 to images.Count-1 do
  begin
  mi:=newItem('',0,FALSE,TRUE,iconmenuclick,0,'');
  mi.ImageIndex:=i;
  icon1.Add(mi);
  end;
end;

function Tmainfrm.saveCFG():boolean;

  procedure proposeUserRegistry();
  const
    MSG = 'Can''t save options there.'
      +#13'Should I try to save to user registry?';
  begin
  if msgDlg(MSG, MB_ICONERROR+MB_YESNO) = IDYES then
    begin
    saveMode:=SM_USER;
    saveCFG();
    end;
  end; // proposeUserRegistry

begin
result:=FALSE;
if srv = NIL then exit;
case saveMode of
	SM_FILE:
  	begin
    if not savefile(cfgPath+CFG_FILE, getCfg()) then
      begin
      proposeUserRegistry();
      exit;
      end;
    result:=TRUE;
    if customizedTpl() then savefile(cfgPath+TPL_FILE, tpl.fullText)
    else deleteFile(cfgPath+TPL_FILE);
    end;
  SM_SYSTEM:
  	begin
    deleteFile(cfgPath+CFG_FILE);
    deleteFile(cfgPath+TPL_FILE);
    deleteRegistry(CFG_KEY);
    if not saveregistry( CFG_KEY, '', getCfg(), HKEY_LOCAL_MACHINE ) then
      begin
      proposeUserRegistry();
      exit;
      end;
    result:=TRUE;
    if customizedTpl() then
      saveregistry( CFG_KEY, TPL_FILE, tpl.fullText, HKEY_LOCAL_MACHINE )
    else deleteRegistry( CFG_KEY, TPL_FILE, HKEY_LOCAL_MACHINE );
    end;
  SM_USER:
  	begin
    deleteFile(cfgPath+CFG_FILE);
    deleteFile(cfgPath+TPL_FILE);
    result:=saveregistry( CFG_KEY, '', getCfg() );
    if customizedTpl() then saveRegistry( CFG_KEY, TPL_FILE, tpl.fullText )
    else deleteRegistry( CFG_KEY, TPL_FILE );
    end;
  end;
if result then
  deleteFile(lastUpdateCheckFN);
end; // saveCFG

procedure TmainFrm.toregistrycurrentuser1Click(Sender: TObject);
begin
saveMode:=SM_USER;
if saveCFG() then
  msgDlg(MSG_OPTIONS_SAVED);
end;

procedure TmainFrm.toregistryallusers1Click(Sender: TObject);
begin
saveMode:=SM_SYSTEM;
saveCFG();
msgDlg(MSG_OPTIONS_SAVED);
end;

procedure TmainFrm.tofile1Click(Sender: TObject);
begin
saveMode:=SM_FILE;
saveCFG();
msgDlg(MSG_OPTIONS_SAVED);
end;

procedure TmainFrm.About1Click(Sender: TObject);
begin msgDlg(format(getRes('copyright'), [VERSION,VERSION_BUILD])) end;

procedure Tmainfrm.purgeConnections();
var
  i: integer;
  data: TconnData;
begin
i:=0;
while i < toDelete.Count do
  begin
  data:=toDelete[i];
  inc(i);
  if data = NIL then continue;
  toDelete[i-1]:=NIL;
  { i had to move the deletion of connBox items from httpEvent to the timer
  { event because it has to be synchronous, because between the call of
  { indexOfObject() and delete() another disconnection could happen.
  { This would lead to invalidate the value returned by indexofObject() .
  { Maybe this problem could be solved with critical sections, i did not try }
  connBox.Items.delete(connBox.items.IndexOfObject(Tobject(data)));
  setupDownloadIcon(data);
  freeAndNIL(data.conn);
  try freeAndNIL(data) except end;
  end;
toDelete.clear();
end; // purgeConnections

procedure Tmainfrm.recalculateGraph();
var
  i: integer;
begin
if (srv = NIL) or quitting then exit;
// shift samples
i:=sizeOf(graph.samplesOut)-sizeOf(graph.samplesOut[0]);
move(graph.samplesOut[0], graph.samplesOut[1], i);
move(graph.samplesIn[0], graph.samplesIn[1], i);
// insert new "out" sample
graph.samplesOut[0]:=srv.bytesSent-graph.lastOut;
graph.lastOut:=srv.bytesSent;
// insert new "in" sample
graph.samplesIn[0]:=srv.bytesReceived-graph.lastIn;
graph.lastIn:=srv.bytesReceived;
// increase the max value
i:=max(graph.samplesOut[0], graph.samplesIn[0]);
if i > graph.maxV then
  begin
  graph.maxV:=i;
  graph.beforeRecalcMax:=100;
  end;
dec(graph.beforeRecalcMax);
if graph.beforeRecalcMax > 0 then exit;
// recalculate max value
graph.maxV:=0;
with graph do
  for i:=0 to length(samplesOut)-1 do
    maxV:=max(maxV, max(samplesOut[i], samplesIn[i]) );
graph.beforeRecalcMax:=100;
end; // recalculateGraph

// parse the version-dependant notice
procedure parseVersionNotice(s:string);
var
  l, msg: string;
begin
while s > '' do
  begin
  l:=trim(chopLine(s));
  // the line has to start with a @ followed by involved versions
  if (length(l) < 2) or (l[1] <> '@') then continue;
  delete(l,1,1);
  // collect the message (until next @-starting line)
  msg:='';
  while (s > '') and (s[1] <> '@') do
    msg:=msg+chopLine(s)+#13;
  // before 2.0 beta14 a bare semicolon-separated string comparison was used
  if filematch(l, VERSION) or filematch(l, '#'+VERSION_BUILD) then
    msgDlg(msg, MB_ICONWARNING);
  end;
end; // parseVersionNotice

procedure promptForUpdating(url:string);
const
  NEW_EXE = 'hfs.new.exe';
  MSG_UPDATE = 'You are invited to use the new version.'#13#13'Update now?';
  MSG_ERROR = 'Communication error';
  MSG_LIMITED = 'The auto-update feature cannot work because it requires the "Only 1 instance" option enabled.'
    +#13#13'Your browser will now be pointed to the update, so you can install it manually.';
  UPDATE_BATCH_FILE = 'hfs.update.bat';
  UPDATE_BATCH = 'START %0:s /WAIT "%1:s" -q'+CRLF
    +'ping 127.0.0.1 -n 3 -w 1000> nul'+CRLF
    +'DEL "%3:shfs.old.exe"'+CRLF
    +'%2:sMOVE "%1:s" "%3:shfs.old.exe"'+CRLF
    +'DEL "%1:s"'+CRLF
    +'MOVE "%4:s" "%1:s"'+CRLF
    +'START %0:s "%1:s"'+CRLF
    +'DEL %%0'+CRLF;
var
  b: boolean;
  size: integer;
begin
if (url = '') or (msgDlg(MSG_UPDATE, MB_YESNO) = IDNO) then exit;
if not monoLib.working then
  begin
  msgDlg(MSG_LIMITED, MB_ICONWARNING);
  exec(url);
  exit;
  end;

checkVfsOnQuit();
progFrm.show('Downloading...', TRUE);
try
  size:=sizeOfFile(NEW_EXE);
  // a previous failed update attempt? avoid re-downloading if not necessary
  if (size > 0) and (httpFileSize(url) = size) then b:=TRUE
  else b:=httpGetFile(url, NEW_EXE, mainfrm.progFrmHttpGetUpdate)
finally progFrm.hide() end;
if progFrm.cancelRequested then
  begin
  deleteFile(NEW_EXE);
  exit;
  end;
if not b then
  begin
  msgDlg(MSG_ERROR, MB_ICONERROR);
  exit;
  end;

try
  progFrm.show('Processing...');
  saveFile(UPDATE_BATCH_FILE, format(UPDATE_BATCH, [
    ifThen(isNT(), '""'),
    paramStr(0),
    ifThen(not mainfrm.keepBakUpdatingChk.checked,'REM '),
    exePath,
    NEW_EXE
  ]));
  execNew(UPDATE_BATCH_FILE);
finally
  progFrm.hide()
  end;
end; // promptForUpdate

function downloadUpdateInfo():Ttpl;
const
  URL = 'http://www.rejetto.com/hfs/hfs.updateinfo.txt';
  ON_DISK = 'hfs.updateinfo.txt';
  MSG_FROMDISK = 'Update info has been read from local file.'
    +#13'To resume normal operation of the updater, delete the file '
      +ON_DISK+' from the HFS program folder.';
var
  s: string;
begin
lastUpdateCheck:=now();
saveFile(lastUpdateCheckFN, '');
fileSetAttr(lastUpdateCheckFN, faHidden);

result:=NIL;
progFrm.show('Requesting...');
try
  // this let the developer to be not connected when testing the parsing
  if not fileExists(ON_DISK) then s:=httpGet(URL)
  else
    begin
    s:=loadFile(ON_DISK);
    msgDlg(MSG_FROMDISK, MB_ICONWARNING);
    end;
finally
  progFrm.hide()
  end;
if pos('[EOF]', s) = 0 then exit;
result:=Ttpl.Create;
result.fullText:=s;
end; // downloadUpdateInfo

procedure Tmainfrm.autoCheckUpdates();
var
  info: Ttpl;
  updateURL, ver, build: string;

  function thereSnew(kind:string):boolean;
  begin
  result:=trim(info['last '+kind+' build']) > VERSION_BUILD;
  if not result then exit;
  updateURL:=trim(info['last '+kind+' url']);
  ver:=trim(info['last '+kind]);
  build:=trim(info['last '+kind+' build']);
  end;

begin
if (VERSION_STABLE and (now()-lastUpdateCheck < 1))
or (now()-lastUpdateCheck < 1/3) then exit;
setStatusBarText('Checking for updates', 5);
info:=downloadUpdateInfo();
if info = NIL then
  begin
  if logOtherEventsChk.checked then add2log('Check update: failed');
  setStatusBarText('Update check failed', 5);
  exit;
  end;
if not thereSnew('stable')
and (not VERSION_STABLE or testerUpdatesChk.checked) then
  thereSnew('untested');
if ver = VERSION then ver:=format('Build #%s (current is #%s)', [build, VERSION_BUILD]);
info.Free;
if logOtherEventsChk.checked then
  if updateURL = '' then add2log('Check update: no new version')
  else add2log('Check update: new version found: '+ver);
promptForUpdating(updateURL);
setStatusBarText('');
end; // autoCheckUpdates

procedure TmainFrm.timerEvent(Sender: TObject);

  procedure calculateETA(data:TconnData; current:real; leftOver:int64);
  var
    i: integer;
  begin
  data.eta.data[data.eta.idx mod ETA_FRAME]:=current;
  inc(data.eta.idx);
  // calculate average
  if data.eta.idx < ETA_FRAME then data.eta.result:=0
  else
    begin
    data.averageSpeed:=0;
    for i:=0 to ETA_FRAME-1 do
      data.averageSpeed:=data.averageSpeed+data.eta.data[i];
    data.averageSpeed:=data.averageSpeed/ETA_FRAME;
    if data.averageSpeed > 0 then
      data.eta.result:=(leftOver/data.averageSpeed)/SECONDS;
    end;
  end; // calculateETA

  procedure everyMinute();
  begin
  if updateDailyChk.Checked and (now()-programStart > 60/SECONDS) then
    autoCheckUpdates();
  if dyndns.url > '' then getExternalAddress(externalIP);
  // purge icons older than 5 minutes, because sometimes icons change
  iconsCache.purge(now()-(5*60)/SECONDS);
  end; // everyMinute

  procedure every10sec();
  begin
  if not stringExists(defaultIP, getPossibleAddresses()) then
    // previous address not available anymore (it happens using dial-up)
    findSimilarIP(defaultIP);
  end; // every10sec

  procedure everySec();
  var
    i, outside: integer;
    data: TconnData;
    _now: Tdatetime;
  begin
  _now:=now();

  // check if the window is outside the visible screen area
  outside:=left;
  for i:=0 to monitor.MonitorNum do
    dec(outside, screen.monitors[i].width);
  if outside > 0 then
    MakeFullyVisible();

  if dyndns.active and (dyndns.url > '') then
    begin
    if externalIP = '' then getExternalAddress(externalIP);
    if not isLocalIP(externalIP) and (externalIP <> dyndns.lastIP) then
      updateDynDNS();
    end;

  updateTrayTip();

  setLastCol();  // sometimes vertical scrollbar appears and we must resize

  if warnManyItems and (filesBox.items.count > MANY_ITEMS_THRESHOLD) then
    begin
    warnManyItems:=FALSE;
    msgDlg(MSG_MANY_ITEMS, MB_ICONWARNING);
    end;

  with autosaveCFG do
    if (every > 0) and ((_now-last)*SECONDS >= every) then
      begin
      last:=_now;
      saveCFG();
      end;
  with autosaveVFS do // we do it only if the filename is already specified
    if (every > 0) and (lastFileOpen > '') and not loadingVFS.disableAutosave
    and ((_now-last)*SECONDS >= every) then
      begin
      last:=_now;
      saveVFS(lastFileOpen);
      end;

  i:=0;
  if assigned(srv) and assigned(srv.conns) then
    while i < srv.conns.count do
      begin
      data:=ThttpConn(srv.conns[i]).data;
      if data = NIL then continue;

      if isReceivingFile(data) then
        calculateETA(data, data.conn.speedIn, data.conn.bytesToPost);
      if isSendingFile(data) then
        calculateETA(data, data.conn.speedOut, data.conn.bytesToSend);
      // connection inactivity timeout
      if (connectionsInactivityTimeout > 0)
      and ((_now-data.lastActivityTime)*SECONDS >= connectionsInactivityTimeout)
      then
        begin
        data.disconnectReason:='inactivity';
        data.conn.disconnect();
        end;
      inc(i);
      end;

  // we got a better ip address
  if searchbetteripChk.checked
  and isLocalIP(defaultIP) and not isLocalIP(getIP()) then
    setDefaultIP(getIP());

  // server inactivity timeout
  if nodownloadtimeout > 0 then
    if (_now-lastActivityTime)*SECONDS > noDownloadTimeout*60 then
      quitASAP:=TRUE;

  if windowState = wsNormal then
    lastWindowRect:=mainfrm.boundsRect;

  { connBox should be repainted upon request by http events. but sometimes
  { it happens to be not updated anymore. this little hack should stay here
  { until we fix the bug. }
  connBox.Repaint();
  end; // everySec

  procedure everyTenth();
  var
    node: Ttreenode;
  begin
  purgeConnections();
  // see the filesBoxEditing event for an explanation of the following lines
  if not filesBox.IsEditing and (remove1.ShortCut = 0) then
    begin
    remove1.ShortCut:=TextToShortCut('Del');
    Paste1.ShortCut:=TextToShortCut('Ctrl+V');
    end;
  with optionsFrm do
    if active and iconsPage.Visible then
      updateIconMap();
  if scrollFilesBox in [SB_LINEUP,SB_LINEDOWN] then
    postMessage(filesBox.Handle, WM_VSCROLL, scrollFilesBox, 0);

  if assigned(filesToAddQ) then
    begin
    node:=filesBox.selected;
    while assigned(filesToAddQ) do
      addFilesFromString(popString(filesToAddQ), node);
    filesBox.selected:=node;
    end;

  if tplFrm.active then
    tplFrm.updateSbar();
  end; // everyTenth

  procedure quit();
  var
    bak: boolean;
  begin
  { close is not effective when lockTimerevent is TRUE, so we force it TRUE.
  { it should not be necessary, but we want to be sure to quit even with bugs. }
  bak:=lockTimerevent;
  lockTimerevent:=FALSE;
  application.MainForm.Close();
  lockTimerevent:=bak;
  end; // quit

begin
if not quitting and quitASAP then quit();
if not timer.enabled or quitting or lockTimerevent then exit;
lockTimerevent:=TRUE;
try
  everyTenth();
  inc(clock);
  if clock mod (60*10) = 0 then everyMinute();
  if clock mod 100 = 0 then every10sec();
  if clock mod 10 = 0 then everySec();
  if clock mod STATUSBAR_REFRESH = 0 then refreshSbar();
  if clock mod graph.rate = 0 then
    begin
    recalculateGraph();
    refreshGraphBox();
    end;
finally
  lockTimerevent:=FALSE
  end;
end; // timerEvent

procedure Tmainfrm.refreshSbar();
var
  pn: integer;

  function addPanel(s:string; al:TAlignment=taCenter):integer;
  begin
  result:=pn;
  inc(pn);
  if sbar.Panels.count < pn then sbar.Panels.Add();
  with sbar.panels[pn-1] do
    begin
    alignment:=al;
    width:=sbar.Canvas.TextWidth(s)+20;
    Text:=s;
    end;
  end; // addPanel

  procedure checkDiskSpace();
  type
    Tdrive = 1..26;
  var
    i: integer;
    drives: set of Tdrive;
    driveLetters: TStringDynArray;
    driveLetter: char;
  begin
  if minDiskSpace <= 0 then exit;
  drives:=[];
  i:=0;
  while i < length(uploadPaths) do
    begin
    include(drives, filenameToDriveByte(uploadPaths[i]));
    inc(i);
    end;
  driveLetters:=NIL;
  for i:=low(Tdrive) to high(Tdrive) do
    if i in drives then
      begin
      driveLetter:=chr(i+ord('A')-1);
      if not directoryExists(driveLetter+':\') then continue;
      if diskfree(i) div MEGA <= minDiskSpace then
        addString(driveLetter, driveLetters);
      end;
  if driveLetters = NIL then exit;
  sbarIdxs.oos:=addPanel( 'Out of space: '+join(',', driveLetters));
  end; // checkDiskSpace

  function getConnectionsString():string;
  var
    i: integer;
  begin
  result:=format('Connections: %d', [srv.conns.Count]);
  if easyMode then exit;
  i:=countIPs();
  if i < srv.conns.count then result:=result+' / '+intToStr(i);
  end;

var
  tempText: string;
begin
if quitting then exit;
fillChar(sbarIdxs, sizeof(sbarIdxs), -1);
if sbarTextTimeout < now() then tempText:=''
else tempText:=sbar.Panels[sbar.Panels.Count-1].text;
pn:=0;
if not easyMode then
  addPanel( getConnectionsString() );
sbarIdxs.out:=addPanel( format('Out: %.1f KB/s',[srv.speedOut/1000]) );
addPanel( format('In: %.1f KB/s',[srv.speedIn/1000]) );
if not easyMode then
  begin
  sbarIdxs.totalOut:=addPanel( format('Total Out: %s',[
    smartSize(outTotalOfs+srv.bytesSent)]) );
  sbarIdxs.totalIn:=addPanel( format('Total In: %s',[
    smartSize(inTotalOfs+srv.bytesReceived)]) );
  sbarIdxs.notSaved:=addPanel( format('VFS: %d items',[filesBox.items.count-1])
    +ifThen(VFSmodified,' - not saved') );
  if not vFsmodified then sbarIdxs.notSaved:=-1;
  end;
checkDiskSpace();

if assigned(banlist) then
  sbarIdxs.banStatus:=addPanel(format('Ban rules: %d', [length(banlist)]));

if customizedTpl() then sbarIdxs.customTpl:=addPanel('Customized template');
// if tempText empty, ensures a final panel terminator
addPanel(tempText, taLeftJustify);
// delete excess panels
while sbar.Panels.count > pn do sbar.Panels.delete(pn);
end; // refreshSbar

procedure Tmainfrm.refreshIPlist();
CONST
  INDEX_FOR_URL = 2;
  INDEX_FOR_NIC = 1;
var
  a: TStringDynArray;
  i: integer;
begin
while IPaddress1.Items[INDEX_FOR_URL].Caption <> '-' do
  IPaddress1.delete(INDEX_FOR_URL);
// fill 'IP address' menu
a:=getPossibleAddresses();
for i:=0 to length(a)-1 do
  mainfrm.IPaddress1.Insert(INDEX_FOR_URL,
    newItem(a[i], 0, a[i]=defaultIP, TRUE, ipmenuclick, 0, '') );

// fill 'Accept connections on' menu
while Acceptconnectionson1.count > INDEX_FOR_NIC  do
  Acceptconnectionson1.delete(INDEX_FOR_NIC);
Anyaddress1.checked:= listenOn = '';
a:=listToArray(localIPlist);
addUniqueString('127.0.0.1', a);
for i:=0 to length(a)-1 do
  Acceptconnectionson1.Insert(INDEX_FOR_NIC,
    newItem( a[i], 0, a[i]=listenOn, TRUE, acceptOnMenuclick, 0, '') );
end; // refreshIPlist

procedure TmainFrm.filesBoxDblClick(Sender: TObject);
begin
if assigned(selectedFile) then setClip(selectedFile.fullURL());
updateUrlBox();
end;

function setBrowsable(f:Tfile; par:integer):boolean;
begin
if f.isFolder() then
  if boolean(par) then exclude(f.flags, FA_BROWSABLE)
  else include(f.flags, FA_BROWSABLE);
result:=TRUE;
end; // setBrowsable

function setDLforbidden(f:Tfile; par:integer):boolean;
begin
if boolean(par) then exclude(f.flags, FA_DL_FORBIDDEN)
else include(f.flags, FA_DL_FORBIDDEN);
result:=TRUE;
end; // setDLforbidden

procedure TmainFrm.Letbrowse1Click(Sender: TObject);
var
  i: integer;
begin
if selectedFile = NIL then exit;
for i:=0 to filesBox.SelectionCount-1 do
  with Tfile(filesBox.Selections[i].data) do
    if isFolder() then
      recursiveApply(setBrowsable, integer((sender as Tmenuitem).checked) );
VFSmodified:=TRUE;
end;

procedure fileMenuSetFlag(sender:Tobject; flagToSet:TfileAttribute;
  filter:TfilterMethod=NIL; negateFilter:boolean=FALSE);
var
  i: integer;
begin
if selectedFile = NIL then exit;
with mainFrm do
  for i:=0 to filesBox.SelectionCount-1 do
    with Tfile(filesBox.Selections[i].data) do
      if not assigned(filter)
      or (negateFilter xor filter(filesBox.Selections[i].data)) then
        if (sender as TmenuItem).checked then exclude(flags, flagToSet)
        else include(flags, flagToSet);
VFSmodified:=TRUE;
with mainFrm do
filesBox.Repaint();
end;

procedure TmainFrm.Hide1Click(Sender: TObject);
begin fileMenuSetFlag(sender, FA_HIDDEN, @Tfile.isRoot, TRUE) end;

procedure TmainFrm.Hidetree1Click(Sender: TObject);
begin fileMenuSetFlag(sender, FA_HIDDENTREE) end;

procedure TmainFrm.filesBoxMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
filesBox.Selected:=filesbox.GetNodeAt(x,y)
end;

procedure setFilesBoxExtras(v:boolean);
begin
{ let disable this silly feature for now
if winVersion <> WV_VISTA then exit;
with mainfrm.filesBox do
  begin
  if isEditing then exit;
  ShowButtons:=v;
  ShowLines:=v;
  end;}
end; // setFilesBoxExtras

procedure TmainFrm.filesBoxMouseEnter(Sender: TObject);
begin
with filesBox do setFilesBoxExtras(TRUE);
end;

procedure TmainFrm.filesBoxMouseLeave(Sender: TObject);
begin
with filesBox do setFilesBoxExtras(focused);
end;

procedure TmainFrm.filesBoxCompare(Sender: TObject; Node1,
  Node2: TTreeNode; Data: Integer; var Compare: Integer);
var
  f1, f2: Tfile;
begin
f1:=Tfile(node1.data);
f2:=Tfile(node2.data);
if (f1 = NIL) or (f2 = NIL) then exit;
if not foldersbeforeChk.checked or (f1.isFolder() = f2.isFolder()) then
  compare:=ansiCompareText(f1.name, f2.name)
else
  if f1.isFolder() then compare:=-1
  else compare:=+1;
end;

procedure TmainFrm.foldersbeforeChkClick(Sender: TObject);
begin rootNode.AlphaSort(TRUE) end;

procedure browse(url:string);
begin
if mainfrm.browseUsingLocalhostChk.Checked then
  begin
  chop('//',url);
  chop('/',url);
  url:='http://localhost:'+srv.port+'/'+url;
  end;
exec(url);
end; // browse

procedure TmainFrm.Browseit1Click(Sender: TObject);
begin
if selectedFile = NIL then exit;
if selectedFile.isLink() then exec(selectedfile.url())
else browse(selectedfile.fullurl())
end;

procedure TmainFrm.Openit1Click(Sender: TObject);
begin
if selectedFile = NIL then exit;
exec('"'+selectedfile.resource+'"')
end;

procedure Tmainfrm.ipmenuclick(sender:Tobject);
var
  ip: string;
begin
ip:=(sender as Tmenuitem).caption;
delete(ip, pos('&',ip), 1);
setDefaultIP(ip);
searchbetteripChk.checked:=FALSE;
end; // ipmenuclick

// returns the last file added
function Tmainfrm.addFilesFromString(files:string; under:Ttreenode=NIL):Tfile;
var
  folderKindFrm: TfolderKindFrm;

  function selectFolderKind():integer;
  begin
  application.restore();
  application.BringToFront();
  Application.CreateForm(TfolderKindFrm, folderKindFrm);
  result:=folderKindFrm.ShowModal();
  folderKindFrm.Free;
  end; // selectFolderKind

const
  MSG1 = '%d item(s) already exists:'#13'%s'#13#13'Continue?';
var
  first: boolean;
  f: Tfile;
  s, fn: string;
  doubles: TStringDynArray;
  res: integer;
  skipComment: boolean;
begin
result:=NIL;
if files = '' then exit;
// warn upon double filenames
doubles:=NIL;
s:=files;
while s > '' do
	begin
  fn:=chopLine(s);
	if isExtention(fn, '.lnk') then fn:=resolveLnk(fn);
  if (length(fn) = 3) and (fn[2] = ':') then fn:=fn[1]+fn[2] // unit root folder
	else fn:=ExtractFileName(fn);
  if existsNodeWithName(fn, under) then addString(fn, doubles);
  end;
if assigned(doubles) then
  begin
  filesBox.Repaint();
  if msgDlg(format(MSG1, [length(doubles), join(', ',doubles)]),
    MB_ICONWARNING+MB_YESNO) <> IDYES then exit;
  end;

f:=NIL;
first:=TRUE;
skipComment:=not singleLine(files);
	repeat
  fn:=chopLine(files);
  if fn = '' then continue;
  f:=Tfile.create(fn);
  if first and (files = '') and f.isFolder() then
    begin
    if addFolderDefault = 'virtual' then res:=mrYes
    else if addFolderDefault = 'real' then res:=mrNo
    else res:=selectFolderKind();

    if isAbortResult(res) then
      begin
      f.free;
      exit;
      end;
    if res = mrYes then include(f.flags, FA_VIRTUAL);
    end;

  addingItemsCounter:=0;
  f.lock();
  try
    f.name:=getUniqueNodeName(f.name, under);
    addFile(f, under, skipComment);
  finally
    f.unlock();
    addingItemsCounter:=-1;
    end;

  first:=FALSE;
  until files = '';
if assigned(f) and autocopyURLonadditionChk.checked then
  setClip(f.fullURL());
result:=f;
end; // addFilesFromString

procedure Tmainfrm.addDropFiles(hnd:Thandle; under:Ttreenode);
var
  i, n: integer;
  buffer: array [0..2000] of char;
  files: string;
begin
if hnd = 0 then exit;
GlobalLock(hnd);
n:=DragQueryFile(hnd,cardinal(-1),NIL,0);
files:='';
buffer:='';
for i:=0 to n-1 do
  begin
  DragQueryFile(hnd,i,@buffer,sizeof(buffer));
  files:=files+buffer+CRLF;
  end;
//DragFinish(hnd);  // this call seems to cause instability, don't know why
GlobalUnlock(hnd);

addFilesFromString(files, under);
end; // addDropFiles

procedure Tmainfrm.WMDropFiles(var msg:TWMDropFiles);
var
  p: Tpoint;
begin
GetCursorPos(p);
p:=filesbox.ScreenToClient(p);
addDropFiles(msg.Drop, filesbox.GetNodeAt(p.x,p.y));
inherited;
end; // WMDropFiles

procedure Tmainfrm.WMQueryEndSession(var msg:TWMQueryEndSession);
begin
windowsShuttingDown:=TRUE;
msg.Result:=1;
inherited;
end; // WMQueryEndSession

procedure Tmainfrm.WMEndSession(var msg:TWMEndSession);
begin
if msg.EndSession then
  begin
  windowsShuttingDown:=TRUE;
  close();
  end;
inherited;
end; // WMEndSession

procedure TmainFrm.WMNCLButtonDown(var msg:TWMNCLButtonDown);
begin
if (msg.hitTest = windows.HTCLOSE) and trayInsteadOfQuitChk.checked then
  begin
  msg.hitTest:=windows.HTCAPTION; // cancel closing
  minimizeToTray();
  end;
inherited;
end;

procedure drawCentered(cnv:Tcanvas; r:Trect; text:string);
begin
drawText(cnv.Handle, pchar(text), length(text), r,
  DT_CENTER+DT_NOPREFIX+DT_VCENTER+DT_END_ELLIPSIS);
end; // drawCentered

procedure TmainFrm.ToolBar3CustomDraw(Sender: TToolBar; const ARect: TRect;
  var DefaultDraw: Boolean);
begin
with filesBox.BoundsRect do
  drawCentered(sender.Canvas,
    rect(left,arect.top+2,right,arect.bottom),'Virtual File System');
with logBox.BoundsRect do
  drawCentered(sender.Canvas,
    rect(left,arect.top+2,right,arect.bottom),'Log');
end;

procedure TmainFrm.splitVMoved(Sender: TObject);
begin
if logBox.width > 0 then lastGoodLogWidth:=logBox.width;
toolbar3.Repaint;
end;

procedure TmainFrm.appEventsShowHint(var HintStr: String;
  var CanShow: Boolean; var HintInfo: THintInfo);

  function reduce(s:string):string;
  begin
  result:=x_template(s, [ #13,' ', #10,'' ]);
  if length(result) > 30 then
    begin
    setlength(result, 29);
    result:=result+'...';
    end;
  end; // reduce

  function fileHint():string;
  const
    INHERITED_LABEL = ' [inherited]';
    EXTERNAL_LABEL = ' [external]';
  var
    f, parent: Tfile;
    s, s2: string;
    inheritd, externl: boolean;

    function flag(lbl:string; att:TfileAttribute; positive:boolean=TRUE):string;
    begin result:=ifThen((att in f.flags) = positive, #13+lbl) end;

    function flagR(lbl:string; att:TfileAttribute; positive:boolean=TRUE):string;
    var
      inh: boolean;
    begin
    result:=ifThen(f.hasRecursive(att, @inh), #13+lbl);
    result:=result+ifThen(inh, INHERITED_LABEL);
    end; // flagR

  begin
  result:=ifThen(HintsfornewcomersChk.checked,'Drag your files here');
  f:=pointedFile();
  if f = NIL then exit;
  parent:=f.parent;

  result:='URL: '+f.url()
    +ifThen(f.isRealFolder() or f.isFile(), #13'Path: '+f.resource);
  if f.isFile() then
    result:=result+format(#13'Size: %s'#13'Downloads: %d',
      [ smartsize(sizeofFile(f.resource)), f.DLcount ]);

  s:=flag('Invisible', FA_HIDDENTREE, TRUE);
  if s = '' then s:=flag('Invisible', FA_HIDDEN);
  result:=result+s
    +flag('Download forbidden', FA_DL_FORBIDDEN)
    +flagR('Don''t log', FA_DONT_LOG);

  if f.isFolder() then
    begin
    if assigned(parent) and parent.hasRecursive(FA_HIDE_EMPTY_FOLDERS) then
        result:=result+#13'Hidden if empty'+INHERITED_LABEL;

    result:=result
      +flag('Not browsable', FA_BROWSABLE, FALSE)
      +flag('Hide empty folders', FA_HIDE_EMPTY_FOLDERS)
      +flagR('Hide extention', FA_HIDE_EXT)
      +flagR('Archivable', FA_ARCHIVABLE)
    end;

  s:=f.getRecursiveFileMask();
  if (s > '') and (f.defaultFileMask = '') then s:=s+INHERITED_LABEL;
  if s > '' then result:=result+#13'Default file mask: '+s;

  s:=join(', ', f.accessList(TRUE, @inheritd));
  if (s > '') and inheritd then s:=s+INHERITED_LABEL;
  if s > '' then result:=result+#13'Access for: '+s;

  if stringExists('*', f.uploadAccounts) then s:='anyone'
  else s:=join(', ', f.uploadAccounts);
  if s > '' then result:=result+#13'Upload for: '+s;

  s:=reduce(f.getDynamicComment());
  if (s > '') and (f.comment = '') then s:=s+EXTERNAL_LABEL;
  if s > '' then result:=result+#13'Comment: '+s;

  s:=reduce(f.getShownRealm());
  if (s > '') and (f.realm = '') then s:=s+INHERITED_LABEL;
  if s > '' then result:=result+#13'Realm: '+s;

  s:=reduce(f.getRecursiveDiffTplAsStr(@inheritd, @externl));
  if s > '' then
    begin
    if inheritd then s:=s+INHERITED_LABEL;
    if externl then s:=s+EXTERNAL_LABEL;
    result:=result+#13'Diff template: '+s;
    end;

  f.getFiltersRecursively(s, s2);
  result:=result
    +ifThen(s>'', #13'Files filter: '+s
      +ifThen(f.filesFilter = '', INHERITED_LABEL))
    +ifThen(s2>'', #13'Folders filter: '+s2
      +ifThen(f.foldersFilter = '', INHERITED_LABEL))
    +flag('Don''t consider as download', FA_DONT_COUNT_AS_DL)
    +ifThen(f.dontCountAsDownloadMask>'',
      #13'Don''t consider as download (mask): '+f.dontCountAsDownloadMask)
  end; // filehint

  function connhint():string;
  var
    data: TconnData;
  begin
  result:=ifThen(HintsForNewcomersChk.checked,
    'This box shows info about current connections');
  data:=pointedConnection();
  if data = NIL then exit;
  result:='Start time: '+DateTimeToStr(data.time)
    +#13'Agent: '+firstNonEmpty(data.agent,'<unknown>')
  end; // connhint

begin
canShow:=mainFrm.Active;
if hintinfo.HintControl = filesBox then hintStr:=filehint();
if hintinfo.HintControl = connBox then hintStr:=connhint();
hintinfo.ReshowTimeout:=500;
end;

procedure TmainFrm.logmenuPopup(Sender: TObject);
begin
Readonly1.Checked:=logBox.ReadOnly;
Banthisaddress1.Visible:= logBox.SelAttributes.Color=ADDRESS_COLOR;
Address2name1.Visible:=not easyMode;
Logfile1.Visible:=not easyMode;
logOnVideoChk.Visible:=not easyMode;
Donotlogaddress1.Visible:=not easyMode;
Apachelogfileformat1.visible:=not easyMode and (logfile.filename>'');
Clearandresettotals1.visible:=not easyMode;
Addresseseverconnected1.Visible:=not easyMode;
end;

procedure TmainFrm.Readonly1Click(Sender: TObject);
begin with logBox do ReadOnly:=not ReadOnly end;

procedure TmainFrm.Clear1Click(Sender: TObject);
begin logBox.Clear() end;

procedure TmainFrm.Clearandresettotals1Click(Sender: TObject);
begin
logBox.clear();
resetTotals();
end;

procedure TmainFrm.Copy1Click(Sender: TObject);
begin
if logBox.SelLength > 0 then setClip(logBox.SelText)
else setClip(logBox.Text)
end;

procedure TmainFrm.Saveas1Click(Sender: TObject);
var
  fn: string;
begin
fn:='';
if PromptForFileName(fn, 'Text file|*.txt', 'txt', 'Save log', '', TRUE) then
  savefile(fn, logBox.text);
end;

procedure TmainFrm.Save1Click(Sender: TObject);
begin savefile('hfs.log', logBox.text) end;

procedure deleteCFG();
begin
deleteFile(lastUpdateCheckFN);
deleteFile(CFG_FILE);
deleteRegistry(CFG_KEY);
deleteRegistry(CFG_KEY, HKEY_LOCAL_MACHINE);
end; // deleteCFG

procedure TmainFrm.Clearoptionsandquit1click(Sender: TObject);
begin
deleteCFG();
saveoptionsonexitChk.Checked:=FALSE;
close();
end;

procedure TmainFrm.connBoxDrawItem(Control: TWinControl; Index: Integer;
  R: TRect; State: TOwnerDrawState);
const
  HCS2STR :array [ThttpConnState] of string = ('idle','requesting',
    'receiving', 'replying', 'replying', 'sending', 'disconnected');
var
  cnv: Tcanvas;
  data: TconnData;
  incomingFile, outgoingFile: string;

  procedure textCenter(s:string);
  begin drawCentered(cnv,r,s) end;

  function getFname():string;
  begin result:=firstNonEmpty(outgoingFile, incomingFile, '-') end;

  function getStatus():string;
  begin
  if outgoingFile > '' then
    begin
    if data.conn.paused then result:='paused'
    else result:=format('%s / %s sent', [dotted(data.conn.bytesSentLastItem),
      dotted(data.conn.bytesPartial)]);
    exit;
    end;
  if incomingFile > '' then
    begin
    result:=format('%s / %s received', [dotted(data.conn.bytesPosted),
      dotted(data.conn.post.length)]);
    exit;
    end;
  result:=HCS2STR[data.conn.state]
    +ifThen(data.conn.state = HCS_IDLE, ' '+intToStr(data.conn.requestCount))
  end; // getStatus

  procedure drawProgress(now,total,lowerbound,upperbound:int64);
  var
    d: real;
    selected: boolean;
    r1: Trect;
    x: integer;
    colors:array [boolean] of Tcolor;
  begin
  if (total <= 0) or (lowerbound >= upperbound) then exit;
  colors[false]:=clWindow;
  colors[true]:=blend(clWindow, clWindowText, 0.25);
  selected:=odSelected in state;
  r1:=rect(r.Left+1,r.Top+1,r.Right-1,r.Bottom-1);
  // paint a shadow for non requested piece of data
  cnv.brush.Color:=blend(clWindow, clHotLight, 0.30);
  cnv.Brush.Style:=bsSolid;
  cnv.FillRect(r1);
  // and shrink the rectangle
  x:=r1.Right-r1.Left;
  cnv.pen.color:=colors[selected];
  cnv.pen.Style:=psSolid;
  if lowerbound > 0 then
    begin
    inc( r1.Left, round(x*lowerbound/total) );
    cnv.MoveTo(r1.Left-1, r1.Top);
    cnv.LineTo(r1.Left-1, r1.Bottom);
    end;
  if upperbound > 0 then dec( r1.Right, round(x*(total-upperbound)/total) );
  // border + non filled part
  cnv.brush.Color:=colors[not selected];
  cnv.Brush.Style:=bsSolid;
  cnv.FillRect(r1);
  // filled part
  d:=now / (upperbound-lowerbound+1);
  if d > 1 then d:=1;
  inc(r1.Left, 1+round(d*(r1.right-r1.Left-2)));
  dec(r1.right); dec(r1.bottom); inc(r1.top);
  cnv.brush.Color:=colors[selected];
  if not IsRectEmpty(r1) then cnv.FillRect(r1);
  // label
  cnv.Font.Name:='Small Fonts';
  cnv.font.Size:=7;
  cnv.font.Color:=clWindowText;
  SetBkMode(cnv.handle, TRANSPARENT);
  inc(r.top);
  textCenter(format('%d%%', [trunc(d*100)]));
  end; // drawProgress

  function getSpeed():string;
  var
    d: real;
  begin
  if isSendingFile(data) then d:=data.conn.speedOut/1000
  else if isReceivingFile(data) then d:=data.conn.speedIn/1000
  else d:=data.averageSpeed/1000;
  if d < 0.001 then textCenter('-')
  else textCenter(format('%.1f KB/s',[d]))
  end; // getSpeed

  function getIP():string;
  begin
  with data.conn do
    result:=nonEmptyConcat('', request.user, '@')+address+':'+port
  end; // getIP

var
  i: integer;
  l: string;
begin
if quitting then exit;
cnv:=connBox.canvas;
cnv.fillRect(r);
data:=TconnData(connBox.items.objects[index]);
if (data = NIL) or data.deleting then exit;

outgoingFile:=ifThen(isSendingFile(data), data.lastFN);
incomingFile:=optAnsi(tpl.utf8, ifThen(isReceivingFile(data), data.conn.post.filename));

for i:=0 to connheader.Sections.Count-1 do
  with connheader.Sections[i] do
    begin
    l:=text;
    r.left:=left;
    r.Right:=right;
    if l = 'IP' then textCenter(getIP());
    if l = 'Filename' then textCenter(getFname());
    if l = 'Status' then textCenter(getStatus());
    if l = 'Speed' then textCenter(getSpeed());
    if l = 'Time left' then textCenter(getETA(data));
    if l = '%' then
      if (outgoingFile > '') or (data.conn.reply.bodyMode = RBM_STREAM) then
        drawProgress( data.conn.bytesSentLastItem, data.conn.bytesFullBody,
          data.conn.reply.firstByte, data.conn.reply.lastByte )
      else
        if incomingFile > '' then
          drawProgress( data.conn.bytesPosted, data.conn.post.length, 0,
            data.conn.post.length);
    end;
end;

procedure TmainFrm.connheaderResize(Sender: TObject);
begin
setlastCol();
connBox.Repaint()
end;

procedure TmainFrm.connheaderSectionResize(HeaderControl: THeaderControl;
  Section: THeaderSection);
begin
setlastCol();
connBox.Repaint();
end;

procedure TmainFrm.connBoxMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
connBox.ItemIndex:=connBox.ItemAtPos(point(x,y), TRUE)
end;

procedure Tmainfrm.setlastCol();
var
  scroll: boolean;
  scrollW: integer;
begin
scroll:=connBox.Count > connBox.ClientHeight div connBox.ItemHeight;
scrollW:=GetSystemMetrics(SM_CXVSCROLL);
with connheader.Sections do
  Items[count-1].width:=connheader.Width-items[count-1].left+1
    -ifThen(scroll,scrollW);
end;

procedure TmainFrm.appEventsMinimize(Sender: TObject);
begin
if not MinimizetotrayChk.Checked then exit;
minimizeToTray();
end;

procedure TmainFrm.appEventsRestore(Sender: TObject);
begin
trayed:=FALSE;
if not showmaintrayiconChk.checked then tray.hide();
end;

procedure Tmainfrm.trayEvent(sender:Tobject; ev:TtrayEvent);
var
  p: Tpoint;
begin
updateTrayTip();
if userInteraction.disabled then exit;
case ev of
  TE_RCLICK:
    begin
    SetForegroundWindow(handle);
    getcursorpos(p);
    menu.Popup(p.x,p.y);
    end;
  TE_2CLICK:
    begin
    application.Restore();
    application.BringToFront();
    end;
  end;
end; procedure TmainFrm.trayicons1Click(Sender: TObject);
begin

end;

// trayEvent

procedure Tmainfrm.downloadtrayEvent(sender:Tobject; ev:TtrayEvent);
var
  p: Tpoint;
begin
if userInteraction.disabled then exit;
connBox.itemIndex:=connBox.Items.IndexOfObject((sender as TtrayIcon).data);
case ev of
  TE_RCLICK:
    begin
    getcursorpos(p);
    connmenu.Popup(p.x,p.y);
    end;
  TE_2CLICK:
    begin
    application.Restore();
    application.BringToFront();
    connBox.setFocus();
    end;
  end;
end; // downloadtrayEvent

procedure Tmainfrm.updateTrayTip();
var
  f: Tfile;
  s: string;
begin
if quitting then exit;
f:=rootFile;
s:='';
if assigned(f) then s:=f.fullURL();

tray.setTip(x_template(trayMsg, [
  '%uptime%', uptimestr(),
  '%url%', s,
  '%ip%', defaultIP,
  '%port%', srv.port,
  '%hits%', intToStr(hitsLogged),
  '%downloads%', intToStr(downloadsLogged),
  '%uploads%', intToStr(uploadsLogged),
  '%version%', VERSION
]));
end; // updateTrayTip

procedure TmainFrm.Restore1Click(Sender: TObject);
begin application.Restore() end;

procedure TmainFrm.Numberofcurrentconnections1Click(Sender: TObject);
begin setTrayShows('connections') end;

procedure TmainFrm.NumberofdifferentIPaddresses1Click(Sender: TObject);
begin setTrayShows('ips') end;

procedure TmainFrm.NumberofdifferentIPaddresseseverconnected1Click(
  Sender: TObject);
begin setTrayShows('ips-ever') end;

procedure TmainFrm.Numberofloggeddownloads1Click(Sender: TObject);
begin setTrayShows('downloads') end;

procedure TmainFrm.Numberofloggedhits1Click(Sender: TObject);
begin setTrayShows('hits') end;

procedure Tmainfrm.setTrayShows(s:string);
begin
trayShows:=s;
repainttray();
end; // setTrayShows

procedure TmainFrm.Exit1Click(Sender: TObject);
begin close() end;

procedure TmainFrm.onDownloadChkClick(Sender: TObject);
begin flashOn:='download' end;

procedure TmainFrm.onconnectionChkClick(Sender: TObject);
begin flashOn:='connection' end;

procedure TmainFrm.never1Click(Sender: TObject);
begin flashOn:='' end;

procedure Tmainfrm.addTray();
begin
repaintTray();
tray.show();
end; // addTray

procedure TmainFrm.Allowarchive1Click(Sender: TObject);
begin fileMenuSetFlag(sender, FA_ARCHIVABLE) end;

procedure TmainFrm.Allowedreferer1Click(Sender: TObject);
const
  MSG = 'Leave empty to disable this feature.'
    +#13'Here you can specify a mask.'
    +#13'When a file is requested, if the mask doesn''t match'
    +#13'the "Referer" HTTP field, the request is rejected.';
begin
inputQuery('Allowed referer', MSG, allowedReferer)
end;

// addtray

procedure TmainFrm.FormShow(Sender: TObject);
begin
if trayed then showWindow(application.handle, SW_HIDE);
updateTrayTip();
end;

procedure TmainFrm.filesBoxDragOver(Sender, Source: TObject; X, Y: Integer;
  State: TDragState; var Accept: Boolean);
const
  THRESHOLD = 10;
var
  src, dst: Tfile;
  i: integer;
begin
scrollFilesBox:=-1;
if y < THRESHOLD then scrollFilesBox:=SB_LINEUP;
if filesBox.Height-y < THRESHOLD then scrollFilesBox:=SB_LINEDOWN;

accept:=FALSE;
if sender <> source then exit; // only move files within filesBox
dst:=pointedFile(FALSE);
if assigned(dst) and not dst.isFolder() then dst:=dst.parent;
if dst = NIL then exit;
for i:=0 to filesBox.SelectionCount-1 do
  with Tfile(filesbox.selections[i].data) do
    if isRoot() or isLocked() then exit;
src:=selectedFile;
accept:=(dst <> src.parent) and (dst <> src);
end;

procedure TmainFrm.filesBoxDragDrop(Sender, Source: TObject; X,Y: Integer);
var
  dst: Ttreenode;
  i, bak: integer;
  nodes: array of Ttreenode;
begin
if selectedFile = NIL then exit;
VFSmodified:=TRUE;
dst:=filesBox.DropTarget;
if not Tfile(dst.Data).isFolder() then dst:=dst.Parent;
// copy list of selected nodes
setlength(nodes, filesBox.SelectionCount);
for i:=0 to filesBox.SelectionCount-1 do nodes[i]:=filesbox.selections[i];
// check for namesakes
for i:=0 to length(nodes)-1 do
  if existsNodeWithName(nodes[i].Text, dst) then
    if msgDlg(MSG_SAME_NAME, MB_ICONWARNING+MB_YESNO) = IDYES then break
    else exit;
// move'em
for i:=0 to length(nodes)-1 do
  begin
  // removing and restoring stateIndex is a workaround to a delphi bug
  bak:=nodes[i].stateIndex;
  nodes[i].stateIndex:=0;
  nodes[i].MoveTo(dst, naAddChild);
  nodes[i].stateIndex:=bak;
  end;
dst.AlphaSort(FALSE);
end;

procedure TmainFrm.EditHTMLtemplate1Click(Sender: TObject);
var
  s: string;
begin
application.restore();
if tplFrm.windowState = wsMinimized then tplFrm.windowState:=wsMaximized;
if tplFrm.Visible then tplFrm.SetFocus()
else
  begin
  s:=tpl.fullText;
  if tpl.utf8 then s:=UTF8toAnsi(s);
  tplFrm.box.lines.text:=s;
  tplFrm.show();
  end;
end;

procedure TmainFrm.refreshConn(conn:ThttpConn);
var
  r: Trect;
begin
r:=connBox.itemRect(connBox.Items.IndexOfObject(conn.data));
invalidateRect(connBox.handle,@r,TRUE);
refreshSbar();
end; // refreshConn

const
  // IDs used for file chunks
  FK_HEAD = 0;
  FK_RESOURCE = 1;
  FK_NAME = 2;
  FK_FLAGS = 3;
  FK_NODE = 4;
  FK_FORMAT_VER = 5;
  FK_CRC = 6;
  FK_COMMENT = 7;
  FK_USERPWD = 8;
  FK_ADDEDTIME = 9;
  FK_DLCOUNT = 10;
  FK_ROOT = 11;
  FK_ACCOUNTS = 12;
  FK_FILESFILTER = 13;
  FK_FOLDERSFILTER = 14;
  FK_ICON_GIF = 15;
  FK_REALM = 16;
  FK_UPLOADACCOUNTS = 17;
  FK_DEFAULTMASK = 18;
  FK_DONTCOUNTASDOWNLOADMASK = 19;
  FK_AUTOUPDATED_FILES = 20;
  FK_DONTCOUNTASDOWNLOAD = 21;
  FK_HFS_VER = 22;
  FK_HFS_BUILD = 23;
  FK_COMPRESSED_ZLIB = 24;
  FK_DIFF_TPL = 25;

function Tmainfrm.getVFS(node:Ttreenode=NIL):string;

  function getAutoupdatedFiles():string;
  var
    i: integer;
    fn: string;
  begin
  result:='';
  i:=0;
  while i < autoupdatedFiles.Count do
    begin
    fn:=autoupdatedFiles[i];
    result:=result+TLV(FK_NODE, TLV(FK_NAME, fn)
      + TLV(FK_DLCOUNT, str_(autoupdatedFiles.getInt(fn))) );
    inc(i);
    end;
  end; // getAutoupdatedFiles

var
  i: integer;
  f: Tfile;
  commonFields: string;
begin
if node = NIL then node:=rootNode;
if node = NIL then exit;
f:=Tfile(node.data);
commonFields:=TLV(FK_FLAGS, str_(f.flags))
    +TLV_NOT_EMPTY(FK_RESOURCE, f.resource)
    +TLV_NOT_EMPTY(FK_COMMENT, f.comment)
    +ifThen(f.user>'', TLV(FK_USERPWD, base64encode(f.user+':'+f.pwd)))
    +TLV_NOT_EMPTY(FK_ACCOUNTS, join(';',f.accounts) )
    +TLV_NOT_EMPTY(FK_UPLOADACCOUNTS, join(';',f.uploadAccounts))
    +TLV_NOT_EMPTY(FK_FILESFILTER, f.filesfilter)
    +TLV_NOT_EMPTY(FK_FOLDERSFILTER, f.foldersfilter)
    +TLV_NOT_EMPTY(FK_REALM, f.realm)
    +TLV_NOT_EMPTY(FK_DEFAULTMASK, f.defaultFileMask)
    +TLV_NOT_EMPTY(FK_DONTCOUNTASDOWNLOADMASK, f.dontCountAsDownloadMask)
    +TLV_NOT_EMPTY(FK_DIFF_TPL, f.diffTpl);

result:='';
if f.isRoot() then
	result:=result+TLV(FK_ROOT, commonFields );
for i:=0 to node.Count-1 do
  result:=result+getVFS(node.item[i]); // recursion
if f.isRoot() then
  begin
  result:=result+TLV_NOT_EMPTY(FK_AUTOUPDATED_FILES, getAutoupdatedFiles() );
  exit;
  end;
// for non-root nodes, subnodes must be calculated first, so to be encapsulated
result:=TLV(FK_NODE, commonFields
  +TLV_NOT_EMPTY(FK_NAME, f.name)
  +TLV(FK_ADDEDTIME, str_(f.atime))
  +TLV(FK_DLCOUNT, str_(f.DLcount))
  +TLV_NOT_EMPTY(FK_ICON_GIF, pic2str(f.icon))
  +result // subnodes
);
end; // getVFS

var
  buildLoading: string;

procedure Tmainfrm.setVFS(vfs:string; node:Ttreenode=NIL);
const
  MSG_BETTERSTOP = #13'Going on may lead to problems.'
    +#13'It is adviced to stop loading.'
    +#13'Stop?';
  MSG_BADCRC = 'This file could be corrupted, because the CRC check failed.';
  MSG_NEWER='This file has been created with a newer and incompatible version.';

var
  data: string;
  f: Tfile;
  initialLength: integer;
  pushed: boolean;
  after: record
    resetLetBrowse: boolean;
    end;

  procedure parseAutoupdatedFiles(data:string);
  var
    s, s2, fn: string;
  begin
  autoupdatedFiles.Clear();
  while popTLV(data, s) = FK_NODE do
    while s > '' do
      begin
      progFrm.progress:= (initialLength - length(vfs) - length(data))
        / initialLength;
      case popTLV(s, s2) of
        FK_NAME: fn:=s2;
        FK_DLCOUNT: autoupdatedFiles.setInt(fn, int_(s2));
        end;
      end;
  end; // parseAutoupdatedFiles

begin
if vfs = '' then exit;
if node = NIL then
  begin
  node:=rootNode;
  uploadPaths:=NIL;
  usersInVFS.reset();
  end;
fillchar(after, sizeof(after), 0);
initialLength:=length(vfs);
node.DeleteChildren();
f:=Tfile(node.data);
f.node:=node;
while vfs > '' do
  case popTLV(vfs, data) of
    -1: break;
    FK_ROOT:
      begin
      setVFS(data, rootNode );
      if buildLoading < '109' then
        include(f.flags, FA_ARCHIVABLE);
      end;
    FK_NODE:
      begin
      if progFrm.cancelRequested then exit;
      pushed:=progFrm.visible;
      if pushed then
        begin
        progFrm.progress:= (initialLength - length(vfs) - length(data)) / initialLength;
        progFrm.push( length(data) / initialLength );
        application.ProcessMessages();
        end;
      setVFS(data, addFile(Tfile.create(''), node, TRUE).node );
      if pushed then progFrm.pop();
      end;
    FK_COMPRESSED_ZLIB: setVFS(ZDecompressStr2(data,31), node);
    FK_FORMAT_VER:
      begin
      if length(data) < 4 then // early versions: '1.0', '1.1'
        begin
        loadingVFS.resetLetBrowse:=TRUE;
        after.resetLetBrowse:=TRUE;
        end;
      if (int_(data) > CURRENT_VFS_FORMAT)
      and (msgDlg(MSG_NEWER+MSG_BETTERSTOP, MB_ICONERROR+MB_YESNO) = IDYES) then
        exit;
      end;
  	FK_CRC:
      if (str_(getCRC(vfs)) <> data)
      and (msgDlg(MSG_BADCRC+MSG_BETTERSTOP,MB_ICONERROR+MB_YESNO) = IDYES) then
      	exit;
    FK_RESOURCE: f.resource:=data;
    FK_NAME:
      begin
      f.name:=data;
      node.text:=data;
      end;
    FK_FLAGS: move(data[1], f.flags, length(data));
  	FK_ADDEDTIME: f.atime:=dt_(data);
    FK_COMMENT: f.comment:=data;
    FK_USERPWD:
    	begin
      data:=base64decode(data);
      f.user:=chop(':',data);
      f.pwd:=data;
      usersInVFS.track(f.user, f.pwd);
      end;
    FK_DLCOUNT: f.DLcount:=int_(data);
    FK_ACCOUNTS: f.accounts:=split(';',data);
    FK_UPLOADACCOUNTS: f.uploadAccounts:=split(';',data);
    FK_FILESFILTER: f.filesfilter:=data;
    FK_FOLDERSFILTER: f.foldersfilter:=data;
    FK_REALM: f.realm:=data;
    FK_DEFAULTMASK: f.defaultFileMask:=data;
    FK_DIFF_TPL: f.diffTpl:=data;
    FK_DONTCOUNTASDOWNLOADMASK: f.dontCountAsDownloadMask:=data;
    FK_DONTCOUNTASDOWNLOAD: // legacy, now moved into flags
      if boolean(data[1]) then include(f.flags, FA_DONT_COUNT_AS_DL);
    FK_ICON_GIF: if data > '' then f.setupImage(str2pic(data));
    FK_AUTOUPDATED_FILES: parseAutoupdatedFiles(data);
    FK_HFS_BUILD: buildLoading:=data;
    FK_HEAD, FK_HFS_VER: ; // recognize these fields, but do nothing
    else loadingVFS.unkFK:=TRUE;
    end;
if FA_VIS_ONLY_ANON in f.flags then
  loadingVFS.visOnlyAnon:=TRUE;
if f.isVirtualFolder() then f.mtime:=f.atime;
if assigned(f.uploadAccounts) and (f.resource > '') then
  addString(f.resource, uploadPaths);
f.setupImage();
if after.resetLetBrowse then
  f.recursiveApply(setBrowsable, integer(FA_BROWSABLE in f.flags));
end; // setVFS

function addVFSheader(vfsdata:string):string;
begin
if mainfrm.filesBox.Items.Count > COMPRESSION_THRESHOLD then
  vfsdata:=TLV(FK_COMPRESSED_ZLIB,
    ZcompressStr2(vfsdata, zcFastest, 31,8, zsDefault) );
result:= TLV(FK_HEAD, 'HFS.VFS')
  +TLV(FK_FORMAT_VER, str_(CURRENT_VFS_FORMAT))
  +TLV(FK_HFS_VER, VERSION)
  +TLV(FK_HFS_BUILD, VERSION_BUILD)
  +TLV(FK_CRC, str_(getCRC(vfsdata)));  // CRC must always be right before data
result:=result+vfsdata
end; // addVFSheader

procedure TmainFrm.Savefilesystem1Click(Sender: TObject);
begin saveVFS() end;

procedure TmainFrm.filesBoxDeletion(Sender: TObject; Node: TTreeNode);
var
  f: Tfile;
begin
f:=node.data;
node.data:=NIL;
// the test on uploadPaths may save some function call
if assigned(f.uploadAccounts) and assigned(uploadPaths) then
  removeString(f.resource, uploadPaths);
try f.free
except
  end;
if node = rootNode then
  rootNode:=NIL;
VFSmodified:=TRUE
end;

procedure TmainFrm.Loadfilesystem1Click(Sender: TObject);
var
  fn: string;
begin
fn:='';
if PromptForFileName(fn, 'VirtualFileSystem|*.vfs', 'vfs', 'Open VFS file') then
  loadVFS(fn);
end;

procedure TmainFrm.connheaderSectionEndDrag(Sender: TObject);
begin connBox.Repaint() end;

procedure TmainFrm.leavedisconnectedconnectionsChkClick(Sender: TObject);
var
  i: integer;
  data: TconnData;
begin
if leavedisconnectedconnectionsChk.checked then exit;
i:=0;
while i < connBox.Count do
  begin
  data:=TconnData(connBox.Items.objects[i]);
  if data.conn.state = HCS_DISCONNECTED then
    begin
    toDelete.Add(data);
    data.deleting:=TRUE;
    end;
  inc(i);
  end;
end;

procedure drawGraphOn(cnv:Tcanvas; colors:TIntegerDynArray=NIL);
var
  i, h, maxV: integer;
  r: Trect;

  procedure drawSample(sample:integer);
  begin
	cnv.moveTo(r.left+i, r.bottom);
  cnv.lineTo(r.Left+i, r.Bottom-1-round(sample*h div maxV));
  end; // drawSample

  function getColor(idx:integer; def:Tcolor):Tcolor;
  begin
  if (length(colors) <= idx) or (colors[idx] = -1) then result:=def
  else result:=colors[idx]
  end; // getColor

begin
r:=cnv.cliprect;
// clear
cnv.brush.color:=getColor(0, clBlack);
cnv.fillrect(r);
// draw grid
cnv.Pen.color:=getColor(1, rgb(0,0,120));
i:=r.left;
while i < r.right do
  begin
  cnv.moveTo(i, r.top);
  cnv.LineTo(i, r.Bottom);
  inc(i,10);
  end;
i:=r.bottom;
while i > r.top do
  begin
  cnv.moveTo(r.left, i);
  cnv.LineTo(r.right, i);
  dec(i,10);
  end;

maxV:=max(graph.maxV, 1);
h:=r.bottom-r.top-1;
// draw graph
cnv.Pen.color:=getColor(2, clFuchsia);
for i:=0 to (r.Right-r.left)-1 do	drawSample(graph.samplesOut[i]);
cnv.Pen.color:=getColor(3, clYellow);
for i:=0 to (r.Right-r.left)-1 do	drawSample(graph.samplesIn[i]);
// text
cnv.Font.Color:=getColor(4, clLtGray);
cnv.Font.Name:='Small Fonts';
cnv.font.size:=7;
SetBkMode(cnv.handle, TRANSPARENT);
cnv.TextOut(r.right-130, 3, format('Top speed: %.1f KB/s', [
  (graph.maxV/1000)*safeDiv(10.0, graph.rate) ]));
if assigned(srv) and (srv.maxSpeed < MAXINT) then
  cnv.TextOut(r.right-130+25, 15, format('Limit: %.1f KB/s', [srv.maxSpeed/1000]));
end; // drawGraphOn

procedure Tmainfrm.refreshGraphBox();
var
  bmp: Tbitmap;
  r: Trect;
begin
if not graphBox.visible then exit;
bmp:=Tbitmap.create;
bmp.Width:=graphBox.Width;
bmp.Height:=graphBox.Height;
r:=bmp.canvas.ClipRect;
drawGraphOn(bmp.canvas);
graphBox.canvas.CopyRect(r,bmp.canvas,r);
bmp.free;
end; // refreshGraphBox

function Tmainfrm.getGraphPic(conn:ThttpConn=NIL):string;
var
  bmp: Tbitmap;
  refresh: string;
  i: integer;
  colors: TIntegerDynArray;
  options: string;
  data: TconnData;

  procedure addColor(c:Tcolor);
  var
    n: integer;
  begin
  n:=length(colors);
  setLength(colors, n+1);
  colors[n]:=c;
  end; // addColor

  procedure fetchColor();
  var
    s: string;
  begin
  s:=chop('x',options);
  if s = '' then exit;
  if not ansiStartsText('cl', s) and not ansiStartsText('$', s) then
    begin
    if length(s) = 6 then s:=s[5]+s[6]+s[3]+s[4]+s[1]+s[2];
    if length(s) = 3 then s:=s[3]+s[3]+s[2]+s[2]+s[1]+s[1];
    s:='$'+s;
    end;
  try addColor(stringToColor(s));
  except addColor(-1) end;
  end; // fetchColor

begin
data:=conn.data;
options:=copy(conn.request.url, 12, MAXINT);
delete(options, pos('?',options), MAXINT);
bmp:=Tbitmap.create;
bmp.Width:=graphBox.Width;
bmp.Height:=graphBox.Height;
colors:=NIL;
if options = '' then
  begin
  // here is an initial support for ?parameters. colors not supported yet.
  try bmp.width:=strToInt(data.params.Values['w']) except end;
  try bmp.height:=min(strToInt(data.params.Values['h']), 300000 div max(1,bmp.width)) except end;
  refresh:=data.params.Values['refresh'];
  end
else
  try
    i:=strToInt(chop('x',options));
    if (i > 0) and (i <= length(graph.samplesIn)) then bmp.Width:=i;
    i:=strToInt(chop('x',options));
    if (i > 0) and (i <= length(graph.samplesIn)) then
      bmp.height:=min(i, 300000 div max(1,bmp.width));
    refresh:=chop('x',options);
    for i:=1 to 5 do fetchColor();
  except
    end;
drawGraphOn(bmp.canvas, colors);
result:=bmp2str(bmp);
bmp.free;
if conn = NIL then exit;
conn.addHeader('Cache-Control: no-cache');
if refresh > '' then
  conn.addHeader('Refresh: '+refresh);
end; // getGraphPic

procedure TmainFrm.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
altPressedForMenu:=(key = 18) and (shift = [ssAlt]);

if shift = [] then
  case key of
    VK_F4: toggleServer();
    VK_F5: setEasyMode(not easyMode);
    VK_F10: popupMainMenu();
    end;
if shift = [ssCtrl] then
	case key of
  	ord('O'): Loadfilesystem1Click(sender);
    ord('S'): saveVFS();
    end;
end;

procedure TmainFrm.Officialwebsite1Click(Sender: TObject);
begin exec('http://www.rejetto.com/hfs/') end;

procedure TmainFrm.showmaintrayiconChkClick(Sender: TObject);
begin
if showmaintrayiconChk.Checked then addTray()
else tray.hide();
end;

function Shell_GetImageLists(var hl,hs:Thandle):boolean;
  stdcall; external 'shell32.dll' index 71;

function getSystemimages():TImageList;
var
  hl, hs: Thandle;
begin
result:=NIL;
if not Shell_GetImageLists(hl,hs) then exit;
result:=Timagelist.Create(NIL);
result.ShareImages:=TRUE;
result.handle:=hs;
end; // loadSystemimages

procedure TmainFrm.Speedlimit1Click(Sender: TObject);
const
  MSG = 'Max bandwidth (KB/s).'#13+MSG_EMPTY_NO_LIMIT;
var
  s: string;
begin
if speedLimit < 0 then s:=''
else s:=floatToStr(speedLimit);
if inputquery('Speed limit', MSG, s) then
	try
  	s:=trim(s);
  	if s = '' then setSpeedLimit(-1)
    else setSpeedLimit(strToFloat(s));
    // a manual set of speedlimit voids the pause command
    Pausestreaming1.Checked:=FALSE;
  except msgDlg(MSG_INVALID_VALUE, MB_ICONERROR)
  end;
end;

procedure Tmainfrm.setSpeedLimit(v:real);
begin
speedLimit:=v;
srv.maxSpeed:=round(v*1000);
speedLimit1.caption:=
	'Speed limit: '+ifThen(v<0, 'disabled', floatToStr(v)+' KB/s' );
end;

procedure Tmainfrm.setnoDownloadTimeout(v:integer);
begin
if v < 0 then v:=0;
if v <> noDownloadTimeout then lastActivityTime:=now();
noDownloadTimeout:=v;
noDownloadTimeout1.caption:=
	'No downloads timeout: '+ifThen(v=0, 'disabled', intToStr(v) );
end;                                        

procedure Tmainfrm.setGraphRate(v:integer);
const
  MSG = 'Graph refresh rate: %d (tenths of second)';
begin
if v < 1 then v:=1;
if graph.rate = v then exit;
graph.rate:=v;
Graphrefreshrate1.caption:=format(MSG, [v]);
// changing rate invalidates previous data
fillChar(graph.samplesOut, sizeof(graph.samplesOut), 0);
fillChar(graph.samplesIn, sizeof(graph.samplesIn), 0);
graph.maxV:=0;
end; // setGraphRate

procedure TmainFrm.Maxconnections1Click(Sender: TObject);
const
  MSG = 'Max simultaneous connections to serve.'
    +#13+MSG_EMPTY_NO_LIMIT;
  MSG2 = 'In this moment there are %d active connections';
var
  s: string;
begin
if maxConnections > 0 then s:=intToStr(maxConnections)
else s:='';
if inputquery('Max connections', MSG, s) then
	try setMaxConnections(strToUInt(s))
  except msgDlg(MSG_INVALID_VALUE, MB_ICONERROR)
  end;
if (maxConnections > 0) and (srv.conns.count > maxConnections) then
  msgDlg(format(MSG2, [srv.conns.count]), MB_ICONWARNING);
end;

procedure TmainFrm.maxDLs1Click(Sender: TObject);
const
  MSG = 'Max simultaneous downloads.'#13+MSG_EMPTY_NO_LIMIT;
  MSG2 = 'In this moment there are %d active downloads';
var
  s: string;
  i: integer;
begin
if maxContempDLs > 0 then s:=intToStr(maxContempDLs)
else s:='';
if inputquery('Max downloads', MSG, s) then
	try setMaxDLs(strToUInt(s))
  except msgDlg(MSG_INVALID_VALUE, MB_ICONERROR)
  end;
if maxContempDLs = 0 then exit;
i:=countDownloads();
if i > maxContempDLs then
  msgDlg(format(MSG2, [i]), MB_ICONWARNING);
end;

procedure TmainFrm.Maxconnectionsfromsingleaddress1Click(Sender: TObject);
const
  MSG = 'Max simultaneous connections to serve from a single IP address.'
    +#13+MSG_EMPTY_NO_LIMIT;
var
  s: string;
  addresses: TStringDynArray;
  i: integer;
begin
if maxConnectionsIP > 0 then s:=intToStr(maxConnectionsIP)
else s:='';
if inputquery('Max connections by IP', MSG, s) then
	try setMaxConnectionsIP(strToUInt(s))
  except msgDlg(MSG_INVALID_VALUE, MB_ICONERROR)
  end;
if maxConnectionsIP = 0 then exit;
addresses:=NIL;
for i:=0 to srv.conns.Count-1 do
  with ThttpConn(srv.conns[i]) do
    if countConnectionsByIP(address) > maxConnectionsIP then
      addUniqueString(address, addresses);
if assigned(addresses) then
  msgDlg(format(MSG_ADDRESSES_EXCEED,[join(#13, addresses)]), MB_ICONWARNING);
end;

procedure TmainFrm.MaxDLsIP1Click(Sender: TObject);
const
  MSG = 'Max simultaneous downloads from a single IP address.'
    +#13+MSG_EMPTY_NO_LIMIT;
var
  s: string;
  addresses: TStringDynArray;
  i: integer;
begin
if maxContempDLsIP > 0 then s:=intToStr(maxContempDLsIP)
else s:='';
if inputquery('Max downloads by IP', MSG, s) then
	try setMaxDLsIP(strToUInt(s))
  except msgDlg(MSG_INVALID_VALUE, MB_ICONERROR)
  end;
if maxContempDLsIP = 0 then exit;
addresses:=NIL;
for i:=0 to srv.conns.Count-1 do
  with ThttpConn(srv.conns[i]) do
    if countDownloads(address) > maxContempDLsIP then
      addUniqueString(address, addresses);
if assigned(addresses) then
  msgDlg(format(MSG_ADDRESSES_EXCEED,[join(#13, addresses)]), MB_ICONWARNING);
end;

procedure TmainFrm.Forum1Click(Sender: TObject);
begin exec('http://www.rejetto.com/forum/') end;

procedure TmainFrm.FAQ1Click(Sender: TObject);
begin exec('http://www.rejetto.com/sw/?faq=hfs') end;

procedure TmainFrm.License1Click(Sender: TObject);
begin exec('http://www.rejetto.com/sw/license.txt') end;

procedure Tmainfrm.pasteFiles();
begin
// try twice
try addDropFiles(clipboard.GetAsHandle(CF_HDROP), filesBox.selected)
except
  try addDropFiles(clipboard.GetAsHandle(CF_HDROP), filesBox.selected)
  except on e:Exception do
    msgDlg(e.message, MB_ICONERROR);
    end
  end;
end;

procedure TmainFrm.Paste1Click(Sender: TObject);
begin pasteFiles() end;

procedure TmainFrm.Addfiles1Click(Sender: TObject);
var
  dlg: TopenDialog;
  i: integer;
begin
dlg:=TopenDialog.create(self);
dlg.Options:=dlg.Options+[ofAllowMultiSelect, ofFileMustExist, ofPathMustExist];
if dlg.Execute() then
	for i:=0 to dlg.files.count-1 do
    addFile(Tfile.create(dlg.files[i]), filesBox.Selected, dlg.Files.count<>1 );
dlg.free;
end;

procedure TmainFrm.Addfolder1Click(Sender: TObject);
begin
if selectFolder('',lastAddedFolder) then
  addFilesFromString(lastAddedFolder, filesBox.selected);
end;

procedure TmainFrm.Splitter1Moved(Sender: TObject);
begin
if graphBox.height > 0 then graph.size:=graphBox.height
else hideGraph();
refreshGraphBox();
end;

procedure TmainFrm.FormPaint(Sender: TObject);
begin refreshGraphBox() end;

procedure TmainFrm.Graphrefreshrate1Click(Sender: TObject);
var
  s: string;
begin
s:=intToStr(graph.rate);
if inputquery('Graph refresh rate', 'Tenths of second',s) then
	try
  	s:=trim(s);
  	if s = '' then setGraphRate(10)
    else setGraphRate(strToInt(s))
  except msgDlg(MSG_INVALID_VALUE, MB_ICONERROR)
  end;
end;

procedure TmainFrm.Pausestreaming1Click(Sender: TObject);
begin
if Pausestreaming1.checked then srv.maxSpeed:=0
else setSpeedLimit(speedLimit)
end;

procedure TmainFrm.Comment1Click(Sender: TObject);
var
  i: integer;
begin
if selectedFile = NIL then exit;
inputComment(selectedFile);
for i:=0 to filesBox.SelectionCount-1 do
  Tfile(filesBox.Selections[i].data).comment:=selectedFile.comment;
end;

procedure TmainFrm.filesBoxCustomDrawItem(Sender: TCustomTreeView;
  Node: TTreeNode; State: TCustomDrawState; var DefaultDraw: Boolean);
var
  f: Tfile;
begin
if not sender.Visible then exit;
f:=Tfile(node.data);
if f = NIL then exit;
if not f.visible then
	with sender.Canvas.Font do
  	style:=style+[fsItalic];
if (f.user > '') or assigned(f.accounts) then node.StateIndex:=12
else node.StateIndex:=-1;
end;

function Tmainfrm.fileAttributeInSelection(fa:TfileAttribute):boolean;
var
  i: integer;
begin
for i:=0 to filesBox.SelectionCount-1 do
  with Tfile(filesBox.Selections[i].data) do
    if FA_LINK in flags then
      begin
      result:=TRUE;
      exit;
      end;
result:=FALSE;
end; // fileAttributeInSelection

procedure TmainFrm.Setuserpass1Click(Sender: TObject);
var
  i: integer;
  user, pwd: string;
  f: Tfile;
begin
if selectedFile = NIL then exit;
if fileAttributeInSelection(FA_LINK)
and (msgDlg(MSG_UNPROTECTED_LINKS, MB_ICONWARNING+MB_YESNO) <> IDYES) then exit;
user:=selectedFile.user;
pwd:=selectedFile.pwd;
if not newuserpassFrm.prompt(user, pwd) then exit;
for i:=0 to filesBox.SelectionCount-1 do
  begin
  f:=filesBox.Selections[i].data;
  usersInVFS.drop(f.user, f.pwd);
  f.user:=user;
  f.pwd:=pwd;
  usersInVFS.track(f.user, f.pwd);
  end;
filesBox.Repaint();
VFSmodified:=TRUE;
end;

procedure TmainFrm.browseBtnClick(Sender: TObject);
begin browse(urlBox.Text) end;

procedure TmainFrm.BanIPaddress1Click(Sender: TObject);
var
  c: ThttpConn;
begin
c:=selectedConnection();
if c = NIL then exit;
banAddress(c.address)
end;

procedure showOptions(page:TtabSheet);
begin
optionsFrm.pageCtrl.ActivePage:=page;
if mainfrm.modalOptionsChk.checked then optionsFrm.ShowModal()
else optionsFrm.show();
end;

procedure TmainFrm.BannedIPaddresses1Click(Sender: TObject);
begin showOptions(optionsFrm.bansPage) end;

procedure Tmainfrm.recentsClick(sender:Tobject);
var
  i: integer;
begin
i:=strToInt((sender as Tmenuitem).Caption[3]);
if i > length(recentFiles) then exit;
if FileExists(recentFiles[i-1]) then loadVFS(recentFiles[i-1])
else
  begin
  msgDlg('The file does not exist anymore', MB_ICONERROR);
  removeString(recentFiles, i-1);
  updateRecentFilesMenu();
  end;
end;

procedure Tmainfrm.updateRecentFilesMenu();
var
  i: integer;
begin
Loadrecentfiles1.Clear();
for i:=0 to length(recentFiles)-1 do
  loadrecentfiles1.Add(
    NewItem( '[&'+intToStr(i+1)+'] '+ExtractFileName(recentFiles[i]), 0, FALSE,
      TRUE, recentsClick, 0, 'recent')
  );
Loadrecentfiles1.Visible:=Loadrecentfiles1.count>0;
end; // updateRecentFilesMenu

procedure Tmainfrm.loadVFS(fn:string);
const
  MSG_TITLE = 'Loading VFS';
  MSG_OLD = 'This file is old and uses different settings.'
    +#13'The "let browse" folder option will be reset.'
    +#13'Re-saving the file will update its format.';
  MSG_UNK_FK = 'This file has been created with a newer version.'
    +#13'Some data was discarded because unknown.'
    +#13'If you save the file now, the discarded data will NOT be saved.';
  MSG_VIS_ONLY_ANON =
    'This VFS file uses the "Visible only to anonymous users" feature.'
  +#13'This feature is not available anymore.'
  +#13'You can achieve similar results by restricting access to @anonymous,'
  +#13'then enabling "List protected items only for allowed users".';
  MSG_AUTO_DISABLED = 'Because of the problems encountered in loading,'
    +#13'automatic saving has been disabled'
    +#13'until you save manually or load another one.';

  function anyAutosavingFeatureEnabled():boolean;
  begin  result:=(autosaveVFS.every > 0) or autosaveVFSchk.checked end;

var
  took: Tdatetime;
begin
progFrm.show('Loading VFS...', TRUE);
disableUserInteraction();
filesBox.hide(); // it seems to speed up a lot
fillchar(loadingVFS, sizeof(loadingVFS), 0);
try
  took:=now();
  initVFS();
  setVFS(loadfile(fn));
finally
  reenableUserInteraction();
  if progFrm.cancelRequested then initVFS()
  else lastFileOpen:=fn;
  progFrm.hide();
  end;
took:=now()-took;
purgeVFSaccounts(); // remove references to unexistant users
filesBox.FullCollapse();
rootNode.Selected:=TRUE;
rootNode.MakeVisible();
filesBox.show();
VFSmodified:=FALSE;
if progFrm.cancelRequested then exit;
if loadingVFS.visOnlyAnon then
  msgDlg(MSG_VIS_ONLY_ANON, MB_ICONWARNING, MSG_TITLE);
if loadingVFS.resetLetBrowse then msgDlg(MSG_OLD, MB_ICONWARNING, MSG_TITLE);
if loadingVFS.unkFK then msgDlg(MSG_UNK_FK, MB_ICONWARNING, MSG_TITLE);

with loadingVFS do disableAutosave:=unkFK or resetLetBrowse or visOnlyAnon;
if loadingVFS.disableAutosave and anyAutosavingFeatureEnabled() then
  msgDlg(MSG_AUTO_DISABLED, MB_ICONWARNING, MSG_TITLE);

setStatusBarText(format('Loaded in %.1f seconds (%s)', [took*SECONDS,fn]), 10);

removestring(fn, recentFiles); // avoid duplicates
insertstring(fn, 0, recentFiles); // insert fn as first element
removeString(recentFiles, MAX_RECENT_FILES, length(recentFiles)); // shrink2max
updateRecentFilesMenu();

if autocopyURLonadditionChk.checked then
  setClip( rootFile.fullURL() );
end;

// loadVFS

procedure TmainFrm.FormKeyUp(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
if (key <> 18) or (shift <> []) then altPressedForMenu:=FALSE
else
  if altPressedForMenu then popupMainMenu();
end;

procedure Tmainfrm.popupMainMenu();
begin
menuBtn.Down:=TRUE;
with menuBtn.clientToScreen(point(0,menuBtn.height)) do
  menu.popup(x,y);
menuBtn.Down:=FALSE;
end;

procedure TmainFrm.portBtnClick(Sender: TObject);
var
  s, bakS: string;
  bakB: boolean;
begin
bakB:=srv.active;
bakS:=port;
s:=port;
  repeat
  if not inputQuery('Port', 'Specify a port to accept connection,'#13'or leave empty to decide automatically.', s) then exit;
  s:=trim(s);
  if not isOnlyDigits(s) then
    begin
    msgDlg('Numbers only', MB_ICONERROR);
    continue;
    end;
  port:=s;
  if srv.active and (s = srv.port) then exit;
  stopServer();
  if startServer() then
    begin
    if not bakB then stopServer();
    exit;
    end;
  sayPortBusy();
  port:=bakS;
  if bakB then startServer();
  until FALSE;
end;

procedure Tmainfrm.updateAlwaysOnTop();
begin
if alwaysOnTopchk.checked then FormStyle:=fsStayOnTop
else formStyle:=fsNormal
end; // updateAlwaysOnTop

procedure TmainFrm.Checkforupdates1Click(Sender: TObject);
const
  MSG_INFO = 'Last stable version: %s'#13'Last untested version: %s'#13;
  MSG_NEWER = 'There''s a new version available online: %s';
  MSG_ERROR = 'Communication error';
var
  updateURL: string;
  info: Ttpl;
begin
progFrm.show('Requesting...');
try info:=downloadUpdateInfo()
finally progFrm.hide() end;

if info = NIL then
  begin
  msgDlg(MSG_ERROR, MB_ICONERROR);
  exit;
  end;

try
  msgDlg(format(MSG_INFO, [ info['last stable'], info['last untested'] ]));

  updateURL:='';
  if trim(info['last stable build']) > VERSION_BUILD then
    begin
    msgDlg(format(MSG_NEWER,[info['last stable']]));
    updateURL:=trim(info['last stable url']);
    end
  else
    if (not VERSION_STABLE or testerUpdatesChk.checked)
    and (trim(info['last untested build']) > VERSION_BUILD) then
      begin
      msgDlg(format(MSG_NEWER,[info['last untested']]));
      updateURL:=trim(info['last untested url']);
      end;

  msgDlg(info['notice'], MB_ICONWARNING);
  parseVersionNotice(info['version notice']);
finally
  info.free;
  end;
promptForUpdating(updateURL);
end;

procedure Tmainfrm.setEasyMode(easy:boolean=TRUE);
begin
easyMode:=easy;
modeBtn.Caption:='You are in '+ifThen(easyMode, 'Easy mode', 'Expert mode');
modeBtn.ImageIndex:=ifThen(easyMode, 29, 35);
if easyMode then hideGraph()
else showGraph();
optionsFrm.mimePage.tabVisible:=not easyMode;
optionsFrm.accountsPage.tabVisible:=not easyMode;
optionsFrm.a2nPage.tabVisible:=not easyMode;
refreshSbar();
end; // switchEasyMode

procedure TmainFrm.Rename1Click(Sender: TObject);
begin
if assigned(selectedFile) then
  filesBox.Selected.EditText()
end;

procedure tplSetText(sender:Tobject; text:string);
begin
tpl.fullText:=text;
if not tpl.utf8 then exit;
tpl.fullText:=ansiToUTF8(text);
end; // tplSetText

procedure TmainFrm.noDownloadtimeout1Click(Sender: TObject);
const
  MSG = 'Enter the number of MINUTES with no download after which the '
       +'program automatically shuts down.'#13'Leave blank to get no timeout.';
var
  s:string;
begin
if noDownloadTimeout > 0 then s:=intToStr(noDownloadTimeout)
else s:='';
if inputquery('No downloads timeout', MSG, s) then
	try setnoDownloadTimeout(strToUInt(s))
  except msgDlg(MSG_INVALID_VALUE, MB_ICONERROR)
  end;
end;

procedure Tmainfrm.initVFS();
var
  f:Tfile;
begin
uploadPaths:=NIL;
if assigned(rootNode) then rootNode.delete();
f:=Tfile.createVirtualFolder('/');
f.flags:=f.flags+[FA_ROOT, FA_ARCHIVABLE];
f.dontCountAsDownloadMask:='*.htm;*.html;*.css';
f.defaultFileMask:='';
rootFile:=f;
addFile(f, NIL, TRUE);
rootNode:=rootFile.node;
VFSmodified:=FALSE;
lastFileOpen:='';
end; // initVFS

procedure TmainFrm.alwaysontopChkClick(Sender: TObject);
begin updateAlwaysOnTop() end;

procedure TmainFrm.hideGraph();
begin
splitter1.hide();
graphBox.hide();
end; // hideGraph

procedure TmainFrm.showGraph();
begin
graphBox.show();
graphBox.Height:=graph.size;
splitter1.show();
end; // showGraph

procedure TmainFrm.Showbandwidthgraph1Click(Sender: TObject);
begin showGraph() end;

procedure TmainFrm.Pause1Click(Sender: TObject);
begin with selectedConnection() do paused:=not paused end;

procedure TmainFrm.MIMEtypes1Click(Sender: TObject);
begin showOptions(optionsFrm.mimePage) end;

procedure TmainFrm.Accounts1Click(Sender: TObject);
begin showOptions(optionsFrm.accountsPage) end;

procedure TmainFrm.CopyURL1Click(Sender: TObject);
var
  i: integer;
  s: string;
begin
s:='';
for i:=0 to filesBox.SelectionCount-1 do
  s:=s+Tfile(filesBox.Selections[i].data).fullURL()+CRLF;
setLength(s, length(s)-2);
setClip(s);
end;

procedure Tmainfrm.copyURLwithPasswordMenuClick(sender:TObject);
var
  a: Paccount;
  user, pwd: string;
  f: Tfile;
begin
if selectedFile = NIL then exit;
user:=(sender as Tmenuitem).caption;
delete(user, pos('&',user), 1);
// protection may have been inherited
f:=selectedFile;
while assigned(f) and not assigned(f.accounts) and (f.user = '') do
  f:=f.parent;

if f.user = user then pwd:=f.pwd
else
  begin
  a:=getAccount( user );
  if assigned(a) then pwd:=a.pwd
  else pwd:='';
  end;
if encodePwdUrlChk.checked then pwd:=totallyEncoded(pwd)
else pwd:=encodeURL(pwd);

setClip( selectedFile.fullURL( encodeURL(user)+':'+pwd ) )
end; // copyURLwithPasswordMenuClick

procedure Tmainfrm.copyURLwithAddressMenuClick(sender:Tobject);
var
  s, addr: string;
  i: integer;
begin
addr:=(sender as Tmenuitem).Caption;
delete(addr, pos('&',addr), 1);

s:='';
for i:=0 to filesBox.SelectionCount-1 do
  s:=s+Tfile(filesBox.Selections[i].data).fullURL('', addr)+CRLF;
setLength(s, length(s)-2);

setClip(s);
end; // copyURLwithAddressMenuClick

procedure TmainFrm.CopyURLwithfingerprint1Click(Sender: TObject);
var
  f: Tfile;
  s, hash: string;
  i: integer;
begin
if selectedFile = NIL then exit;
s:='';
try
  for i:=0 to filesBox.SelectionCount-1 do
    begin
    f:=filesBox.Selections[i].data;

    progFrm.show('Hashing '+f.name, TRUE);
    progFrm.progress:=i / filesBox.SelectionCount;
    application.ProcessMessages();

    hash:=loadFingerprint(f.resource);
    if (hash = '') and f.isFile() then
      begin
      progFrm.push( 1/filesBox.SelectionCount );
      try hash:=createFingerprint(f.resource);
      finally progFrm.pop() end;
      if saveNewFingerprintsChk.checked and (hash > '') then
        saveFile(f.resource+'.md5', hash);
      end;
    if progFrm.cancelRequested then exit;
    s:=s + f.fullURL() + nonEmptyConcat('#!md5!', hash) + CRLF;
    end;
finally progFrm.hide() end;
setLength(s, length(s)-2);

urlBox.Text:=getTill(#13, s);
setClip(s);
end;

procedure Tmainfrm.iconMenuClick(sender:Tobject);
var
  i: integer;
  mi: Tmenuitem;
  fn: string;
begin
if selectedFile = NIL then exit;
mi:=sender as Tmenuitem;
if mi.HelpContext = ICONMENU_NEW then
  begin
  if not promptForFileName(fn) then exit;
  i:=getImageIndexForFile(fn);
  if i < 0 then exit;
  mi.ImageIndex:=i;
  end;
for i:=0 to filesBox.SelectionCount-1 do
  with Tfile(filesBox.Selections[i].data) do
    if not isRoot() then setupImage(mi.ImageIndex);
VFSmodified:=TRUE;
filesBox.Repaint();
end; // iconmenuclick

function createAccountOnTheFly():Paccount;
var
  acc: Taccount;
  i: integer;
begin
result:=NIL;
  repeat
  if not newuserpassFrm.prompt(acc.user, acc.pwd)
  or (acc.user = '') then exit;

  if getAccount(acc.user) = NIL then break;
  msgDlg('Username already exists', MB_ICONERROR)
  until false;
acc.enabled:=TRUE;
acc.noLimits:=FALSE;
i:=length(accounts);
setLength(accounts, i+1);
accounts[i]:=acc;
result:=@accounts[i];
end; // createAccountOnTheFly

procedure Tmainfrm.uploadMenuClick(sender:Tobject);
var
  mi: Tmenuitem;
  itWas, itIs: boolean;
  i: integer;
  username: string;
  acc: Paccount;
begin
if selectedFile = NIL then exit;
mi:=sender as Tmenuitem;
case mi.helpContext of
  UPLOADMENU_ANYONE: username:='*';
  UPLOADMENU_NEW_ACCOUNT:
    begin
    acc:=createAccountOnTheFly();
    if acc = NIL then exit;
    username:=acc.user;
    end;
    
  else username:=mi.Caption;
  end;
for i:=0 to filesBox.SelectionCount-1 do
  with Tfile(filesBox.Selections[i].data) do
    if isRealFolder() then
      begin
      itWas:=uploadAccounts <> NIL;
      if mi.checked then removeString(username, uploadAccounts)
      else addUniqueString(username, uploadAccounts);
      itIs:=uploadAccounts <> NIL;
      if itWas = itIs then continue;
      // WARNING: toggleString() can't be used here
      if itIs then addString(resource, uploadPaths)
      else removeString(resource, uploadPaths);
      end;
VFSmodified:=TRUE;
end; // uploadmenuclick

procedure Tmainfrm.restrictAccessClick(sender:Tobject);
var
  i: integer;
  mi: Tmenuitem;
  all: TStringDynArray;
  b: boolean;
  acc: Paccount;
begin
if selectedFile = NIL then exit;
if fileAttributeInSelection(FA_LINK)
and (msgDlg(MSG_UNPROTECTED_LINKS, MB_ICONWARNING+MB_YESNO) <> IDYES) then exit;

mi:=sender as Tmenuitem;
case mi.helpContext of
  RESTRICTMENU_ALL_ACCOUNTS:
    begin
    all:=NIL;
    for i:=0 to length(accounts)-1 do
      addUniqueString(accounts[i].user, all);
    for i:=0 to filesBox.SelectionCount-1 do
      with Tfile(filesBox.Selections[i].data) do
        begin
        b:=stringExists(USER_ANONYMOUS, accounts);
        accounts:=all;
        if b then addString(USER_ANONYMOUS, accounts); 
        end;
    end;
  RESTRICTMENU_NO_RESTRICTION:  // NO RESTRICTION
    for i:=0 to filesBox.SelectionCount-1 do
      Tfile(filesBox.Selections[i].data).accounts:=NIL;
  RESTRICTMENU_NEW_ACCOUNT:
    begin
    acc:=createAccountOnTheFly();
    if acc = NIL then exit;
    for i:=0 to filesBox.SelectionCount-1 do
      addUniqueString(acc.user, Tfile(filesBox.Selections[i].data).accounts);
    end;
    
  else// specific account and @anonymous
    for i:=0 to filesBox.SelectionCount-1 do
      with Tfile(filesBox.Selections[i].data) do
        if mi.Checked then removeString(mi.caption, accounts)
        else addUniqueString(mi.caption, accounts);
  end;
VFSmodified:=TRUE;
filesBox.Repaint();
end; // restrictAccessClick

procedure TmainFrm.Filesfilter1Click(Sender: TObject);
const
  MSG = 'Please specify a filter that will be applied to files';
var
  i: integer;
  s: string;
begin
if selectedFile = NIL then exit;
s:=selectedFile.filesFilter;
if not inputquery('Filter', MSG, s) then exit;
for i:=0 to filesBox.SelectionCount-1 do
  with Tfile(filesBox.Selections[i].data) do
    if isFolder() then filesFilter:=s;
VFSmodified:=TRUE;
end;

procedure TmainFrm.Foldersfilter1Click(Sender: TObject);
const
  MSG = 'Please specify a filter that will be applied to folders';
var
  i: integer;
  s: string;
begin
if selectedFile = NIL then exit;
s:=selectedFile.foldersFilter;
if not inputquery('Filter', MSG, s) then exit;
for i:=0 to filesBox.SelectionCount-1 do
  with Tfile(filesBox.Selections[i].data) do
    if isFolder() then foldersFilter:=s;
VFSmodified:=TRUE;
end;

procedure TmainFrm.traymessage1Click(Sender: TObject);
begin showOptions(optionsFrm.trayPage) end;

procedure TmainFrm.Guide1Click(Sender: TObject);
begin exec('http://www.rejetto.com/hfs/guide/') end;

procedure Tmainfrm.saveVFS(fn:string='');
begin
if fn = '' then
  begin
  fn:=lastFileOpen;
  if not promptForFileName(fn, 'VirtualFileSystem|*.vfs', 'vfs', 'Save VFS', '', TRUE) then
    exit;
  end;
lastFileOpen:=fn;
if not savefile(fn, addVFSheader(getVFS())) then
  begin
  msgDlg('Error saving', MB_ICONERROR);
  exit;
  end;
VFSmodified:=FALSE;
loadingVFS.disableAutosave:=FALSE;
end; // saveVFS

procedure TmainFrm.filesBoxAddition(Sender: TObject; Node: TTreeNode);
begin VFSmodified:=TRUE end;

procedure Tmainfrm.beforeQuit();
begin
quitting:=TRUE;
checkVfsOnQuit();
timer.enabled:=FALSE;
if saveOptionsOnExitChk.checked then saveCFG();
// we disconnectAll() before srv.free, so we can purgeConnections()
if assigned(srv) then srv.disconnectAll(TRUE);
purgeConnections();
freeAndNIL(srv);
freeAndNIL(tray);
freeAndNIL(tray_ico);
end; // beforequit

procedure TmainFrm.FormClose(Sender: TObject; var Action: TCloseAction);
begin beforeQuit() end;

procedure TmainFrm.Logfile1Click(Sender: TObject);
const
  MSG = 'This function does not save any previous information to the log file.'
    +#13'Instead, it saves all information that appears in the log box in real-time (from when you click "OK", below).'
    +#13'Specify a filename for the log.'
    +#13'If you leave the filename blank, no log file is saved.'
    +#13
    +#13'Here are some symbols you can use in the filename to split the log:'
    +#13'  %d% -- day of the month (1..31)'
    +#13'  %m% -- month (1..12)'
    +#13'  %y% -- year (2000..)'
    +#13'  %dow% -- day of the week (0..6)'
    +#13'  %w% -- week of the year (1..53)'
    +#13'  %user% -- username surrounded by parenthesis';
begin
InputQuery('Log file', MSG, logFile.filename)
end;

procedure TmainFrm.Font1Click(Sender: TObject);
var
  dlg: TFontDialog;
begin
dlg:=TFontDialog.Create(NIL);
dlg.Font.name:=logFontName;
dlg.Font.size:=logFontSize;
if dlg.Execute then
  begin
  logBox.font.Assign(dlg.Font);
  logFontName:=dlg.Font.Name;
  logFontSize:=dlg.Font.size;
  end;
dlg.free;
end;

procedure TmainFrm.SetURL1Click(Sender: TObject);
const
  MSG = 'Please insert an URL for the link'
    +#13
    +#13'Do not forget to specify http:// or whatever'
    +#13'%%ip%% will be translated to your address';
var
  i: integer;
  s: string;
begin
if selectedFile = NIL then exit;
s:=selectedFile.resource;
if not inputquery('Set URL', MSG, s) then exit;
for i:=0 to filesBox.SelectionCount-1 do
  with Tfile(filesBox.Selections[i].data) do
    if FA_LINK in flags then resource:=s;
VFSmodified:=TRUE;
end;

procedure TmainFrm.Resetuserpass1Click(Sender: TObject);
var
  i: integer;
  f: Tfile;
begin
for i:=0 to filesBox.SelectionCount-1 do
  begin
  f:=filesBox.Selections[i].data;
  usersInVFS.drop(f.user, f.pwd);
  f.user:='';
  f.pwd:='';
  end;
VFSmodified:=TRUE;
filesBox.Repaint();
end;

procedure TmainFrm.Switchtovirtual1Click(Sender: TObject);
var
  f: Tfile;
  under: Ttreenode;
  i: integer;
  bakIcon: integer;
  someLocked: boolean;
begin
if selectedFile = NIL then exit;
someLocked:=FALSE;
for i:=0 to filesBox.SelectionCount-1 do
  with Tfile(filesBox.Selections[i].data) do
    if isRealFolder() and not isRoot() then
      if isLocked() then someLocked:=TRUE
      else
        begin
        bakIcon:=icon;
        f:=Tfile.create(resource);
        under:=node.Parent;
        include(f.flags, FA_VIRTUAL);
        node.Delete();
        addFile(f, under, TRUE);
        f.setupImage(bakIcon);
        f.node.Focused:=TRUE;
        end;
VFSmodified:=TRUE;
if someLocked then msgDlg(MSG_SOME_LOCKED, MB_ICONWARNING);
end;

procedure TmainFrm.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
begin
if confirmexitChk.checked and not windowsShuttingDown and not quitASAP then
  if msgDlg('Quit?', MB_ICONQUESTION+MB_YESNO) = IDNO then
    begin
    canclose:=FALSE;
    exit;
    end;
stopAddingItems:=TRUE;
if lockTimerevent or not applicationFullyInitialized then
  begin
  quitASAP:=TRUE;
  canClose:=FALSE;
  end;
{ it's better to switch off this flag, because some software
{ queried after us may prevent Windows from shutting down, but
{ the flag would stay set, while Windows is no more shutting down }
windowsShuttingDown:=FALSE;
end;

procedure TmainFrm.Customizedrealm2Click(Sender: TObject);
const
  MSG = 'The realm string is shown on the user/pass dialog of the browser.'
    +#13'This realm will be used for selected files and their descendants.';
var
  i: integer;
begin
if selectedFile = NIL then exit;
with selectedFile do
	if not inputquery('Realm', format(MSG, [doubleAmp(name)]), realm) then exit
  else realm:=trim(realm);
for i:=0 to filesBox.SelectionCount-1 do
  Tfile(filesBox.Selections[i].data).realm:=selectedFile.realm;
VFSmodified:=TRUE;
end;

procedure TmainFrm.Loginrealm1Click(Sender: TObject);
const
  MSG = 'The realm string is shown on the user/pass dialog of the browser.'
    +#13'Here you can customize the realm for the login button';
begin
if not inputquery('Login realm', MSG, loginRealm) then exit;
loginRealm:=trim(loginRealm);
end;

procedure TmainFrm.Introduction1Click(Sender: TObject);
begin exec('http://www.rejetto.com/hfs/guide/intro.html') end;

function resetDLs(f:Tfile; par:integer):boolean;
begin f.DLcount:=0; result:=TRUE end;

procedure TmainFrm.Resetfileshits1Click(Sender: TObject);
begin
repaintTray();
rootFile.recursiveApply(resetDLs);
autoupdatedFiles.clear();
VFSmodified:=TRUE;
end;

procedure TmainFrm.persistentconnectionsChkClick(Sender: TObject);
begin
srv.persistentConnections:=persistentconnectionsChk.Checked;
if not srv.persistentConnections then
  Kickidleconnections1Click(NIL);
end;

procedure TmainFrm.Kickidleconnections1Click(Sender: TObject);
var
  i: integer;
begin
i:=0;
while i < srv.conns.count do
  begin
  with ThttpConn(srv.conns[i]) do
    if state = HCS_IDLE then
      begin
      TconnData(data).disconnectReason:='kicked';
      disconnect();
      end;
  inc(i);
  end;
end;

procedure TmainFrm.Connectionsinactivitytimeout1Click(Sender: TObject);
const
  MSG = 'The connection is kicked after a timeout.'
    +#13'Specify in seconds.'
    +#13'Leave blank to get no timeout.';
var
  s:string;
begin
if connectionsInactivityTimeout <= 0 then s:=''
else s:=intToStr(connectionsInactivityTimeout);
if not inputquery('Connection inactivity timeout', MSG, s) then exit;
try connectionsInactivityTimeout:=strToUInt(s)
except msgDlg(MSG_INVALID_VALUE, MB_ICONERROR) end;
end;

procedure TmainFrm.splitHMoved(Sender: TObject);
begin if connPnl.height > 0 then lastGoodConnHeight:=connPnl.height end;

procedure TmainFrm.Clearfilesystem1Click(Sender: TObject);
const
  MSG = 'All changes will be lost'#13'Continue?';
begin
checkIfOnlyCountersChanged();
if VFSmodified and (msgDlg(MSG,MB_ICONQUESTION+MB_YESNO) = IDNO) then exit;
initVFS();
end;

procedure TmainFrm.Defaultfilemask1Click(Sender: TObject);
const
  MSG = 'When a folder is browsed, the default file mask is used to find a '
    +'file to serve in place of the folder page. If no file is found, the '
  	+'folder page is served.';
var
  i: integer;
  s: string;
begin
if selectedFile = NIL then exit;
s:=selectedFile.defaultFileMask;
if not inputquery('Default file mask', MSG, s) then exit;
for i:=0 to filesBox.SelectionCount-1 do
  with Tfile(filesBox.Selections[i].data) do
    if isFolder() then defaultFileMask:=s;
VFSmodified:=TRUE;
end;

procedure TmainFrm.Dontconsiderasadownloadmask1Click(Sender: TObject);
const
  MSG = 'Files matching this filemask are not considered for global downloads '
       +'counter. Moreover they never get tray icon.';
var
  s: string;
  i: integer;
begin
if selectedFile = NIL then exit;
s:=selectedFile.dontCountAsDownloadMask;
if not inputquery('Don''t consider as download', MSG, s) then exit;
for i:=0 to filesBox.SelectionCount-1 do
  with Tfile(filesBox.Selections[i].data) do
    if isFolder() then dontCountAsDownloadMask:=s;
VFSmodified:=TRUE;
end;

function isMultiInstance():boolean;
begin
result:=FALSE;
if monoLib.working then exit;
result:=TRUE;
msgDlg(MSG_SINGLE_INSTANCE, MB_ICONERROR);
end; // isMultiInstance

function isIntegratedInShell():boolean;
begin
result:=(loadregistry('*\shell\Add to HFS\command', '', HKEY_CLASSES_ROOT) > '')
  and (loadregistry('Folder\shell\Add to HFS\command','',HKEY_CLASSES_ROOT) >'')
  and (loadregistry('.vfs', '', HKEY_CLASSES_ROOT) > '')
  and (loadregistry('.vfs\shell\Open\command', '', HKEY_CLASSES_ROOT) > '')
end; // isIntegratedInShell

function integrateInShell():boolean;

  function addToContextMenuFor(kind:string):boolean;
  begin
  // for some time we have to delete the old key
  deleteRegistry(kind+'\shell\HFS', HKEY_CLASSES_ROOT);
  result:=saveRegistry(kind+'\shell\Add to HFS\command','',
    '"'+paramstr(0)+'" "%1"',HKEY_CLASSES_ROOT);
  end;

begin
result:=addToContextMenuFor('*')
  and addToContextMenuFor('Folder')
  and saveregistry('.vfs','','HFS file system', HKEY_CLASSES_ROOT)
  and saveregistry('.vfs\shell\Open\command','','"'+paramstr(0)+'" "%1"',
    HKEY_CLASSES_ROOT)
end; // integrateInShell

procedure disintegrateShell();
begin
deleteRegistry('*\shell\Add to HFS', HKEY_CLASSES_ROOT);
deleteRegistry('*\shell\HFS', HKEY_CLASSES_ROOT);
deleteRegistry('Folder\shell\Add to HFS', HKEY_CLASSES_ROOT);
deleteRegistry('Folder\shell\HFS', HKEY_CLASSES_ROOT);
deleteRegistry('.vfs\shell\Open\command', HKEY_CLASSES_ROOT);
deleteRegistry('.vfs', HKEY_CLASSES_ROOT);
end; // disintegrateShell

procedure uninstall();
const
  BATCH_FILE = 'hfs.uninstall.bat';
  BATCH = 'START "" /WAIT "%s" -q'+CRLF
    +'DEL "%0:s"'+CRLF
    +'DEL %%0'+CRLF;
begin
if isMultiInstance() then exit;
mainfrm.saveoptionsonexitChk.Checked:=FALSE;
disintegrateShell();
deleteCFG();
saveFile(BATCH_FILE, format(BATCH,[paramStr(0)]));
quitASAP:=TRUE;
execNew(BATCH_FILE);
end; // uninstall

procedure Tmainfrm.processParams_before(var params:TStringDynArray;
  allowed:string='');
var
  i, consume: integer;

  procedure optDelay();
  var
    n: integer;
  begin
  consume:=2;
  if i >= length(params)-1 then exit;
  n:=strToInt(params[i+1]);
  if n <= 0 then exit;
  sleep(n*100);
  end; // optDelay

  procedure optIniPath();
  begin
  consume:=2;
  if i >= length(params)-1 then exit;
  cfgPath:=IncludeTrailingPathDelimiter(params[i+1]);
  end; // optIniPath

  procedure optAdditionalIni();
  var
    fn: string;
  begin
  consume:=2;
  if i >= length(params)-1 then exit;
  fn:=params[i+1];
  if not fileExists(fn) then fn:=cfgPath+fn;
  if not fileExists(fn) then exit;
  setcfg(loadFile(fn));
  end;

begin
// MEMO: see if FindCmdLineSwitch() can be useful for the job below 
i:=1;
while i < length(params) do
  begin
  if (length(params[i]) = 2) and (params[i][1] = '-')
  and ((allowed = '') or (pos(params[i][2], allowed) > 0)) then
    begin
    consume:=1; // number of params an option takes
    case params[i][2] of
      'd': optDelay();
      'q': quitASAP:=TRUE;
      'u': uninstall();
      'i': optIniPath();
      'a': optAdditionalIni();
      end;
    for consume:=1 to consume do removeString(params, i);
    continue;
    end;
  inc(i);
  end;
end; // processParams_before

procedure Tmainfrm.processParams_after(var params:TStringDynArray);
begin
popString(params);
if (length(params) = 1) and isExtention(params[0], '.vfs') then
  // note: 2 .vfs files will be treated as any file
  mainfrm.loadVFS(params[0])
else
  { parameters are also passed by other instances via sendMessage().
  { since this operation may require user interaction, it must be queued
  { because those instances wouldn't quit until the dialog is closed. }
  addArray(filesToAddQ, params);
end; // processParams_after

procedure TmainFrm.Numberofloggeduploads1Click(Sender: TObject);
begin setTrayShows('uploads') end;

procedure Tmainfrm.compressReply(cd:TconnData);
const
  BAD_IE_THRESHOLD = 2000; // under this size (few bytes less, really) old IE versions will go nuts with UTF-8 pages
var
  s: string;
  m: TStringDynArray;
begin
if not compressedbrowsingChk.checked then exit;
s:=cd.conn.reply.body;
if s = '' then exit;
if pos('gzip', cd.conn.getHeader('Accept-Encoding')) = 0 then exit;
// workaround for IE6 pre-SP2 bug
if (cd.workaroundForIEutf8  = toDetect) and (cd.agent > '') then
  if (reMatch(cd.agent, '^MSIE ([0-9]+)\.', FALSE, @m) > 0) and (m[1][1] in ['4'..'6']) then
    cd.workaroundForIEutf8:=yes
  else
    cd.workaroundForIEutf8:=no;
s:=ZcompressStr2(s, zcFastest, 31,8,zsDefault);
if (cd.workaroundForIEutf8  = yes) and (length(s) < BAD_IE_THRESHOLD) then exit;
cd.conn.addHeader('Content-Encoding: gzip');
cd.conn.reply.body:=s;
end; // compressReply

procedure TmainFrm.Flagfilesaddedrecently1Click(Sender: TObject);
const
  MSG = 'Enter the number of MINUTES files stay flagged from their addition.'
    +#13'Leave blank to disable.';
var
  s: string;
begin
if filesStayFlaggedForMinutes <= 0 then s:=''
else s:=intToStr(filesStayFlaggedForMinutes);
if InputQuery('Flag new files', MSG, s) then
	try
  	s:=trim(s);
  	if s = '' then filesStayFlaggedForMinutes:=0
    else filesStayFlaggedForMinutes:=strToInt(s);
  except msgDlg(MSG_INVALID_VALUE, MB_ICONERROR)
  end;
end;

procedure TmainFrm.Flagasnew1Click(Sender: TObject);
var
  i: integer;
begin
if selectedFile = NIL then exit;
for i:=0 to filesBox.SelectionCount-1 do
  Tfile(filesBox.Selections[i].data).atime:=now();
VFSmodified:=TRUE;
end;

function removeFlagNew(f:Tfile; par:integer):boolean;
begin
result:=TRUE;
f.atime:=now()-succ(filesStayFlaggedForMinutes)/(24*60)
end; // removeFlagNew

procedure TmainFrm.Resetnewflag1Click(Sender: TObject);
var
  i: integer;
begin
if selectedFile = NIL then exit;
for i:=0 to filesBox.SelectionCount-1 do
  Tfile(filesBox.Selections[i].data).recursiveApply(removeFlagNew);
VFSmodified:=TRUE;
end;

procedure Tmainfrm.setStatusBarText(s:string; lastFor:integer);
begin
with sbar.Panels[sbar.Panels.Count-1] do
  begin
  Alignment:=taLeftJustify;
  Text:=s;
  end;
sbarTextTimeout:=now()+lastFor/SECONDS;
end;

procedure TmainFrm.Donate1Click(Sender: TObject);
begin exec('http://www.rejetto.com/hfs-donate') end;

procedure TmainFrm.Donotlogaddress1Click(Sender: TObject);
const
  MSG = 'Any event from the following IP address mask will be not logged.';
begin
InputQuery('Do not log address', MSG, dontLogAddressMask)
end;

procedure TmainFrm.Custom1Click(Sender: TObject);
const
  MSG = 'Specify your addresses, each per line';
var
  s: string;
begin
s:=join(CRLF, customIPs);
if inputQueryLong('Custom IP addresses', MSG, s) then
  begin
  customIPs:=split(CRLF, s);
  removeStrings('', customIPs);
  setDefaultIP(defaultIP); // change the address if it is not available anymore
  end;
end;

procedure TmainFrm.Findexternaladdress1Click(Sender: TObject);
begin
if getExternalAddress(externalIP) then
  begin
  setDefaultIP(externalIP);
  msgDlg(externalIP);
  end
else msgDlg(MSG_ERROR_EXTERNAL, MB_ICONERROR);
end;

procedure TmainFrm.sbarDblClick(Sender: TObject);
var
  i: integer;
begin
i:=whatStatusPanel(sbar,sbar.ScreenToClient(mouse.CursorPos).X);
if (i = sbarIdxs.totalIn) or (i = sbarIdxs.totalOut) then
  if msgDlg('Do you want to reset total in/out?', MB_YESNO) = IDYES then
    begin
    outTotalOfs:=-srv.bytesSent;
    inTotalOfs:=-srv.bytesReceived;
    end;
if i = sbarIdxs.banStatus then BannedIPaddresses1Click(NIL);
if i = sbarIdxs.customTpl then EditHTMLtemplate1Click(NIL);
if i = sbarIdxs.oos then Minimumdiskspace1Click(NIL);
if i = sbarIdxs.out then Speedlimit1Click(NIL);
if i = sbarIdxs.notSaved then Savefilesystem1Click(NIL); 
end;

procedure TmainFrm.sbarMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
// since right click is not used for now, it will act as double click
if button = mbRight then
  sbarDblClick(sender);
end;

procedure forceDynDNSupdate(url:string='');
begin
if url > '' then dyndns.url:=url;
dyndns.active:=TRUE;
dyndns.lastIP:='';
externalIP:='';
end; // forceDynDNSupdate

procedure TmainFrm.Custom2Click(Sender: TObject);
const
  MSG = 'Enter URL for updating.'
    +#13'%ip% will be translated to your external IP.';
begin
if inputQuery('Enter URL', MSG, dyndns.url) then forceDynDNSupdate();
end;

function dynDNSinputUserPwd():boolean;
begin
result:=inputQuery('Enter user', 'Enter user', dyndns.user)
  and (dyndns.user > '')
  and inputQuery('Enter password', 'Enter password', dyndns.pwd)
  and (dyndns.pwd > '')
end; // dynDNSinputUserPwd

function dynDNSinputHost():boolean;
begin
result:=FALSE;
while true do
  begin
  if not inputQuery('Enter host', 'Enter domain (full form!)', dyndns.host)
  or (dyndns.host = '') then exit;
  if pos('://', dyndns.host) > 0 then
    chop('://', dyndns.host);
  if pos('.', dyndns.host) > 0 then
    begin
    result:=TRUE;
    exit;
    end;
  msgDlg('Please, enter it in the FULL form, with dots', MB_ICONERROR);
  end;
end; // dynDNSinputHost

procedure TmainFrm.NoIPtemplate1Click(Sender: TObject);
begin
if not dynDNSinputUserPwd() or not dynDNSinputHost() then exit;
forceDynDNSupdate('http://dynupdate.no-ip.com/dns?username='+dyndns.user
  +'&password='+dyndns.pwd+'&hostname='+dyndns.host+'&ip=%ip%');
end;

procedure TmainFrm.CJBtemplate1Click(Sender: TObject);
begin
if not dynDNSinputUserPwd() then exit;
forceDynDNSupdate('http://www.cjb.net/cgi-bin/dynip.cgi?username='+dyndns.user
  +'&password='+dyndns.pwd+'&ip=%ip%');
dyndns.host:=dyndns.user+'.cjb.net';
end;

procedure TmainFrm.DynDNStemplate1Click(Sender: TObject);
begin
if not dynDNSinputUserPwd() or not dynDNSinputHost() then exit;
forceDynDNSupdate('http://'+dyndns.user+':'+dyndns.pwd
  +'@members.dyndns.org/nic/update?hostname='+dyndns.host
  +'&myip=%ip%&wildcard=NOCHG&backmx=NOCHG&mx=NOCHG&system=dyndns');
end;

procedure TmainFrm.Whyisuploaddisabled1Click(Sender: TObject);
begin msgDLg(getRes('uploadDisabled')) end;

procedure TmainFrm.Minimumdiskspace1Click(Sender: TObject);
const
  MSG = 'The upload will fail if your disk has less than the specified amount of free MegaBytes.';
var
  s: string;
begin
if minDiskSpace <= 0 then s:=''
else s:=intToStr(minDiskSpace);
if InputQuery('Min disk space', MSG, s) then
	try
  	s:=trim(s);
  	if s = '' then minDiskSpace:=0
    else minDiskSpace:=strToInt(s);
  except msgDlg(MSG_INVALID_VALUE, MB_ICONERROR)
  end;
end;

function Tmainfrm.ipPointedInLog():string;
var
  i: integer;
  s: string;
begin
s:=logbox.Lines[logbox.caretpos.y];
i:=logbox.CaretPos.X;
while (i > 1) and (s[i] <> ' ') do dec(i);
delete(s,1,i);
s:=chop(' ',s);
if pos('@', s) > 0 then chop('@',s);
result:=trim(chop(':', s))
end; // ipPointedInLog

procedure TmainFrm.Banthisaddress1Click(Sender: TObject);
begin banAddress(ipPointedInLog()) end;

procedure TmainFrm.Address2name1Click(Sender: TObject);
begin showOptions(optionsFrm.a2nPage) end;

procedure TmainFrm.Addresseseverconnected1Click(Sender: TObject);
begin
if modalOptionsChk.checked then ipsEverFrm.ShowModal()
else ipsEverFrm.show()
end;

procedure TmainFrm.Renamepartialuploads1Click(Sender: TObject);
const
  MSG = 'This string will be appended to the filename.'
    +#13
    +#13'If you need more control, enter a string with %name% in it, and this symbol will be replaced by the original filename.';
begin
InputQuery('Rename partial uploads', MSG, renamePartialUploads)
end;

procedure TmainFrm.SelfTest1Click(Sender: TObject);
const
  MSG_BEFORE = 'Here you can test if your server does work on the Internet.'
    +#13'If you are not interested in serving files over the Internet, this is not for you.'
    +#13
    +#13'We''ll now perform a test involving network activity.'
    +#13'In order to complete this test, you may need to allow HFS''s activity in your firewall, by clicking Allow on the warning prompt.';
  MSG_OK = 'The test is successful. The server should be working fine.';
  MSG_3 = 'You may be behind a router or firewall.';
  MSG_6 = 'You are behind a router.'
    +#13'Ensure it is configured to forward port %s to your computer';
  MSG_7 = 'You may be behind a firewall.'
    +#13'Ensure nothing is blocking HFS.';

  function doTheTest(host:string; port:string=''):string;
  begin
  result:='';
  if progFrm.cancelRequested then exit;
  if port = '' then port:=srv.port;
  progFrm.show('Testing '+host+':'+port+' ...', TRUE);
  if not srv.active and not startServer() then exit;
    repeat
    selfTesting:=TRUE;
    try result:=httpGet('www.rejetto.com/hfs/test.php?port='+port+'&host='+host
      +'&natted='+YESNO[localIPlist.IndexOf(externalIP)<0] );
    finally selfTesting:=FALSE end;

    if (result ='') or (result[1] <> '4') or progFrm.cancelRequested then break;
    sleep(1000);
    until progFrm.cancelRequested;
  end; // doTheTest

  function successful(s:string):boolean;
  begin result:=(s > '') and (s[1] = '1') end;

var
  best: record host, res: string; end;

  procedure tryDifferentHosts();
  var
    i: integer;
    tries: TStringDynArray;
    s: string;
  begin
  if externalIP = '' then
    begin
    progFrm.show('Retrieving external address...');
    getExternalAddress(externalIP);
    end;
  tries:=getPossibleAddresses();
  // ensure defaultIP is the first one 
  insertString(defaultIP, 0, tries);
  uniqueStrings(tries);
  // remove local addresses
  for i:=length(tries)-1 downto 0 do
    if isLocalIP(tries[i]) then removeString(tries, i);

  best.res:='';
  for i:=0 to length(tries)-1 do
    begin
    progFrm.progress:=succ(i)/succ(length(tries));
    s:=doTheTest(tries[i]);
    if s = '' then continue;
    if (best.res = '') or (best.res[1] > s[1]) then
      begin
      best.res:=s;
      best.host:=tries[i];
      if successful(s) then break;
      end;
    end;
  end; // tryDifferentHosts

  procedure tryDifferentPorts();
  var
    i: integer;
    tries: TStringDynArray;
    bak: record
      port: string;
      active: boolean;
      end;
    s: string;
  begin
  // build list of ports we'll test
  tries:=NIL;
  addArray(tries, ['80','81','8000','8080','1','9000','1000','7000']);
  removeString(srv.port, tries); // already tested

  bak.active:=srv.active;
  bak.port:=port;
  for i:=0 to length(tries)-1 do
    begin
    progFrm.progress:=succ(i)/succ(length(tries));
    port:=tries[i];
    stopServer();
    if not startServer() then continue;
    s:=doTheTest(defaultIP);
    if successful(s) then break;
    end;
  if successful(s) or (best.res = '') then
    begin
    best.res:=s;
    best.host:=defaultIP;
    end
  else
    begin
    port:=bak.port;
    stopServer();
    if bak.active then startServer();
    end;
  end; // tryDifferentPorts

var
  s: string;
begin
if msgDlg(MSG_BEFORE, MB_ICONWARNING+MB_OKCANCEL) <> IDOK then exit;

if not srv.active and not startServer() then
  begin
  port:='';
  if not startServer() then
    begin
    msgDlg('Unable to switch the server on', MB_ICONERROR);
    exit;
    end;
  end;

disableUserInteraction();
progFrm.show('Self testing...');
try
  best.res:='';
  progFrm.push(0.5);
  tryDifferentHosts();
  progFrm.pop();

  progFrm.push(0.5);
  if not successful(best.res) then tryDifferentPorts();
  progFrm.pop();
finally
  reenableUserInteraction();
  progFrm.hide();
  end;

s:=best.res;
if successful(s) then
  begin
  msgDlg(MSG_OK);
  if best.host <> defaultIP then setDefaultIP(best.host);
  exit;
  end;

if progFrm.cancelRequested then
  begin
  msgDlg('Test cancelled');
  exit;
  end;

// error
if s = '' then
  if httpGet(ALWAYS_ON_WEB_SERVER+'/test') = '' then
    s:='Your internet connection does not work'
  else
    s:='Sorry, the test is unavailable at the moment'
else
  begin
  case s[1] of
    '3': s:=MSG_3;
    '6': s:=format(MSG_6, [port]);
    '7': s:=MSG_7;
    end;
  s:='The test failed: server does not answer.'#13#13+s;
  end;
msgDlg(s, MB_ICONERROR);
end;

procedure TmainFrm.Opendirectlyinbrowser1Click(Sender: TObject);
const
  MSG = '"Suggest" the browser to open directly the specified files.'
    +#13'Other files should pop up a save dialog.';
begin
InputQuery('Open directly in browser', MSG, openInBrowser)
end;

procedure TmainFrm.noPortInUrlChkClick(Sender: TObject);
const
  MSG = 'You should not use this option unless you really know its meaning.'
    +#13'Continue?';
begin
if noPortInUrlChk.Checked and (msgDlg(MSG, MB_YESNO) = ID_YES) then
  mainfrm.updateUrlBox()
else
  noPortInUrlChk.Checked:=FALSE;
end;

procedure TmainFrm.Editresource1Click(Sender: TObject);
const
  CAPTION = 'Edit resource';
var
  oldRes, oldName, res: string;
  done, nameSync: boolean;
begin
if (selectedFile = NIL) or (FA_VIRTUAL in selectedFile.flags) then exit;
res:=selectedFile.resource;
oldRes:=res;
oldName:=selectedFile.name;
// name sync, only if the name was not customized
nameSync:= selectedFile.name = ExtractFileName(selectedFile.resource);
if selectedFile.isFolder then done:=selectFolder(CAPTION, res)
else done:=PromptForFileName(res, '', '', CAPTION);
if done then VFSmodified:=TRUE;
selectedFile.setResource(res);
if not nameSync then selectedFile.setName(oldName);
selectedFile.setupImage();
end;

procedure TmainFrm.modeBtnClick(Sender: TObject);
begin setEasyMode(not easyMode) end;

procedure TmainFrm.Shellcontextmenu1Click(Sender: TObject);
begin
if isIntegratedInShell() then disintegrateShell()
else if integrateInShell() then msgDlg(MSG_ADD_TO_HFS)
else msgDlg(MSG_ERROR_REGISTRY, MB_ICONERROR);
end;

procedure TmainFrm.menuBtnClick(Sender: TObject);
begin popupMainMenu() end;

procedure TmainFrm.menuPopup(Sender: TObject);
var
  i: integer;
begin
refreshIPlist();
for i:=1 to Fingerprints1.Count-1 do
  Fingerprints1.items[i].Enabled:=fingerprintsChk.checked;

SwitchON1.imageIndex:=ifThen(srv.active, 11, 4);
SwitchON1.caption:='Switch '+ifThen(srv.active, 'OFF', 'ON');

Showbandwidthgraph1.Visible:=not graphBox.Visible;
Shellcontextmenu1.Checked:=isIntegratedInShell();
Restore1.visible:=trayed;
Numberofcurrentconnections1.Checked:= trayShows='connections';
Numberofloggedhits1.checked:= trayShows='hits';
Numberofloggeddownloads1.checked:= trayShows='downloads';
Numberofloggeduploads1.Checked:= trayShows='uploads';
NumberofdifferentIPaddresses1.Checked:= trayShows='ips';
NumberofdifferentIPaddresseseverconnected1.Checked:= trayShows='ips-ever';
ondownloadChk.checked:= flashOn='download';
onconnectionChk.checked:= flashOn='connection';
never1.checked:= flashOn='';
defaultToVirtualChk.Checked:= addFolderDefault='virtual';
defaultToRealChk.Checked:= addFolderDefault='real';
askFolderKindChk.Checked:= addFolderDefault='';
name1.Checked:= TRUE;
time1.Checked:= defSorting='time';
size1.checked:= defSorting='size';
hits1.Checked:= defSorting='hits';
Renamepartialuploads1.Enabled:=not deletePartialUploadsChk.checked;
Seelastserverresponse1.Visible:= dyndns.lastResult>'';
Disable1.Visible:= dyndns.url>'';
RunHFSwhenWindowsstarts1.Checked:= fileExists(startupFilename);
// point out where the options will automatically be saved
tofile1.Default:= saveMode=SM_FILE;
toregistrycurrentuser1.default:= saveMode=SM_USER;
toregistryallusers1.Default:= saveMode=SM_SYSTEM;

Addfiles2.Visible:=easyMode;
Addfolder2.Visible:=easyMode;
Defaultsorting1.Visible:=not easyMode;
sendHFSidentifierChk.Visible:=not easyMode;
URLencoding1.Visible:=not easyMode;
persistentconnectionsChk.Visible:=not easyMode;
DMbrowserTplChk.Visible:=not easyMode;
MIMEtypes1.Visible:=not easyMode;
compressedbrowsingChk.Visible:=not easyMode;
modalOptionsChk.Visible:=not easyMode;
Allowedreferer1.visible:=not easyMode;
Fingerprints1.Visible:=not easyMode;
findExtOnStartupChk.Visible:=not easyMode;
listfileswithsystemattributeChk.Visible:=not easyMode;
Custom1.Visible:=not easyMode;
noPortInUrlChk.Visible:=not easyMode;
DynamicDNSupdater1.Visible:=not easyMode;
only1instanceChk.Visible:=not easyMode;
Flashtaskbutton1.Visible:=not easyMode;
HintsfornewcomersChk.Visible:=not easyMode;
Graphrefreshrate1.Visible:=not easyMode;
foldersbeforeChk.Visible:=not easyMode;
listfileswithhiddenattributeChk.Visible:=not easyMode;
saveTotalsChk.Visible:=not easyMode;
trayfordownloadChk.Visible:=not easyMode;
Accounts1.visible:=not easyMode;
VirtualFileSystem1.Visible:=not easyMode;
Pausestreaming1.Visible:=not easyMode;
Maxconnections1.Visible:=not easyMode;
Maxconnectionsfromsingleaddress1.Visible:=not easyMode;
maxIPsDLing1.Visible:=not easyMode;
maxIPs1.Visible:=not easyMode;
MaxDLsIP1.Visible:=not easyMode;
Connectionsinactivitytimeout1.Visible:=not easyMode;
minimumDiskSpace1.Visible:=not easyMode;
editHTMLtemplate1.Visible:=not easyMode;
shellcontextmenu1.Visible:=not easyMode;
useCommentAsRealmChk.Visible:=not easyMode;
openDirectlyInBrowser1.Visible:=not easyMode;
keepBakUpdatingChk.Visible:=not easyMode;
loginRealm1.Visible:=not easyMode;
DumprequestsChk.Visible:=not easyMode;
logBytesreceivedChk.Visible:=not easyMode;
logBytessentChk.Visible:=not easyMode;
logconnectionsChk.Visible:=not easyMode;
logDisconnectionsChk.Visible:=not easyMode;
autoCommentChk.Visible:=not easyMode;
traymessage1.Visible:=not easyMode;
showmaintrayiconChk.Visible:=not easyMode;
numberOfLoggedHits1.Visible:=not easyMode;
Showcustomizedoptions1.Visible:=not easyMode;
enableNoDefaultChk.visible:=not easyMode;
browseUsingLocalhostChk.visible:=not easyMode;
useISOdateChk.visible:=not easyMode;
Addicons1.Visible:=not easyMode;
Acceptconnectionson1.Visible:=not easyMode;
numberFilesOnUploadChk.Visible:=not easyMode;
Renamepartialuploads1.Visible:=not easyMode;
deletePartialUploadsChk.Visible:=not easyMode;
end;

function Tmainfrm.finalInit():boolean;

  function paramsAsArray():TStringDynArray;
  var
    i: integer;
  begin
  i:=paramCount();
  setlength(result, i+1);
  for i:=0 to i do result[i]:=paramStr(i);
  end; // paramsAsArray

  function getBrowserPath():string;
  var
    i: integer;
  begin
  result:=loadRegistry('HTTP\shell\open\command', '', HKEY_CLASSES_ROOT);
  if result = '' then exit;
  if result[1] = '"' then
    begin
    result:=copy(result, 2, posEx('"', result, 2)-2);
    exit;
    end;
  i:=pos(' ', result);
  if i = 0 then exit;  
  delete(result, i, MAXINT);  
  end; // getBrowserPath

var
  cfgLoaded: boolean;
  params: TStringDynArray;
  i: integer;
begin
result:=FALSE;
// some windows versions do not support multiline tray tips
if winVersion < WV_2000 then trayNL:='  ';

trayMsg:='%ip%'
  +trayNL+'Uptime: %uptime%'
  +trayNL+'Downloads: %downloads%';


startingImagesCount:=mainfrm.images.count;
srv:=ThttpSrv.create;
srv.autoFreeDisconnectedClients:=FALSE;
srv.onEvent:=httpEvent;
//with toolbar2.Font do style:=style+[fsBold];
i:=getImageIndexForFile(getBrowserPath());
browseBtn.ImageIndex:=i;
tray_ico:=Ticon.create;
tray:=Ttrayicon.create(self);
DragAcceptFiles(handle, true);
caption:=format('HFS ~ HTTP File Server %s%sBuild %s',
  [VERSION, stringOfChar(' ',80), VERSION_BUILD]);
application.Title:=format('HFS %s (%s)', [VERSION, VERSION_BUILD]);
setSpeedLimit(-1);
setGraphRate(10);
setMaxConnections(0);
setMaxConnectionsIP(0);
setMaxDLs(0);
setMaxDLsIP(0);
setMaxIPs(0);
setMaxIPsDLing(0);
setnoDownloadTimeout(0);
setAutosave(autosaveCFG, 0);
setAutosave(autosaveVFS, 0);
setAutoFingerprint(0);

autosaveVFS.minimum:=5;
autosaveVFS.menu:=autosaveevery1;
autosaveCFG.minimum:=5;
autosaveCFG.menu:=autosaveevery2;

params:=paramsAsArray();
processParams_before(params, 'i');

initVFS();
setFilesBoxExtras(winVersion <> WV_VISTA);

defaultCfg:=getCfg();

cfgLoaded:=FALSE;
// if SHIFT is pressed skip configuration loading
if not holdingKey(VK_SHIFT) then
  cfgLoaded:=loadCFG();

// CTRL avoids the only1instance setting
if not holdingKey(VK_CONTROL)
and only1instanceChk.checked and not monoLib.isMaster() then
  begin
  monoLib.sendParams();
  result:=FALSE;
  quitASAP:=TRUE;
  end;

processParams_before(params);

if not quitASAP then
  begin

  if not cfgLoaded then
    begin
    startServer();
    if not isIntegratedInShell() and (shellExtFrm.showModal() = mrYes) then
      if not integrateInShell() then
        msgDlg(MSG_ERROR_REGISTRY, MB_ICONERROR);
    end;

  if findExtOnStartupChk.checked and getExternalAddress(externalIP) then
    setDefaultIP(externalIP);

  end;

// no address set or not available anymore
if not stringExists(defaultIP, getPossibleAddresses()) then
  setDefaultIP(getIP());

progFrm:=TprogressForm.create;
progFrm.preventBackward:=TRUE;
updateUrlBox();
application.HintPause:=100;
splitV.AutoSnap:=FALSE;
splitV.AutoSnap:=TRUE;
splitV.update;
graph.size:=graphBox.height;
if not quitASAP then
  begin
  if autocopyURLonStartChk.Checked then
    setClip( rootFile.fullURL() );
  if reloadonstartupChk.checked and (lastFileOpen > '') then
    if fileExists(lastFileOpen) then loadVFS(lastFileOpen)
    else lastFileOpen:='';
  end;
processParams_after(params);

if not quitASAP then
  begin
  if not cfgLoaded then
    setEasyMode(easyMode);
  tray.setIcon(tray_ico);
  tray.onEvent:=trayEvent;
  if showmaintrayiconChk.checked then addTray();
  end;
timer.Enabled:=TRUE;
applicationFullyInitialized:=TRUE;
if not quitASAP then show()
else application.showmainform:=FALSE;
if startminimizedChk.checked then application.Minimize();
if findExtOnStartupChk.checked and (externalIP = '') then
  setStatusBarText('Search for external address failed', 30);
updatePortBtn();
end; // finalInit

procedure TmainFrm.Dontconsiderasdownload1Click(Sender: TObject);
begin fileMenuSetFlag(sender, FA_DONT_COUNT_AS_DL, @Tfile.isFile) end;

procedure TmainFrm.Dontlog1Click(Sender: TObject);
begin fileMenuSetFlag(sender, FA_DONT_LOG, @Tfile.isFileOrFolder) end;

function expertModeNeededMsg():string;
begin result:=ifThen(easyMode, 'Switch to expert mode.') end;

procedure TmainFrm.Dontlogsomefiles1Click(Sender: TObject);
begin
msgDlg(expertModeNeededMsg()
  +#13'Select the files/folder you don''t want to be logged,'
  +#13'then right click and select "Don''t log".');
end;

procedure TmainFrm.progFrmHttpGetUpdate(sender:Tobject; buffer:pointer;
  len:integer);
begin
with sender as ThttpCli do
  begin
  progFrm.progress:=safeDiv(0.0+RcvdCount, contentLength);
  if progFrm.cancelRequested then abort();
  end;
end; // progFrmUpdate

procedure TmainFrm.UninstallHFS1Click(Sender: TObject);
begin
if isMultiInstance() then exit;
if msgDlg('Delete HFS and all settings?',MB_ICONQUESTION+MB_YESNO) <> IDYES then
  exit;
uninstall();
end;

procedure TmainFrm.maxIPs1Click(Sender: TObject);
const
  MSG = 'Max simultaneous addresses.'#13+MSG_EMPTY_NO_LIMIT;
  MSG2 = 'In this moment there are %d different addresses';
var
  s: string;
  i: integer;
begin
if maxIPs > 0 then s:=intToStr(maxIPs)
else s:='';
if inputquery('Max addresses', MSG, s) then
	try setMaxIPs(strToUInt(s))
  except msgDlg(MSG_INVALID_VALUE, MB_ICONERROR)
  end;
if maxIPs = 0 then exit;
i:=countIPs();
if i > maxIPs then
  msgDlg(format(MSG2, [i]), MB_ICONWARNING);
end;

procedure TmainFrm.maxIPsDLing1Click(
  Sender: TObject);
const
  MSG = 'Max simultaneous addresses downloading.'#13+MSG_EMPTY_NO_LIMIT;
  MSG2 = 'In this moment there are %d different addresses downloading';
var
  s: string;
  i: integer;
begin
if maxIPsDLing > 0 then s:=intToStr(maxIPsDLing)
else s:='';
if inputquery('Max addresses downloading', MSG, s) then
	try setMaxIPsDLing(strToUInt(s))
  except msgDlg(MSG_INVALID_VALUE, MB_ICONERROR)
  end;
if maxIPsDLing = 0 then exit;
i:=countIPs(TRUE);
if i > maxIPsDLing then
  msgDlg(format(MSG2, [i]), MB_ICONWARNING);
end;

procedure TmainFrm.Nodownload1Click(Sender: TObject);
var
  i: integer;
begin
if selectedFile = NIL then exit;
for i:=0 to filesBox.SelectionCount-1 do
  with Tfile(filesBox.Selections[i].data) do
    recursiveApply(setDLforbidden, integer((sender as Tmenuitem).checked) );
VFSmodified:=TRUE;
end;

procedure TmainFrm.Autohideemptyfolders1Click(Sender: TObject);
begin fileMenuSetFlag(sender, FA_HIDE_EMPTY_FOLDERS) end;

procedure TmainFrm.Autosaveevery1Click(Sender: TObject);
begin autosaveClick(autosaveVFS, 'file system') end;

procedure TmainFrm.Autosaveevery2Click(Sender: TObject);
begin autosaveClick(autosaveCFG, 'options') end;

procedure TmainFrm.Apachelogfileformat1Click(Sender: TObject);
const
  MSG = 'Here you can specify how to format the log file complying Apache standard.'
    +#13'Leave blank to get bare copy of screen on file.'
    +#13
    +#13'Example:'
    +#13'   %h %l %u %t "%r" %>s %b';
begin
InputQuery('Apache log file format', MSG, logFile.apacheFormat)
end;

procedure TmainFrm.Bindroottorealfolder1Click(Sender: TObject);
var
  f: Tfile;
begin
f:=selectedFile;
if (f = NIL) or not f.isVirtualFolder() or not f.isRoot() then exit;
if not selectFolder('', f.resource) then exit;
f.setResource(f.resource);
exclude(f.flags, FA_VIRTUAL);
VFSmodified:=TRUE;
end;

procedure TmainFrm.Unbindroot1Click(Sender: TObject);
var
  f: Tfile;
begin
f:=selectedFile;
if (f = NIL) or not f.isRealFolder() or not f.isRoot() then exit;
f.setResource('');
include(f.flags, FA_VIRTUAL);
VFSmodified:=TRUE;
end;

procedure TmainFrm.SwitchON1Click(Sender: TObject);
begin toggleServer() end;

procedure TmainFrm.Switchtorealfolder1Click(Sender: TObject);
var
  i: integer;
  someLocked: boolean;
begin
if selectedFile = NIL then exit;
someLocked:=FALSE;
for i:=0 to filesBox.SelectionCount-1 do
  with Tfile(filesBox.Selections[i].data) do
    if isVirtualFolder() and not isRoot() and (resource > '') then
      if isLocked() then someLocked:=TRUE
      else
        begin
        exclude(flags, FA_VIRTUAL);
        setResource(resource);
        setupImage();
        node.DeleteChildren();
        end;
VFSmodified:=TRUE;
if someLocked then msgDlg(MSG_SOME_LOCKED, MB_ICONWARNING);
end;

procedure TmainFrm.abortBtnClick(Sender: TObject);
begin stopAddingItems:=TRUE end;

procedure TmainFrm.Seelastserverresponse1Click(Sender: TObject);
var
  fn: string;
begin
if pos('<html>', lowercase(dyndns.lastResult)) = 0 then
  begin
  msgDlg(dyndns.lastResult);
  exit;
  end;
fn:=saveTempFile(dyndns.lastResult);
if fn = '' then
  begin
  msgDlg(MSG_NO_TEMP, MB_ICONERROR);
  exit;
  end;
renameFile(fn, fn+'.html');
exec(fn+'.html');
end;

procedure TmainFrm.Showcustomizedoptions1Click(Sender: TObject);
var
  default: TStrings;
  current, defV, v, k: string;
  diff: string;
begin
default:=TStringList.create;
default.text:=defaultCfg;
current:=getCfg();
diff:='';
while current > '' do
  begin
  v:=chopLine(current);
  k:=chop('=', v);
  if AnsiEndsStr('-width', k) or AnsiEndsStr('-height', k)
  or stringExists(k, ['active','window','graph-visible','graph-size','ip',
    'accounts', 'dynamic-dns-user', 'ips-ever']) then continue;
  if k = 'dynamic-dns-updater' then // remove login data
    begin
    v:=base64decode(v);
    chop('//',v);
    v:=chop('/',v);
    if ansiContainsStr(v, '@') then chop('@',v); 
    v:='...'+v+'...';
    end;
  defV:=default.values[k];
  if defV <> v then
    diff:=diff+k+'='+v+CRLF+'default: '+defV+CRLF+CRLF;
  end;
default.free;

diffFrm.memoBox.text:=diff;
diffFrm.showModal();
end;

procedure TmainFrm.useISOdateChkClick(Sender: TObject);
begin applyISOdateFormat() end;

procedure TmainFrm.Difftemplate1Click(Sender: TObject);
const
  MSG = 'Here you can put a partial template that will overlap the main one.';
var
  i: integer;
  accept: boolean;
begin
if selectedFile = NIL then exit;
i:=ifThen( selectedFile=lastDiffTpl.f, lastDiffTpl.ofs);
accept:=inputqueryLong('Diff template', MSG, selectedFile.diffTpl, i);
lastDiffTpl.f:=selectedFile;
lastDiffTpl.ofs:=inputQueryLongDlg.inputBox.selstart;
if not accept then exit;
VFSmodified:=TRUE;
for i:=0 to filesBox.SelectionCount-1 do
  with Tfile(filesBox.Selections[i].data) do
    if isFolder() then
      diffTpl:=selectedFile.diffTpl;
end;

procedure TmainFrm.RunHFSwhenWindowsstarts1Click(Sender: TObject);
begin
if (sender as Tmenuitem).Checked then deleteFile(startupFilename)
else createShellLink(startupFilename, paramStr(0))
end;

procedure TmainFrm.minimizeToTray();
begin
application.Minimize();
addTray();
showWindow(application.handle, SW_HIDE); // hide taskbar button
trayed:=TRUE;
end; // minimizeToTray

procedure TmainFrm.askFolderKindChkClick(Sender: TObject);
begin addFolderDefault:='' end;

procedure TmainFrm.defaultToVirtualChkClick(Sender: TObject);
begin addFolderDefault:='virtual' end;

procedure TmainFrm.defaultToRealChkClick(Sender: TObject);
begin addFolderDefault:='real' end;

procedure TmainFrm.Hidefileextentioninlisting1Click(Sender: TObject);
begin fileMenuSetFlag(sender, FA_HIDE_EXT) end;

procedure TmainFrm.Addicons1Click(Sender: TObject);
var
  files: TStringDynArray;
  i, n: integer;
begin
if not selectFiles('', files) then exit;
n:=images.Count;
for i:=0 to length(files)-1 do
  getImageIndexForFile(files[i]);
n:=images.Count-n;
msgDlg(format('%d new icons added',[n]));
end;

procedure TmainFrm.Iconmasks1Click(Sender: TObject);
begin showOptions(optionsFrm.iconsPage) end;

procedure TmainFrm.Anyaddress1Click(Sender: TObject);
begin
listenOn:='';
restartServer();
end;

procedure Tmainfrm.acceptOnMenuclick(sender:Tobject);
begin
listenOn:=(sender as Tmenuitem).caption;
delete(listenOn, pos('&',listenOn), 1);
restartServer();
end; // acceptOnMenuclick

procedure TmainFrm.filesBoxEndDrag(Sender, Target: TObject; X, Y: Integer);
begin
scrollFilesBox:=-1;
filesBox.Refresh();
end;

procedure TmainFrm.filesBoxEnter(Sender: TObject);
begin setFilesBoxExtras(TRUE) end;

procedure TmainFrm.filesBoxExit(Sender: TObject);
begin setFilesBoxExtras(filesBox.MouseInClient) end;

procedure TmainFrm.Disable1Click(Sender: TObject);
begin
dyndns.url:='';
msgDlg('Dynamic DNS updater disabled');
end;

procedure TmainFrm.saveNewFingerprintsChkClick(Sender: TObject);
const
  MSG = 'This option creates an .md5 file for every new calculated fingerprint.'
    +#13'Use with care to get not your disk invaded by these files.';
begin
if saveNewFingerprintsChk.Checked then
  msgDlg(MSG, MB_ICONWARNING);
end;

procedure TmainFrm.Createfingerprintonaddition1Click(Sender: TObject);
const
  MSG = 'When you add files and no fingerprint is found, it is calculated.'
    +#13'To avoid long waitings, set a limit to file size (in KiloBytes).'
    +#13'Leave empty to disable, and have no fingerprint created.';
var
  s: string;
begin
if autoFingerprint = 0 then s:=''
else s:=intToStr(autoFingerprint);
if not inputquery('Auto fingerprint', MSG, s) then exit;
try setAutoFingerprint(strToUInt(s))
except msgDlg(MSG_INVALID_VALUE, MB_ICONERROR) end;
end;

procedure TmainFrm.pwdInPagesChkClick(Sender: TObject);
const
  MSG = 'This feature is INCOMPATIBLE with Internet Explorer.';
begin
if pwdInPagesChk.Checked
and (msgDlg(MSG, MB_ICONWARNING+MB_OKCANCEL) = IDOK) then
  begin
  msgDlg(MSG_ENABLED);
  exit;
  end;
pwdInPagesChk.checked:=FALSE;
msgDlg(MSG_DISABLED)
end;

procedure TmainFrm.Howto1Click(Sender: TObject);
begin msgDLg(getRes('uploadHowTo')) end;

procedure TmainFrm.Name1Click(Sender: TObject);
begin defSorting:='name' end;

procedure TmainFrm.Size1Click(Sender: TObject);
begin defSorting:='size' end;

procedure TmainFrm.Time1Click(Sender: TObject);
begin defSorting:='time' end;

procedure TmainFrm.Hits1Click(Sender: TObject);
begin defSorting:='hits' end;

procedure TmainFrm.Resettotals1Click(Sender: TObject);
begin resetTotals() end;

INITIALIZATION

programStart:=now();
randomize();
exePath:=extractFilePath(ExpandFileName(paramStr(0)));
cfgPath:=exePath;
// we give priority to exePath because some people often clear the temp folder
tmpPath:=exePath;
if savefile(tmpPath+'test.tmp','') then deleteFile(tmpPath+'test.tmp')
else tmpPath:=getTempDir();
chDir(exePath); // sometimes people mess with the working directory, so we force it to the exe path 
lastUpdateCheckFN:=tmpPath+'HFS last update check.tmp';
defaultTpl:=getRes('defaultTpl');
tpl_help:=getRes('tplHlp');
tpl:=Ttpl.create;
tpl.fulltext:=defaultTpl;
dmBrowserTpl:=Ttpl.create;
dmBrowserTpl.fulltext:=getRes('dmBrowserTpl');
filelistTpl:=Ttpl.create;
filelistTpl.fulltext:=getRes('filelistTpl');
trayShows:='downloads';
flashOn:='download';
runningOnRemovable:=DRIVE_REMOVABLE = getDriveType(PansiChar(exePath[1]+':\'));

toDelete:=Tlist.create;
usersInVFS:=TusersInVFS.Create;

tplFrm:=TeditFrm.create(mainfrm);
tplFrm.default:=defaultTpl;
tplFrm.caption:='HTML template';
tplFrm.help:=tpl_help;
tplFrm.onSetText:=tplSetText;

openInBrowser:='*.htm;*.html;*.jpg;*.jpeg;*.gif;*.png;*.txt;*.swf;*.svg';
addArray(MIMEtypes, [
	'*.htm;*.html', 'text/html',
  '*.jpg;*.jpeg;*.jpe', 'image/jpeg',
  '*.gif', 'image/gif',
  '*.png', 'image/png',
  '*.bmp', 'image/bmp',
  '*.ico', 'image/x-icon',
  '*.mpeg;*.mpg;*.mpe', 'video/mpeg',
  '*.avi', 'video/x-msvideo',
  '*.txt', 'text/plain',
  '*.css', 'text/css'
]);

systemimages:=getSystemimages();
saveMode:=SM_USER;
getDir(0, lastAddedFolder);
autoupdatedFiles:=TstringToIntHash.create;
iconsCache:=TiconsCache.create;
dyndns.active:=TRUE;
connectionsInactivityTimeout:=60; // 1 minute
startupFilename:=getShellFolder('Startup')+'\HFS.lnk';

logfile.apacheZoneString:=ifThen(GMToffset < 0, '-','+')
  +format('%.2d%.2d', [abs(GMToffset div 60), abs(GMToffset mod 60)]);

FINALIZATION

progFrm.free;
tplFrm.Free;
toDelete.free;
tpl.free;
filelistTpl.free;
autoupdatedFiles.free;
iconsCache.free;
usersInVFS.free;

end.
